@@ -16,6 +16,61 @@
 //! * Get an [`SvgHandle`] from the [`Loader`].
 //! * Create a [`CairoRenderer`] for the [`SvgHandle`] and render to a Cairo context.
 //!
+//! You can put the following in your `Cargo.toml`:
+//!
+//! ```toml
+//! [dependencies]
+//! librsvg = { git="https://gitlab.gnome.org/GNOME/librsvg" }
+//! cairo-rs = "0.7.0"
+//! glib = "0.8.0"                                    # only if you need streams
+//! gio = { version="0.7.0", features=["v2_48"] }     # likewise
+//! ```
+//!
+//! # Example
+//!
+//! ```
+//!
+//! const WIDTH: i32 = 640;
+//! const HEIGHT: i32 = 480;
+//!
+//! fn main() {
+//!     // Loading from a file
+//!
+//!     let handle = librsvg::Loader::new().read_path("example.svg").unwrap();
+//!
+//!     let surface = cairo::ImageSurface::create(cairo::Format::ARgb32, WIDTH, HEIGHT).unwrap();
+//!     let cr = cairo::Context::new(&surface);
+//!
+//!     let renderer = librsvg::CairoRenderer::new(&handle);
+//!     renderer.render_document(
+//!         &cr,
+//!         &cairo::Rectangle {
+//!             x: 0.0,
+//!             y: 0.0,
+//!             width: f64::from(WIDTH),
+//!             height: f64::from(HEIGHT),
+//!         },
+//!     ).unwrap();
+//!
+//!     // Loading from a static SVG asset
+//!
+//!     let bytes = glib::Bytes::from_static(
+//!         br#"<?xml version="1.0" encoding="UTF-8"?>
+//!             <svg xmlns="http://www.w3.org/2000/svg" width="50" height="50">
+//!                 <rect id="foo" x="10" y="10" width="30" height="30"/>
+//!             </svg>
+//!         "#
+//!     );
+//!     let stream = gio::MemoryInputStream::new_from_bytes(&bytes);
+//!
+//!     let handle = librsvg::Loader::new().read_stream(
+//!         &stream,
+//!         None::<&gio::File>,          // no base file as this document has no references
+//!         None::<&gio::Cancellable>,   // no cancellable
+//!     ).unwrap();
+//! }
+//! ```
+//!
 //! [`Loader`]: struct.Loader.html
 //! [`SvgHandle`]: struct.SvgHandle.html
 //! [`CairoRenderer`]: struct.CairoRenderer.html
@@ -87,10 +142,8 @@
 //! [SVG 2]: https://www.w3.org/TR/SVG2/
 
 #![warn(unused)]
-use cairo;
-use gio;
-use glib::{self, prelude::*};
-use rsvg_internals;
+
+use glib::prelude::*;
 use url::Url;
 
 use std::path::Path;
@@ -100,13 +153,8 @@ use gio::{Cancellable, FileExt};
 use rsvg_internals::{Dpi, Handle, LoadOptions};
 
 pub use rsvg_internals::{
-    DefsLookupErrorKind,
-    HrefError,
-    Length as InternalLength,
-    LengthUnit,
-    LoadingError,
-    RenderingError,
-    RsvgLength as Length,
+    DefsLookupErrorKind, HrefError, Length as InternalLength, LengthUnit, LoadingError,
+    RenderingError, RsvgLength as Length,
 };
 
 /// Builder for loading an [`SvgHandle`][SvgHandle].
@@ -191,8 +239,6 @@ impl Loader {
     /// ```ignore
     /// # // Test is ignored because "make distcheck" breaks, as the output file
     /// # // can't be written to the read-only srcdir.
-    /// use cairo;
-    /// use librsvg;
     ///
     /// let svg_handle = librsvg::Loader::new()
     ///     .keep_image_data()
@@ -218,8 +264,6 @@ impl Loader {
     /// # Example:
     ///
     /// ```
-    /// use librsvg;
-    ///
     /// let svg_handle = librsvg::Loader::new()
     ///     .read_path("example.svg")
     ///     .unwrap();
@@ -235,9 +279,6 @@ impl Loader {
     ///
     /// # Example:
     /// ```
-    /// use gio;
-    /// use librsvg;
-    ///
     /// let svg_handle = librsvg::Loader::new()
     ///     .read_file(&gio::File::new_for_path("example.svg"), None::<&gio::Cancellable>)
     ///     .unwrap();
@@ -268,8 +309,6 @@ impl Loader {
     ///
     /// ```
     /// use gio::prelude::*;
-    /// use gio;
-    /// use librsvg;
     ///
     /// let file = gio::File::new_for_path("example.svg");
     ///
@@ -407,12 +446,7 @@ impl<'a> CairoRenderer<'a> {
         IntrinsicDimensions {
             width: d.width.map(Into::into),
             height: d.height.map(Into::into),
-            vbox: d.vbox.map(|v| cairo::Rectangle {
-                x: v.x,
-                y: v.y,
-                width: v.width,
-                height: v.height,
-            }),
+            vbox: d.vbox.map(|v| cairo::Rectangle::from(v.0)),
         }
     }
 
