@@ -2,18 +2,9 @@
 extern crate criterion;
 use criterion::Criterion;
 
-use cairo;
-use cairo_sys;
-use rsvg_internals;
-
 use rsvg_internals::rect::IRect;
 use rsvg_internals::surface_utils::shared_surface::{
-    AlphaOnly,
-    Horizontal,
-    NotAlphaOnly,
-    SharedImageSurface,
-    SurfaceType,
-    Vertical,
+    AlphaOnly, Horizontal, NotAlphaOnly, SharedImageSurface, SurfaceType, Vertical,
 };
 
 const SURFACE_SIDE: i32 = 512;
@@ -36,7 +27,7 @@ fn bench_box_blur(c: &mut Criterion) {
             } else {
                 SurfaceType::SRgb
             };
-            let input_surface = SharedImageSurface::new(input_surface, surface_type).unwrap();
+            let input_surface = SharedImageSurface::wrap(input_surface, surface_type).unwrap();
 
             let mut output_surface =
                 cairo::ImageSurface::create(cairo::Format::ARgb32, SURFACE_SIDE, SURFACE_SIDE)
