@@ -1,4 +1,4 @@
-//! Handling of `preserveAspectRatio` values
+//! Handling of `preserveAspectRatio` values.
 //!
 //! This module handles `preserveAspectRatio` values [per the SVG specification][spec].
 //! We have an [`AspectRatio`] struct which encapsulates such a value.
@@ -22,14 +22,11 @@
 
 use std::ops::Deref;
 
-use cairo;
-
-use crate::error::ValueErrorKind;
-use crate::float_eq_cairo::ApproxEqCairo;
+use crate::error::*;
 use crate::parsers::Parse;
-use crate::parsers::ParseError;
+use crate::rect::Rect;
 use crate::viewbox::ViewBox;
-use cssparser::{CowRcStr, Parser};
+use cssparser::{BasicParseError, Parser};
 
 #[derive(Debug, Copy, Clone, PartialEq, Eq)]
 pub struct AspectRatio {
@@ -125,25 +122,29 @@ impl AspectRatio {
         }
     }
 
-    pub fn compute(&self, vbox: &ViewBox, viewport: &cairo::Rectangle) -> (f64, f64, f64, f64) {
+    pub fn compute(&self, vbox: &ViewBox, viewport: &Rect) -> Rect {
         match self.align {
-            None => (viewport.x, viewport.y, viewport.width, viewport.height),
+            None => *viewport,
 
             Some(Align { x, y, fit }) => {
-                let w_factor = viewport.width / vbox.width;
-                let h_factor = viewport.height / vbox.height;
+                let (vb_width, vb_height) = vbox.0.size();
+                let (vp_width, vp_height) = viewport.size();
+
+                let w_factor = vp_width / vb_width;
+                let h_factor = vp_height / vb_height;
+
                 let factor = match fit {
                     FitMode::Meet => w_factor.min(h_factor),
                     FitMode::Slice => w_factor.max(h_factor),
                 };
 
-                let w = vbox.width * factor;
-                let h = vbox.height * factor;
+                let w = vb_width * factor;
+                let h = vb_height * factor;
 
-                let xpos = x.compute(viewport.x, viewport.width, w);
-                let ypos = y.compute(viewport.y, viewport.height, h);
+                let xpos = x.compute(viewport.x0, vp_width, w);
+                let ypos = y.compute(viewport.y0, vp_height, h);
 
-                (xpos, ypos, w, h)
+                Rect::new(xpos, ypos, xpos + w, ypos + h)
             }
         }
     }
@@ -153,7 +154,7 @@ impl AspectRatio {
     pub fn viewport_to_viewbox_transform(
         &self,
         vbox: Option<ViewBox>,
-        viewport: &cairo::Rectangle,
+        viewport: &Rect,
     ) -> Option<cairo::Matrix> {
         // width or height set to 0 disables rendering of the element
         // https://www.w3.org/TR/SVG/struct.html#SVGElementWidthAttribute
@@ -161,88 +162,74 @@ impl AspectRatio {
         // https://www.w3.org/TR/SVG/struct.html#ImageElementWidthAttribute
         // https://www.w3.org/TR/SVG/painting.html#MarkerWidthAttribute
 
-        if viewport.width.approx_eq_cairo(0.0) || viewport.height.approx_eq_cairo(0.0) {
+        if viewport.is_empty() {
             return None;
         }
 
         // the preserveAspectRatio attribute is only used if viewBox is specified
         // https://www.w3.org/TR/SVG/coords.html#PreserveAspectRatioAttribute
         if let Some(vbox) = vbox {
-            if vbox.width.approx_eq_cairo(0.0) || vbox.height.approx_eq_cairo(0.0) {
+            if vbox.0.is_empty() {
                 // Width or height of 0 for the viewBox disables rendering of the element
                 // https://www.w3.org/TR/SVG/coords.html#ViewBoxAttribute
                 None
             } else {
-                let (x, y, w, h) = self.compute(&vbox, viewport);
+                let r = self.compute(&vbox, viewport);
                 let mut matrix = cairo::Matrix::identity();
-                matrix.translate(x, y);
-                matrix.scale(w / vbox.width, h / vbox.height);
-                matrix.translate(-vbox.x, -vbox.y);
+                matrix.translate(r.x0, r.y0);
+                matrix.scale(r.width() / vbox.0.width(), r.height() / vbox.0.height());
+                matrix.translate(-vbox.0.x0, -vbox.0.y0);
                 Some(matrix)
             }
         } else {
             let mut matrix = cairo::Matrix::identity();
-            matrix.translate(viewport.x, viewport.y);
+            matrix.translate(viewport.x0, viewport.y0);
             Some(matrix)
         }
     }
 }
 
-fn parse_align_xy(ident: &CowRcStr) -> Result<Option<(X, Y)>, ValueErrorKind> {
+fn parse_align_xy<'i>(parser: &mut Parser<'i, '_>) -> Result<Option<(X, Y)>, BasicParseError<'i>> {
     use self::Align1D::*;
 
-    match ident.as_ref() {
-        "none" => Ok(None),
+    parse_identifiers!(
+        parser,
 
-        "xMinYMin" => Ok(Some((X(Min), Y(Min)))),
-        "xMidYMin" => Ok(Some((X(Mid), Y(Min)))),
-        "xMaxYMin" => Ok(Some((X(Max), Y(Min)))),
+        "none" => None,
 
-        "xMinYMid" => Ok(Some((X(Min), Y(Mid)))),
-        "xMidYMid" => Ok(Some((X(Mid), Y(Mid)))),
-        "xMaxYMid" => Ok(Some((X(Max), Y(Mid)))),
+        "xMinYMin" => Some((X(Min), Y(Min))),
+        "xMidYMin" => Some((X(Mid), Y(Min))),
+        "xMaxYMin" => Some((X(Max), Y(Min))),
 
-        "xMinYMax" => Ok(Some((X(Min), Y(Max)))),
-        "xMidYMax" => Ok(Some((X(Mid), Y(Max)))),
-        "xMaxYMax" => Ok(Some((X(Max), Y(Max)))),
+        "xMinYMid" => Some((X(Min), Y(Mid))),
+        "xMidYMid" => Some((X(Mid), Y(Mid))),
+        "xMaxYMid" => Some((X(Max), Y(Mid))),
 
-        _ => Err(ValueErrorKind::Parse(ParseError::new("invalid alignment"))),
-    }
+        "xMinYMax" => Some((X(Min), Y(Max))),
+        "xMidYMax" => Some((X(Mid), Y(Max))),
+        "xMaxYMax" => Some((X(Max), Y(Max))),
+    )
 }
 
-fn parse_fit_mode(s: &str) -> Result<FitMode, ValueErrorKind> {
-    match s {
-        "meet" => Ok(FitMode::Meet),
-        "slice" => Ok(FitMode::Slice),
-        _ => Err(ValueErrorKind::Parse(ParseError::new("invalid fit mode"))),
-    }
+fn parse_fit_mode<'i>(parser: &mut Parser<'i, '_>) -> Result<FitMode, BasicParseError<'i>> {
+    parse_identifiers!(
+        parser,
+        "meet" => FitMode::Meet,
+        "slice" => FitMode::Slice,
+    )
 }
 
 impl Parse for AspectRatio {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<AspectRatio, ValueErrorKind> {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<AspectRatio, ParseError<'i>> {
         let defer = parser
             .try_parse(|p| p.expect_ident_matching("defer"))
             .is_ok();
 
-        let align_xy = parser.try_parse(|p| {
-            p.expect_ident()
-                .map_err(|_| ValueErrorKind::Parse(ParseError::new("expected identifier")))
-                .and_then(|ident| parse_align_xy(ident))
-        })?;
+        let align_xy = parser.try_parse(|p| parse_align_xy(p))?;
 
-        let fit = parser
-            .try_parse(|p| {
-                p.expect_ident()
-                    .map_err(|_| ValueErrorKind::Parse(ParseError::new("expected identifier")))
-                    .and_then(|ident| parse_fit_mode(ident))
-            })
-            .unwrap_or_default();
+        let fit = parser.try_parse(|p| parse_fit_mode(p)).unwrap_or_default();
 
-        parser
-            .expect_exhausted()
-            .map_err(|_| ValueErrorKind::Parse(ParseError::new("extra data in AspectRatio")))?;
+        parser.expect_exhausted()?;
 
         let align = align_xy.map(|(x, y)| Align { x, y, fit });
 
@@ -254,17 +241,13 @@ impl Parse for AspectRatio {
 mod tests {
     use super::*;
     use crate::float_eq_cairo::ApproxEqCairo;
-    use crate::rect::RectangleExt;
-    use cairo::Rectangle;
 
     #[test]
     fn parsing_invalid_strings_yields_error() {
         assert!(AspectRatio::parse_str("").is_err());
         assert!(AspectRatio::parse_str("defer").is_err());
         assert!(AspectRatio::parse_str("defer foo").is_err());
-        assert!(AspectRatio::parse_str("defer xmidymid").is_err());
         assert!(AspectRatio::parse_str("defer xMidYMid foo").is_err());
-        assert!(AspectRatio::parse_str("xmidymid").is_err());
         assert!(AspectRatio::parse_str("xMidYMid foo").is_err());
         assert!(AspectRatio::parse_str("defer xMidYMid meet foo").is_err());
     }
@@ -340,139 +323,87 @@ mod tests {
         );
     }
 
-    fn assert_quadruples_equal(a: &(f64, f64, f64, f64), b: &(f64, f64, f64, f64)) {
-        assert_approx_eq_cairo!(a.0, b.0);
-        assert_approx_eq_cairo!(a.1, b.1);
-        assert_approx_eq_cairo!(a.2, b.2);
-        assert_approx_eq_cairo!(a.3, b.3);
+    fn assert_rect_equal(r1: &Rect, r2: &Rect) {
+        assert_approx_eq_cairo!(r1.x0, r2.x0);
+        assert_approx_eq_cairo!(r1.y0, r2.y0);
+        assert_approx_eq_cairo!(r1.x1, r2.x1);
+        assert_approx_eq_cairo!(r1.y1, r2.y1);
     }
 
     #[test]
     fn aligns() {
+        let viewbox = ViewBox(Rect::from_size(1.0, 10.0));
+
         let foo = AspectRatio::parse_str("xMinYMin meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::from_size(0.1, 1.0));
 
         let foo = AspectRatio::parse_str("xMinYMin slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, 0.0, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::from_size(10.0, 100.0));
 
         let foo = AspectRatio::parse_str("xMinYMid meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::from_size(0.1, 1.0));
 
         let foo = AspectRatio::parse_str("xMinYMid slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, -49.5, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(0.0, -49.5, 10.0, 100.0 - 49.5));
 
         let foo = AspectRatio::parse_str("xMinYMax meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::from_size(0.1, 1.0));
 
         let foo = AspectRatio::parse_str("xMinYMax slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, -99.0, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(0.0, -99.0, 10.0, 1.0));
 
         let foo = AspectRatio::parse_str("xMidYMin meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(4.95, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(4.95, 0.0, 4.95 + 0.1, 1.0));
 
         let foo = AspectRatio::parse_str("xMidYMin slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, 0.0, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::from_size(10.0, 100.0));
 
         let foo = AspectRatio::parse_str("xMidYMid meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(4.95, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(4.95, 0.0, 4.95 + 0.1, 1.0));
 
         let foo = AspectRatio::parse_str("xMidYMid slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, -49.5, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(0.0, -49.5, 10.0, 100.0 - 49.5));
 
         let foo = AspectRatio::parse_str("xMidYMax meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(4.95, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(4.95, 0.0, 4.95 + 0.1, 1.0));
 
         let foo = AspectRatio::parse_str("xMidYMax slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, -99.0, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(0.0, -99.0, 10.0, 1.0));
 
         let foo = AspectRatio::parse_str("xMaxYMin meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(9.9, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(9.9, 0.0, 10.0, 1.0));
 
         let foo = AspectRatio::parse_str("xMaxYMin slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, 0.0, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::from_size(10.0, 100.0));
 
         let foo = AspectRatio::parse_str("xMaxYMid meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(9.9, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(9.9, 0.0, 10.0, 1.0));
 
         let foo = AspectRatio::parse_str("xMaxYMid slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, -49.5, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(0.0, -49.5, 10.0, 100.0 - 49.5));
 
         let foo = AspectRatio::parse_str("xMaxYMax meet").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(9.9, 0.0, 0.1, 1.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(9.9, 0.0, 10.0, 1.0));
 
         let foo = AspectRatio::parse_str("xMaxYMax slice").unwrap();
-        let foo = foo.compute(
-            &ViewBox::new(0.0, 0.0, 1.0, 10.0),
-            &Rectangle::new(0.0, 0.0, 10.0, 1.0),
-        );
-        assert_quadruples_equal(&foo, &(0.0, -99.0, 10.0, 100.0));
+        let foo = foo.compute(&viewbox, &Rect::from_size(10.0, 1.0));
+        assert_rect_equal(&foo, &Rect::new(0.0, -99.0, 10.0, 1.0));
     }
 }
