@@ -13,7 +13,6 @@
 //! }
 //!
 //! #baz { stroke-width: 42; }
-//!
 //! ```
 //! The example contains three **rules**, the first one is an **at-rule*,
 //! the other two are **qualified rules**.
@@ -74,10 +73,14 @@
 //! the rules in the stylesheets and gather the matches; then sort the
 //! matches by specificity and apply the result to each element.
 
-use cssparser::*;
+use cssparser::{
+    self, match_ignore_ascii_case, parse_important, AtRuleParser, AtRuleType, BasicParseErrorKind,
+    CowRcStr, DeclarationListParser, DeclarationParser, Parser, ParserInput, QualifiedRuleParser,
+    RuleListParser, SourceLocation, ToCss, _cssparser_internal_to_lowercase,
+};
 use selectors::attr::{AttrSelectorOperation, CaseSensitivity, NamespaceConstraint};
 use selectors::matching::{ElementSelectorFlags, MatchingContext, MatchingMode, QuirksMode};
-use selectors::{self, OpaqueElement, SelectorImpl, SelectorList};
+use selectors::{OpaqueElement, SelectorImpl, SelectorList};
 
 use std::cmp::Ordering;
 use std::fmt;
@@ -120,10 +123,9 @@ impl<'i> DeclarationParser<'i> for DeclParser {
         &mut self,
         name: CowRcStr<'i>,
         input: &mut Parser<'i, 't>,
-    ) -> Result<Declaration, cssparser::ParseError<'i, ValueErrorKind>> {
-        let prop_name = QualName::new(None, ns!(svg), LocalName::from(name.as_ref()));
-        let property =
-            parse_property(&prop_name, input, true).map_err(|e| input.new_custom_error(e))?;
+    ) -> Result<Declaration, ParseError<'i>> {
+        let prop_name = QualName::new(None, ns!(), LocalName::from(name.as_ref()));
+        let property = parse_property(&prop_name, input, true)?;
 
         let important = input.try_parse(parse_important).is_ok();
 
@@ -151,13 +153,14 @@ impl<'i> AtRuleParser<'i> for DeclParser {
 pub struct RuleParser;
 
 /// Errors from the CSS parsing process
-pub enum CssParseErrorKind<'i> {
+#[derive(Debug)]
+pub enum ParseErrorKind<'i> {
     Selector(selectors::parser::SelectorParseErrorKind<'i>),
 }
 
-impl<'i> From<selectors::parser::SelectorParseErrorKind<'i>> for CssParseErrorKind<'i> {
-    fn from(e: selectors::parser::SelectorParseErrorKind) -> CssParseErrorKind {
-        CssParseErrorKind::Selector(e)
+impl<'i> From<selectors::parser::SelectorParseErrorKind<'i>> for ParseErrorKind<'i> {
+    fn from(e: selectors::parser::SelectorParseErrorKind) -> ParseErrorKind {
+        ParseErrorKind::Selector(e)
     }
 }
 
@@ -186,7 +189,7 @@ pub enum Rule {
 // Required to implement the `Prelude` associated type in `cssparser::QualifiedRuleParser`
 impl<'i> selectors::Parser<'i> for RuleParser {
     type Impl = Selector;
-    type Error = CssParseErrorKind<'i>;
+    type Error = ParseErrorKind<'i>;
 
     fn default_namespace(&self) -> Option<<Self::Impl as SelectorImpl>::NamespaceUrl> {
         Some(ns!(svg))
@@ -226,7 +229,7 @@ impl<'i> selectors::Parser<'i> for RuleParser {
 impl<'i> QualifiedRuleParser<'i> for RuleParser {
     type Prelude = SelectorList<Selector>;
     type QualifiedRule = Rule;
-    type Error = CssParseErrorKind<'i>;
+    type Error = ParseErrorKind<'i>;
 
     fn parse_prelude<'t>(
         &mut self,
@@ -242,7 +245,13 @@ impl<'i> QualifiedRuleParser<'i> for RuleParser {
         input: &mut Parser<'i, 't>,
     ) -> Result<Self::QualifiedRule, cssparser::ParseError<'i, Self::Error>> {
         let declarations = DeclarationListParser::new(input, DeclParser)
-            .filter_map(Result::ok) // ignore invalid property name or value
+            .filter_map(|r| match r {
+                Ok(decl) => Some(decl),
+                Err(e) => {
+                    rsvg_log!("Invalid declaration; ignoring: {:?}", e);
+                    None
+                }
+            })
             .collect();
 
         Ok(Rule::QualifiedRule(QualifiedRule {
@@ -259,14 +268,16 @@ impl<'i> AtRuleParser<'i> for RuleParser {
     type PreludeBlock = ();
     type PreludeNoBlock = AtRulePrelude;
     type AtRule = Rule;
-    type Error = CssParseErrorKind<'i>;
+    type Error = ParseErrorKind<'i>;
 
     fn parse_prelude<'t>(
         &mut self,
         name: CowRcStr<'i>,
         input: &mut Parser<'i, 't>,
-    ) -> Result<AtRuleType<Self::PreludeNoBlock, Self::PreludeBlock>, ParseError<'i, Self::Error>>
-    {
+    ) -> Result<
+        AtRuleType<Self::PreludeNoBlock, Self::PreludeBlock>,
+        cssparser::ParseError<'i, Self::Error>,
+    > {
         match_ignore_ascii_case! { &name,
             "import" => {
                 // FIXME: at the moment we ignore media queries
@@ -311,6 +322,10 @@ impl selectors::parser::NonTSPseudoClass for NonTSPseudoClass {
     fn is_user_action_state(&self) -> bool {
         false
     }
+
+    fn has_zero_specificity(&self) -> bool {
+        false
+    }
 }
 
 /// Dummy type required by the SelectorImpl trait
@@ -355,7 +370,7 @@ impl SelectorImpl for Selector {
 /// RsvgNode is an alias for rctree::Node, so we can't implement
 /// `selectors::Element` directly on it.  We implement it on the
 /// `RsvgElement` wrapper instead.
-#[derive(Clone)]
+#[derive(Clone, PartialEq)]
 pub struct RsvgElement(RsvgNode);
 
 impl From<RsvgNode> for RsvgElement {
@@ -410,7 +425,7 @@ impl selectors::Element for RsvgElement {
                 return sibling.map(|n| n.into());
             }
 
-            sibling = self.0.previous_sibling();
+            sibling = sib.previous_sibling();
         }
 
         None
@@ -425,7 +440,7 @@ impl selectors::Element for RsvgElement {
                 return sibling.map(|n| n.into());
             }
 
-            sibling = self.0.next_sibling();
+            sibling = sib.next_sibling();
         }
 
         None
@@ -512,6 +527,16 @@ impl selectors::Element for RsvgElement {
             .unwrap_or(false)
     }
 
+    fn exported_part(&self, _name: &LocalName) -> Option<LocalName> {
+        // unsupported
+        None
+    }
+
+    fn imported_part(&self, _name: &LocalName) -> Option<LocalName> {
+        // unsupported
+        None
+    }
+
     fn is_part(&self, _name: &LocalName) -> bool {
         // unsupported
         false
@@ -588,8 +613,7 @@ impl<'a> PartialOrd for Match<'a> {
 
 impl<'a> PartialEq for Match<'a> {
     fn eq(&self, other: &Self) -> bool {
-        self.origin == other.origin
-            && self.specificity == other.specificity
+        self.origin == other.origin && self.specificity == other.specificity
     }
 }
 
@@ -632,7 +656,13 @@ impl Stylesheet {
         let mut parser = Parser::new(&mut input);
 
         RuleListParser::new_for_stylesheet(&mut parser, RuleParser)
-            .filter_map(Result::ok) // ignore invalid rules
+            .filter_map(|r| match r {
+                Ok(rule) => Some(rule),
+                Err(e) => {
+                    rsvg_log!("Invalid rule; ignoring: {:?}", e);
+                    None
+                }
+            })
             .for_each(|rule| match rule {
                 Rule::AtRule(AtRule::Import(url)) => {
                     // ignore invalid imports
@@ -735,3 +765,112 @@ pub fn cascade(root: &mut RsvgNode, stylesheets: &[Stylesheet]) {
     let values = ComputedValues::default();
     root.cascade(&values);
 }
+
+#[cfg(test)]
+mod tests {
+    use super::*;
+    use gio;
+    use glib::{self, prelude::*};
+    use selectors::Element;
+
+    use crate::allowed_url::Fragment;
+    use crate::document::Document;
+    use crate::handle::LoadOptions;
+
+    fn load_document(input: &'static [u8]) -> Document {
+        let bytes = glib::Bytes::from_static(input);
+        let stream = gio::MemoryInputStream::new_from_bytes(&bytes);
+
+        Document::load_from_stream(
+            &LoadOptions::new(None),
+            &stream.upcast(),
+            None::<&gio::Cancellable>,
+        )
+        .unwrap()
+    }
+
+    #[test]
+    fn impl_element() {
+        let document = load_document(
+            br#"<?xml version="1.0" encoding="UTF-8"?>
+<svg xmlns="http://www.w3.org/2000/svg" id="a">
+  <rect id="b" x="10" y="10" width="30" height="30"/>
+  <circle id="c" cx="10" cy="10" r="10"/>
+  <rect id="d" class="foo bar"/>
+</svg>
+"#,
+        );
+
+        let a = document
+            .lookup(&Fragment::new(None, "a".to_string()))
+            .unwrap();
+        let b = document
+            .lookup(&Fragment::new(None, "b".to_string()))
+            .unwrap();
+        let c = document
+            .lookup(&Fragment::new(None, "c".to_string()))
+            .unwrap();
+        let d = document
+            .lookup(&Fragment::new(None, "d".to_string()))
+            .unwrap();
+
+        // Node types
+
+        assert!(a.borrow().get_type() == NodeType::Svg);
+        assert!(b.borrow().get_type() == NodeType::Rect);
+        assert!(c.borrow().get_type() == NodeType::Circle);
+        assert!(d.borrow().get_type() == NodeType::Rect);
+
+        let a = RsvgElement(a);
+        let b = RsvgElement(b);
+        let c = RsvgElement(c);
+        let d = RsvgElement(d);
+
+        // Tree navigation
+
+        assert_eq!(a.parent_element(), None);
+        assert_eq!(b.parent_element(), Some(a.clone()));
+        assert_eq!(c.parent_element(), Some(a.clone()));
+        assert_eq!(d.parent_element(), Some(a.clone()));
+
+        assert_eq!(b.next_sibling_element(), Some(c.clone()));
+        assert_eq!(c.next_sibling_element(), Some(d.clone()));
+        assert_eq!(d.next_sibling_element(), None);
+
+        assert_eq!(b.prev_sibling_element(), None);
+        assert_eq!(c.prev_sibling_element(), Some(b.clone()));
+        assert_eq!(d.prev_sibling_element(), Some(c.clone()));
+
+        // Other operations
+
+        assert!(a.has_local_name(&LocalName::from("svg")));
+
+        assert!(a.has_namespace(&ns!(svg)));
+
+        assert!(!a.is_same_type(&b));
+        assert!(b.is_same_type(&d));
+
+        assert!(a.has_id(&LocalName::from("a"), CaseSensitivity::AsciiCaseInsensitive));
+        assert!(!b.has_id(
+            &LocalName::from("foo"),
+            CaseSensitivity::AsciiCaseInsensitive
+        ));
+
+        assert!(d.has_class(
+            &LocalName::from("foo"),
+            CaseSensitivity::AsciiCaseInsensitive
+        ));
+        assert!(d.has_class(
+            &LocalName::from("bar"),
+            CaseSensitivity::AsciiCaseInsensitive
+        ));
+
+        assert!(!a.has_class(
+            &LocalName::from("foo"),
+            CaseSensitivity::AsciiCaseInsensitive
+        ));
+
+        assert!(d.is_empty());
+        assert!(!a.is_empty());
+    }
+}
