@@ -1,4 +1,8 @@
-use cairo;
+//! The main context structure which drives the drawing process.
+
+use once_cell::sync::Lazy;
+use regex::{Captures, Regex};
+use std::borrow::Cow;
 use std::cell::RefCell;
 use std::convert::TryFrom;
 use std::rc::{Rc, Weak};
@@ -6,25 +10,31 @@ use std::rc::{Rc, Weak};
 use crate::allowed_url::Fragment;
 use crate::aspect_ratio::AspectRatio;
 use crate::bbox::BoundingBox;
-use crate::clip_path::{ClipPath, ClipPathUnits};
 use crate::coord_units::CoordUnits;
+use crate::dasharray::Dasharray;
 use crate::document::Document;
 use crate::dpi::Dpi;
 use crate::error::{AcquireError, RenderingError};
 use crate::filters;
 use crate::gradient::{LinearGradient, RadialGradient};
-use crate::length::Dasharray;
 use crate::limits;
-use crate::mask::Mask;
+use crate::marker;
 use crate::node::{CascadedValues, NodeDraw, NodeType, RsvgNode};
 use crate::paint_server::{PaintServer, PaintSource};
+use crate::path_builder::*;
 use crate::pattern::Pattern;
 use crate::properties::ComputedValues;
 use crate::property_defs::{
-    ClipRule, FillRule, ShapeRendering, StrokeDasharray, StrokeLinecap, StrokeLinejoin,
+    ClipRule, FillRule, Opacity, Overflow, ShapeRendering, StrokeDasharray, StrokeLinecap,
+    StrokeLinejoin,
+};
+use crate::rect::{Rect, TransformRect};
+use crate::shapes::Markers;
+use crate::structure::{ClipPath, Mask, Symbol, Use};
+use crate::surface_utils::{
+    shared_surface::ExclusiveImageSurface, shared_surface::SharedImageSurface,
+    shared_surface::SurfaceType,
 };
-use crate::rect::RectangleExt;
-use crate::surface_utils::shared_surface::SharedImageSurface;
 use crate::unit_interval::UnitInterval;
 use crate::viewbox::ViewBox;
 
@@ -79,13 +89,9 @@ pub struct DrawingCtx {
 
     initial_affine: cairo::Matrix,
 
-    rect: cairo::Rectangle,
+    rect: Rect,
     dpi: Dpi,
 
-    // This is a mitigation for SVG files that try to instance a huge number of
-    // elements via <use>, recursive patterns, etc.  See limits.rs for details.
-    num_elements_acquired: usize,
-
     cr_stack: Vec<cairo::Context>,
     cr: cairo::Context,
 
@@ -104,7 +110,7 @@ impl DrawingCtx {
         document: Rc<Document>,
         node: Option<&RsvgNode>,
         cr: &cairo::Context,
-        viewport: &cairo::Rectangle,
+        viewport: Rect,
         dpi: Dpi,
         measuring: bool,
         testing: bool,
@@ -115,13 +121,9 @@ impl DrawingCtx {
         // while the code gets refactored not to need special cases for that.
 
         let (rect, vbox) = if measuring {
-            (
-                cairo::Rectangle::new(0.0, 0.0, 1.0, 1.0),
-                ViewBox::new(0.0, 0.0, 1.0, 1.0),
-            )
+            let unit_rect = Rect::from_size(1.0, 1.0);
+            (unit_rect, ViewBox(unit_rect))
         } else {
-            let rect = *viewport;
-
             // https://www.w3.org/TR/SVG2/coords.html#InitialCoordinateSystem
             //
             // "For the outermost svg element, the SVG user agent must
@@ -134,14 +136,9 @@ impl DrawingCtx {
             // "... the initial viewport coordinate system (and therefore
             // the initial user coordinate system) must have its origin at
             // the top/left of the viewport"
-            let vbox = ViewBox {
-                x: 0.0,
-                y: 0.0,
-                width: viewport.width,
-                height: viewport.height,
-            };
+            let vbox = ViewBox(Rect::from_size(viewport.width(), viewport.height()));
 
-            (rect, vbox)
+            (viewport, vbox)
         };
 
         let mut view_box_stack = Vec::new();
@@ -154,7 +151,6 @@ impl DrawingCtx {
             initial_affine,
             rect,
             dpi,
-            num_elements_acquired: 0,
             cr_stack: Vec::new(),
             cr: cr.clone(),
             view_box_stack: Rc::new(RefCell::new(view_box_stack)),
@@ -165,13 +161,15 @@ impl DrawingCtx {
         };
 
         if let Some(node) = node {
-            draw_ctx.add_node_and_ancestors_to_stack(node);
+            for n in node.ancestors() {
+                draw_ctx.drawsub_stack.push(n.clone());
+            }
         }
 
         draw_ctx
     }
 
-    pub fn toplevel_viewport(&self) -> cairo::Rectangle {
+    pub fn toplevel_viewport(&self) -> Rect {
         self.rect
     }
 
@@ -188,7 +186,7 @@ impl DrawingCtx {
     }
 
     pub fn empty_bbox(&self) -> BoundingBox {
-        BoundingBox::new(&self.cr.get_matrix())
+        BoundingBox::new().with_affine(self.cr.get_matrix())
     }
 
     // FIXME: Usage of this function is more less a hack... The caller
@@ -201,18 +199,18 @@ impl DrawingCtx {
     }
 
     // Temporary hack while we unify surface/cr/affine creation
-    pub fn push_cairo_context(&mut self, cr: cairo::Context) {
+    fn push_cairo_context(&mut self, cr: cairo::Context) {
         self.cr_stack.push(self.cr.clone());
         self.cr = cr;
     }
 
     // Temporary hack while we unify surface/cr/affine creation
-    pub fn pop_cairo_context(&mut self) {
+    fn pop_cairo_context(&mut self) {
         self.cr = self.cr_stack.pop().unwrap();
     }
 
     fn size_for_temporary_surface(&self) -> (i32, i32) {
-        let (viewport_width, viewport_height) = (self.rect.width, self.rect.height);
+        let (viewport_width, viewport_height) = (self.rect.width(), self.rect.height());
 
         let (scaled_width, scaled_height) = self
             .initial_affine_with_offset()
@@ -257,13 +255,13 @@ impl DrawingCtx {
     pub fn get_view_params(&self) -> ViewParams {
         let view_box_stack = self.view_box_stack.borrow();
         let last = view_box_stack.len() - 1;
-        let stack_top = &view_box_stack[last];
+        let top_rect = &view_box_stack[last].0;
 
         ViewParams {
             dpi_x: self.dpi.x(),
             dpi_y: self.dpi.y(),
-            view_box_width: stack_top.width,
-            view_box_height: stack_top.height,
+            view_box_width: top_rect.width(),
+            view_box_height: top_rect.height(),
             view_box_stack: None,
         }
     }
@@ -278,7 +276,7 @@ impl DrawingCtx {
     pub fn push_view_box(&self, width: f64, height: f64) -> ViewParams {
         self.view_box_stack
             .borrow_mut()
-            .push(ViewBox::new(0.0, 0.0, width, height));
+            .push(ViewBox(Rect::from_size(width, height)));
 
         ViewParams {
             dpi_x: self.dpi.x(),
@@ -292,104 +290,167 @@ impl DrawingCtx {
     pub fn push_new_viewport(
         &self,
         vbox: Option<ViewBox>,
-        viewport: &cairo::Rectangle,
+        viewport: Rect,
         preserve_aspect_ratio: AspectRatio,
         clip_mode: Option<ClipMode>,
     ) -> Option<ViewParams> {
-        if let Some(ref clip) = clip_mode {
-            if *clip == ClipMode::ClipToViewport {
-                self.clip(viewport.x, viewport.y, viewport.width, viewport.height);
-            }
+        if let Some(ClipMode::ClipToViewport) = clip_mode {
+            self.clip(viewport);
         }
 
         preserve_aspect_ratio
-            .viewport_to_viewbox_transform(vbox, viewport)
+            .viewport_to_viewbox_transform(vbox, &viewport)
             .and_then(|matrix| {
                 self.cr.transform(matrix);
 
                 if let Some(vbox) = vbox {
-                    if let Some(ref clip) = clip_mode {
-                        if *clip == ClipMode::ClipToVbox {
-                            self.clip(vbox.x, vbox.y, vbox.width, vbox.height);
-                        }
+                    if let Some(ClipMode::ClipToVbox) = clip_mode {
+                        self.clip(vbox.0);
                     }
 
-                    Some(self.push_view_box(vbox.width, vbox.height))
+                    Some(self.push_view_box(vbox.0.width(), vbox.0.height()))
                 } else {
                     Some(self.get_view_params())
                 }
             })
     }
 
-    // Use this function when looking up urls to other nodes, and when you expect
-    // the node to be of a particular type. This function does proper recursion
-    // checking and thereby avoids infinite loops.
-    //
-    // Nodes acquired by this function must be released in reverse
-    // acquiring order.
-    //
-    // Specify an empty slice for `node_types` if you want a node of any type.
-    //
-    // Malformed SVGs, for example, may reference a marker by its IRI, but
-    // the object referenced by the IRI is not a marker.
-    //
-    // Note that if you acquire a node, you have to release it before trying to
-    // acquire it again.  If you acquire a node "#foo" and don't release it before
-    // trying to acquire "foo" again, you will obtain a None the second time.
     pub fn acquire_node(
         &mut self,
         fragment: &Fragment,
         node_types: &[NodeType],
     ) -> Result<AcquiredNode, AcquireError> {
-        self.num_elements_acquired += 1;
-
-        if self.num_elements_acquired > limits::MAX_REFERENCED_ELEMENTS {
-            return Err(AcquireError::MaxReferencesExceeded);
-        }
-
         self.acquired_nodes.acquire(fragment, node_types)
     }
 
-    pub fn acquire_node_ref(&mut self, node: &RsvgNode) -> Result<AcquiredNode, AcquireError> {
-        self.acquired_nodes.push_node_ref(node)
-    }
-
-    // Returns (clip_in_user_space, clip_in_object_space), both Option<RsvgNode>
-    fn get_clip_in_user_and_object_space(
-        &mut self,
-        clip_uri: Option<&Fragment>,
-    ) -> (Option<RsvgNode>, Option<RsvgNode>) {
-        clip_uri
-            .and_then(|fragment| self.acquire_node(fragment, &[NodeType::ClipPath]).ok())
-            .and_then(|acquired| {
-                let clip_node = acquired.get().clone();
-
-                let ClipPathUnits(units) = clip_node.borrow().get_impl::<ClipPath>().get_units();
-
-                if units == CoordUnits::UserSpaceOnUse {
-                    Some((Some(clip_node), None))
-                } else {
-                    assert!(units == CoordUnits::ObjectBoundingBox);
-                    Some((None, Some(clip_node)))
-                }
-            })
-            .unwrap_or((None, None))
-    }
-
     fn clip_to_node(
         &mut self,
         clip_node: &Option<RsvgNode>,
         bbox: &BoundingBox,
     ) -> Result<(), RenderingError> {
         if let Some(node) = clip_node {
-            let node_data = node.borrow();
-            let clip_path = node_data.get_impl::<ClipPath>();
-            clip_path.to_cairo_context(&node, self, &bbox)
+            let units = node.borrow().get_impl::<ClipPath>().get_units();
+
+            if units == CoordUnits::ObjectBoundingBox && bbox.rect.is_none() {
+                // The node being clipped is empty / doesn't have a
+                // bounding box, so there's nothing to clip!
+                return Ok(());
+            }
+
+            let cascaded = CascadedValues::new_from_node(node);
+
+            let matrix = if units == CoordUnits::ObjectBoundingBox {
+                let bbox_rect = bbox.rect.as_ref().unwrap();
+
+                Some(cairo::Matrix::new(
+                    bbox_rect.width(),
+                    0.0,
+                    0.0,
+                    bbox_rect.height(),
+                    bbox_rect.x0,
+                    bbox_rect.y0,
+                ))
+            } else {
+                None
+            };
+
+            self.with_saved_matrix(matrix, &mut |dc| {
+                let cr = dc.get_cairo_context();
+
+                // here we don't push a layer because we are clipping
+                let res = node.draw_children(&cascaded, dc, true);
+
+                cr.clip();
+
+                res
+            })
+            .and_then(|_bbox|
+                // Clipping paths do not contribute to bounding boxes (they should,
+                // but we need Real Computational Geometry(tm), so ignore the
+                // bbox from the clip path.
+                Ok(()))
         } else {
             Ok(())
         }
     }
 
+    fn generate_cairo_mask(
+        &mut self,
+        mask: &Mask,
+        mask_node: &RsvgNode,
+        affine: cairo::Matrix,
+        bbox: &BoundingBox,
+    ) -> Result<Option<cairo::ImageSurface>, RenderingError> {
+        if bbox.rect.is_none() {
+            // The node being masked is empty / doesn't have a
+            // bounding box, so there's nothing to mask!
+            return Ok(None);
+        }
+
+        let bbox_rect = bbox.rect.as_ref().unwrap();
+        let (bb_x, bb_y) = (bbox_rect.x0, bbox_rect.y0);
+        let (bb_w, bb_h) = bbox_rect.size();
+
+        let cascaded = CascadedValues::new_from_node(mask_node);
+        let values = cascaded.get();
+
+        let mask_units = mask.get_units();
+
+        let mask_rect = {
+            let params = if mask_units == CoordUnits::ObjectBoundingBox {
+                self.push_view_box(1.0, 1.0)
+            } else {
+                self.get_view_params()
+            };
+
+            mask.get_rect(&values, &params)
+        };
+
+        let mask_affine = cairo::Matrix::multiply(&mask_node.borrow().get_transform(), &affine);
+
+        let mask_content_surface = self.create_surface_for_toplevel_viewport()?;
+
+        // Use a scope because mask_cr needs to release the
+        // reference to the surface before we access the pixels
+        {
+            let mask_cr = cairo::Context::new(&mask_content_surface);
+            mask_cr.set_matrix(mask_affine);
+
+            let bbtransform = cairo::Matrix::new(bb_w, 0.0, 0.0, bb_h, bb_x, bb_y);
+
+            self.push_cairo_context(mask_cr);
+
+            if mask_units == CoordUnits::ObjectBoundingBox {
+                self.clip(bbtransform.transform_rect(&mask_rect));
+            } else {
+                self.clip(mask_rect);
+            }
+
+            let _params = if mask.get_content_units() == CoordUnits::ObjectBoundingBox {
+                self.get_cairo_context().transform(bbtransform);
+                self.push_view_box(1.0, 1.0)
+            } else {
+                self.get_view_params()
+            };
+
+            let res = self.with_discrete_layer(mask_node, values, false, &mut |dc| {
+                mask_node.draw_children(&cascaded, dc, false)
+            });
+
+            self.pop_cairo_context();
+
+            res?;
+        }
+
+        let Opacity(opacity) = values.opacity;
+
+        let mask = SharedImageSurface::wrap(mask_content_surface, SurfaceType::SRgb)?
+            .to_mask(opacity)?
+            .into_image_surface()?;
+
+        Ok(Some(mask))
+    }
+
     pub fn with_discrete_layer(
         &mut self,
         node: &RsvgNode,
@@ -416,7 +477,7 @@ impl DrawingCtx {
                 let affine_at_start = dc.cr.get_matrix();
 
                 let (clip_in_user_space, clip_in_object_space) =
-                    dc.get_clip_in_user_and_object_space(clip_uri);
+                    get_clip_in_user_and_object_space(&mut dc.acquired_nodes, clip_uri);
 
                 // Here we are clipping in user space, so the bbox doesn't matter
                 dc.clip_to_node(&clip_in_user_space, &dc.empty_bbox())?;
@@ -456,17 +517,24 @@ impl DrawingCtx {
                     let bbox = if let Ok(ref bbox) = res {
                         *bbox
                     } else {
-                        BoundingBox::new(&affines.for_temporary_surface)
+                        BoundingBox::new().with_affine(affines.for_temporary_surface)
                     };
 
                     // Filter
 
                     let source_surface = if let Some(filter_uri) = filter {
-                        let child_surface =
-                            cairo::ImageSurface::try_from(dc.cr.get_target()).unwrap();
-                        let img_surface =
-                            dc.run_filter(filter_uri, node, values, &child_surface, bbox)?;
-                        // turn into a Surface
+                        // The target surface has multiple references.
+                        // We need to copy it to a new surface to have a unique
+                        // reference to be able to safely access the pixel data.
+                        let child_surface = SharedImageSurface::copy_from_surface(
+                            &cairo::ImageSurface::try_from(dc.cr.get_target()).unwrap(),
+                        )?;
+
+                        let img_surface = dc
+                            .run_filter(filter_uri, node, values, child_surface, bbox)?
+                            .into_image_surface()?;
+
+                        // turn ImageSurface into a Surface
                         (*img_surface).clone()
                     } else {
                         dc.cr.get_target()
@@ -491,18 +559,20 @@ impl DrawingCtx {
                             let mask_node = acquired.get();
 
                             res = res.and_then(|bbox| {
-                                mask_node
-                                    .borrow()
-                                    .get_impl::<Mask>()
-                                    .generate_cairo_mask(&mask_node, &affines, dc, &bbox)
-                                    .and_then(|mask_surf| {
-                                        if let Some(surf) = mask_surf {
-                                            dc.cr.set_matrix(affines.compositing);
-                                            dc.cr.mask_surface(&surf, 0.0, 0.0);
-                                        }
-                                        Ok(())
-                                    })
-                                    .map(|_: ()| bbox)
+                                dc.generate_cairo_mask(
+                                    &mask_node.borrow().get_impl::<Mask>(),
+                                    &mask_node,
+                                    affines.for_temporary_surface,
+                                    &bbox,
+                                )
+                                .and_then(|mask_surf| {
+                                    if let Some(surf) = mask_surf {
+                                        dc.cr.set_matrix(affines.compositing);
+                                        dc.cr.mask_surface(&surf, 0.0, 0.0);
+                                    }
+                                    Ok(())
+                                })
+                                .map(|_: ()| bbox)
                             });
                         } else {
                             rsvg_log!(
@@ -535,11 +605,12 @@ impl DrawingCtx {
 
     fn initial_affine_with_offset(&self) -> cairo::Matrix {
         let mut initial_with_offset = self.initial_affine;
-        initial_with_offset.translate(self.rect.x, self.rect.y);
+        initial_with_offset.translate(self.rect.x0, self.rect.y0);
         initial_with_offset
     }
 
-    /// Saves the current Cairo matrix, runs the draw_fn, and restores the matrix
+    /// Saves the current Cairo matrix, applies a transform if specified,
+    /// runs the draw_fn, and restores the original matrix
     ///
     /// This is slightly cheaper than a `cr.save()` / `cr.restore()`
     /// pair, but more importantly, it does not reset the whole
@@ -547,14 +618,21 @@ impl DrawingCtx {
     /// was set by the `draw_fn`.
     pub fn with_saved_matrix(
         &mut self,
+        transform: Option<cairo::Matrix>,
         draw_fn: &mut dyn FnMut(&mut DrawingCtx) -> Result<BoundingBox, RenderingError>,
     ) -> Result<BoundingBox, RenderingError> {
         let matrix = self.cr.get_matrix();
+
+        if let Some(t) = transform {
+            self.cr.transform(t);
+        }
+
         let res = draw_fn(self);
+
         self.cr.set_matrix(matrix);
 
         if let Ok(bbox) = res {
-            let mut orig_matrix_bbox = BoundingBox::new(&matrix);
+            let mut orig_matrix_bbox = BoundingBox::new().with_affine(matrix);
             orig_matrix_bbox.insert(&bbox);
             Ok(orig_matrix_bbox)
         } else {
@@ -573,14 +651,34 @@ impl DrawingCtx {
         res
     }
 
+    /// Wraps the draw_fn in a link to the given target
+    pub fn with_link_tag(
+        &mut self,
+        link_target: &str,
+        draw_fn: &mut dyn FnMut(&mut DrawingCtx) -> Result<BoundingBox, RenderingError>,
+    ) -> Result<BoundingBox, RenderingError> {
+        const CAIRO_TAG_LINK: &str = "Link";
+
+        let attributes = format!("uri='{}'", escape_link_target(link_target));
+
+        let cr = self.get_cairo_context();
+        cr.tag_begin(CAIRO_TAG_LINK, &attributes);
+
+        let res = draw_fn(self);
+
+        cr.tag_end(CAIRO_TAG_LINK);
+
+        res
+    }
+
     fn run_filter(
         &mut self,
         filter_uri: &Fragment,
         node: &RsvgNode,
         values: &ComputedValues,
-        child_surface: &cairo::ImageSurface,
+        child_surface: SharedImageSurface,
         node_bbox: BoundingBox,
-    ) -> Result<cairo::ImageSurface, RenderingError> {
+    ) -> Result<SharedImageSurface, RenderingError> {
         match self.acquire_node(filter_uri, &[NodeType::Filter]) {
             Ok(acquired) => {
                 let filter_node = acquired.get();
@@ -602,10 +700,10 @@ impl DrawingCtx {
 
                 // Non-existing filters must act as null filters (that is, an
                 // empty surface is returned).
-                Ok(cairo::ImageSurface::create(
-                    cairo::Format::ARgb32,
-                    child_surface.get_width(),
-                    child_surface.get_height(),
+                Ok(SharedImageSurface::empty(
+                    child_surface.width(),
+                    child_surface.height(),
+                    child_surface.surface_type(),
                 )?)
             }
         }
@@ -631,17 +729,6 @@ impl DrawingCtx {
         );
     }
 
-    fn acquire_paint_server(&mut self, fragment: &Fragment) -> Result<AcquiredNode, AcquireError> {
-        self.acquire_node(
-            fragment,
-            &[
-                NodeType::LinearGradient,
-                NodeType::RadialGradient,
-                NodeType::Pattern,
-            ],
-        )
-    }
-
     pub fn set_source_paint_server(
         &mut self,
         ps: &PaintServer,
@@ -656,7 +743,7 @@ impl DrawingCtx {
             } => {
                 let mut had_paint_server = false;
 
-                match self.acquire_paint_server(iri) {
+                match acquire_paint_server(&mut self.acquired_nodes, iri) {
                     Ok(acquired) => {
                         let node = acquired.get();
 
@@ -783,14 +870,50 @@ impl DrawingCtx {
         res.and_then(|_: ()| Ok(bbox))
     }
 
-    pub fn clip(&self, x: f64, y: f64, w: f64, h: f64) {
-        let cr = self.get_cairo_context();
+    pub fn draw_path(
+        &mut self,
+        builder: &PathBuilder,
+        node: &RsvgNode,
+        values: &ComputedValues,
+        markers: Markers,
+        clipping: bool,
+    ) -> Result<BoundingBox, RenderingError> {
+        if !builder.is_empty() {
+            let bbox = self.with_discrete_layer(node, values, clipping, &mut |dc| {
+                let cr = dc.get_cairo_context();
 
-        cr.rectangle(x, y, w, h);
+                builder.to_cairo(&cr)?;
+
+                if clipping {
+                    cr.set_fill_rule(cairo::FillRule::from(values.clip_rule));
+                    Ok(dc.empty_bbox())
+                } else {
+                    cr.set_fill_rule(cairo::FillRule::from(values.fill_rule));
+                    dc.stroke_and_fill(&cr, values)
+                }
+            })?;
+
+            if markers == Markers::Yes {
+                marker::render_markers_for_path_builder(builder, self, values, clipping)?;
+            }
+
+            Ok(bbox)
+        } else {
+            Ok(self.empty_bbox())
+        }
+    }
+
+    pub fn clip(&self, rect: Rect) {
+        let cr = self.get_cairo_context();
+        cr.rectangle(rect.x0, rect.y0, rect.width(), rect.height());
         cr.clip();
     }
 
-    pub fn get_snapshot(&self, surface: &cairo::ImageSurface) {
+    pub fn get_snapshot(
+        &self,
+        width: i32,
+        height: i32,
+    ) -> Result<SharedImageSurface, cairo::Status> {
         // TODO: as far as I can tell this should not render elements past the last (topmost) one
         // with enable-background: new (because technically we shouldn't have been caching them).
         // Right now there are no enable-background checks whatsoever.
@@ -806,18 +929,25 @@ impl DrawingCtx {
         //
         // CSS Compositing and Blending, "isolation" property:
         //   https://www.w3.org/TR/compositing-1/#isolation
-        let cr = cairo::Context::new(&surface);
-        for (depth, draw) in self.cr_stack.iter().enumerate() {
-            let affines = CompositingAffines::new(
-                draw.get_matrix(),
-                self.initial_affine_with_offset(),
-                depth,
-            );
+        let mut surface = ExclusiveImageSurface::new(width, height, SurfaceType::SRgb)?;
 
-            cr.set_matrix(affines.for_snapshot);
-            cr.set_source_surface(&draw.get_target(), 0.0, 0.0);
-            cr.paint();
-        }
+        surface.draw(&mut |cr| {
+            for (depth, draw) in self.cr_stack.iter().enumerate() {
+                let affines = CompositingAffines::new(
+                    draw.get_matrix(),
+                    self.initial_affine_with_offset(),
+                    depth,
+                );
+
+                cr.set_matrix(affines.for_snapshot);
+                cr.set_source_surface(&draw.get_target(), 0.0, 0.0);
+                cr.paint();
+            }
+
+            Ok(())
+        })?;
+
+        surface.share()
     }
 
     pub fn lookup_image(&self, href: &str) -> Result<SharedImageSurface, RenderingError> {
@@ -826,33 +956,34 @@ impl DrawingCtx {
             .map_err(|_| RenderingError::InvalidHref)
     }
 
-    pub fn draw_node_on_surface(
+    pub fn draw_node_to_surface(
         &mut self,
         node: &RsvgNode,
         cascaded: &CascadedValues<'_>,
-        surface: &cairo::ImageSurface,
         affine: cairo::Matrix,
-        width: f64,
-        height: f64,
-    ) -> Result<BoundingBox, RenderingError> {
+        width: i32,
+        height: i32,
+    ) -> Result<SharedImageSurface, RenderingError> {
+        let surface = cairo::ImageSurface::create(cairo::Format::ARgb32, width, height)?;
+
         let save_cr = self.cr.clone();
         let save_rect = self.rect;
 
-        let cr = cairo::Context::new(surface);
-        cr.set_matrix(affine);
+        {
+            let cr = cairo::Context::new(&surface);
+            cr.set_matrix(affine);
 
-        self.cr = cr;
-        self.rect.x = 0.0;
-        self.rect.y = 0.0;
-        self.rect.width = width;
-        self.rect.height = height;
+            self.cr = cr;
 
-        let res = self.draw_node_from_stack(cascaded, node, false);
+            self.rect = Rect::from_size(f64::from(width), f64::from(height));
+
+            let _ = self.draw_node_from_stack(cascaded, node, false)?;
+        }
 
         self.cr = save_cr;
         self.rect = save_rect;
 
-        res
+        Ok(SharedImageSurface::wrap(surface, SurfaceType::SRgb)?)
     }
 
     pub fn draw_node_from_stack(
@@ -883,16 +1014,111 @@ impl DrawingCtx {
         res
     }
 
-    pub fn add_node_and_ancestors_to_stack(&mut self, node: &RsvgNode) {
-        self.drawsub_stack.push(node.clone());
-        if let Some(ref parent) = node.parent() {
-            self.add_node_and_ancestors_to_stack(parent);
+    pub fn draw_from_use_node(
+        &mut self,
+        node: &RsvgNode,
+        cascaded: &CascadedValues<'_>,
+        clipping: bool,
+    ) -> Result<BoundingBox, RenderingError> {
+        let node_data = node.borrow();
+        let use_ = node_data.get_impl::<Use>();
+
+        // <use> is an element that is used directly, unlike
+        // <pattern>, which is used through a fill="url(#...)"
+        // reference.  However, <use> will always reference another
+        // element, potentially itself or an ancestor of itself (or
+        // another <use> which references the first one, etc.).  So,
+        // we acquire the <use> element itself so that circular
+        // references can be caught.
+        let _self_acquired = self.acquired_nodes.acquire_ref(node).map_err(|e| {
+            if let AcquireError::CircularReference(_) = e {
+                rsvg_log!("circular reference in element {}", node);
+                RenderingError::CircularReference
+            } else {
+                unreachable!();
+            }
+        })?;
+
+        let link = use_.get_link();
+        if link.is_none() {
+            return Ok(self.empty_bbox());
+        }
+
+        let acquired = match self.acquire_node(link.unwrap(), &[]) {
+            Ok(acquired) => acquired,
+
+            Err(AcquireError::CircularReference(node)) => {
+                rsvg_log!("circular reference in element {}", node);
+                return Err(RenderingError::CircularReference);
+            }
+
+            Err(AcquireError::MaxReferencesExceeded) => {
+                return Err(RenderingError::InstancingLimit);
+            }
+
+            Err(AcquireError::InvalidLinkType(_)) => unreachable!(),
+
+            Err(AcquireError::LinkNotFound(fragment)) => {
+                rsvg_log!("element {} references nonexistent \"{}\"", node, fragment);
+                return Ok(self.empty_bbox());
+            }
+        };
+
+        let values = cascaded.get();
+        let params = self.get_view_params();
+        let use_rect = use_.get_rect(values, &params);
+
+        // width or height set to 0 disables rendering of the element
+        // https://www.w3.org/TR/SVG/struct.html#UseElementWidthAttribute
+        if use_rect.is_empty() {
+            return Ok(self.empty_bbox());
+        }
+
+        let child = acquired.get();
+
+        if child.borrow().get_type() != NodeType::Symbol {
+            let cr = self.get_cairo_context();
+            cr.translate(use_rect.x0, use_rect.y0);
+
+            self.with_discrete_layer(node, values, clipping, &mut |dc| {
+                dc.draw_node_from_stack(
+                    &CascadedValues::new_from_values(&child, values),
+                    &child,
+                    clipping,
+                )
+            })
+        } else {
+            let node_data = child.borrow();
+            let symbol = node_data.get_impl::<Symbol>();
+
+            let clip_mode = if !values.is_overflow()
+                || (values.overflow == Overflow::Visible && child.borrow().is_overflow())
+            {
+                Some(ClipMode::ClipToVbox)
+            } else {
+                None
+            };
+
+            self.with_discrete_layer(node, values, clipping, &mut |dc| {
+                let _params = dc.push_new_viewport(
+                    symbol.get_viewbox(),
+                    use_rect,
+                    symbol.get_preserve_aspect_ratio(),
+                    clip_mode,
+                );
+
+                child.draw_children(
+                    &CascadedValues::new_from_values(&child, values),
+                    dc,
+                    clipping,
+                )
+            })
         }
     }
 }
 
 #[derive(Debug)]
-pub struct CompositingAffines {
+struct CompositingAffines {
     pub outside_temporary_surface: cairo::Matrix,
     pub initial: cairo::Matrix,
     pub for_temporary_surface: cairo::Matrix,
@@ -946,10 +1172,44 @@ impl CompositingAffines {
     }
 }
 
+// Returns (clip_in_user_space, clip_in_object_space), both Option<RsvgNode>
+fn get_clip_in_user_and_object_space(
+    acquired_nodes: &mut AcquiredNodes,
+    clip_uri: Option<&Fragment>,
+) -> (Option<RsvgNode>, Option<RsvgNode>) {
+    clip_uri
+        .and_then(|fragment| acquired_nodes.acquire(fragment, &[NodeType::ClipPath]).ok())
+        .and_then(|acquired| {
+            let clip_node = acquired.get().clone();
+
+            let units = clip_node.borrow().get_impl::<ClipPath>().get_units();
+
+            match units {
+                CoordUnits::UserSpaceOnUse => Some((Some(clip_node), None)),
+                CoordUnits::ObjectBoundingBox => Some((None, Some(clip_node))),
+            }
+        })
+        .unwrap_or((None, None))
+}
+
+fn acquire_paint_server(
+    acquired_nodes: &mut AcquiredNodes,
+    fragment: &Fragment,
+) -> Result<AcquiredNode, AcquireError> {
+    acquired_nodes.acquire(
+        fragment,
+        &[
+            NodeType::LinearGradient,
+            NodeType::RadialGradient,
+            NodeType::Pattern,
+        ],
+    )
+}
+
 fn compute_stroke_and_fill_box(cr: &cairo::Context, values: &ComputedValues) -> BoundingBox {
     let affine = cr.get_matrix();
 
-    let mut bbox = BoundingBox::new(&affine);
+    let mut bbox = BoundingBox::new().with_affine(affine);
 
     // Dropping the precision of cairo's bezier subdivision, yielding 2x
     // _rendering_ time speedups, are these rather expensive operations
@@ -965,19 +1225,28 @@ fn compute_stroke_and_fill_box(cr: &cairo::Context, values: &ComputedValues) ->
     // paths for the icon's shape.  We need to be able to compute the bounding
     // rectangle's extents, even when it has no fill nor stroke.
 
-    let fb = BoundingBox::new(&affine).with_ink_extents(cr.fill_extents());
+    let (x0, y0, x1, y1) = cr.fill_extents();
+    let fb = BoundingBox::new()
+        .with_affine(affine)
+        .with_ink_rect(Rect::new(x0, y0, x1, y1));
     bbox.insert(&fb);
 
     // Bounding box for stroke
 
     if values.stroke.0 != PaintServer::None {
-        let sb = BoundingBox::new(&affine).with_ink_extents(cr.stroke_extents());
+        let (x0, y0, x1, y1) = cr.stroke_extents();
+        let sb = BoundingBox::new()
+            .with_affine(affine)
+            .with_ink_rect(Rect::new(x0, y0, x1, y1));
         bbox.insert(&sb);
     }
 
     // objectBoundingBox
 
-    let ob = BoundingBox::new(&affine).with_extents(cr.path_extents());
+    let (x0, y0, x1, y1) = cr.path_extents();
+    let ob = BoundingBox::new()
+        .with_affine(affine)
+        .with_rect(Rect::new(x0, y0, x1, y1));
     bbox.insert(&ob);
 
     // restore tolerance
@@ -987,6 +1256,19 @@ fn compute_stroke_and_fill_box(cr: &cairo::Context, values: &ComputedValues) ->
     bbox
 }
 
+/// escape quotes and backslashes with backslash
+fn escape_link_target(value: &str) -> Cow<'_, str> {
+    static REGEX: Lazy<Regex> = Lazy::new(|| Regex::new(r"['\\]").unwrap());
+
+    REGEX.replace_all(value, |caps: &Captures<'_>| {
+        match caps.get(0).unwrap().as_str() {
+            "'" => "\\'".to_owned(),
+            "\\" => "\\\\".to_owned(),
+            _ => unreachable!(),
+        }
+    })
+}
+
 impl From<StrokeLinejoin> for cairo::LineJoin {
     fn from(j: StrokeLinejoin) -> cairo::LineJoin {
         match j {
@@ -1034,48 +1316,6 @@ impl From<ShapeRendering> for cairo::Antialias {
     }
 }
 
-#[derive(Default, Clone, Copy, Debug, PartialEq)]
-#[repr(C)]
-pub struct RsvgRectangle {
-    pub x: f64,
-    pub y: f64,
-    pub width: f64,
-    pub height: f64,
-}
-
-impl From<cairo::Rectangle> for RsvgRectangle {
-    fn from(r: cairo::Rectangle) -> RsvgRectangle {
-        RsvgRectangle {
-            x: r.x,
-            y: r.y,
-            width: r.width,
-            height: r.height,
-        }
-    }
-}
-
-impl From<ViewBox> for RsvgRectangle {
-    fn from(vb: ViewBox) -> RsvgRectangle {
-        RsvgRectangle {
-            x: vb.x,
-            y: vb.y,
-            width: vb.width,
-            height: vb.height,
-        }
-    }
-}
-
-impl From<RsvgRectangle> for cairo::Rectangle {
-    fn from(r: RsvgRectangle) -> cairo::Rectangle {
-        cairo::Rectangle {
-            x: r.x,
-            y: r.y,
-            width: r.width,
-            height: r.height,
-        }
-    }
-}
-
 pub struct AcquiredNode {
     stack: Option<Rc<RefCell<NodeStack>>>,
     node: RsvgNode,
@@ -1097,8 +1337,18 @@ impl AcquiredNode {
     }
 }
 
+/// This helper struct is used when looking up urls to other nodes.
+/// Its methods do recursion checking and thereby avoid infinite loops.
+///
+/// Malformed SVGs, for example, may reference a marker by its IRI, but
+/// the object referenced by the IRI is not a marker.
+///
+/// Note that if you acquire a node, you have to release it before trying to
+/// acquire it again.  If you acquire a node "#foo" and don't release it before
+/// trying to acquire "foo" again, you will obtain a None the second time.
 struct AcquiredNodes {
     document: Rc<Document>,
+    num_elements_acquired: usize,
     node_stack: Rc<RefCell<NodeStack>>,
 }
 
@@ -1106,6 +1356,7 @@ impl AcquiredNodes {
     fn new(document: Rc<Document>) -> AcquiredNodes {
         AcquiredNodes {
             document,
+            num_elements_acquired: 0,
             node_stack: Rc::new(RefCell::new(NodeStack::new())),
         }
     }
@@ -1120,8 +1371,8 @@ impl AcquiredNodes {
             // Maybe we should have the following stages:
             //   - load main SVG XML
             //
-            //   - load secondary SVG XML and other files like images;
-            //     all document::Resources and document::Images loaded
+            //   - load secondary SVG XML and other files like images; all document::Resources and
+            //     document::Images loaded
             //
             //   - Now that all files are loaded, resolve URL references
             AcquireError::LinkNotFound(fragment.clone())
@@ -1139,15 +1390,27 @@ impl AcquiredNodes {
         }
     }
 
+    /// Acquires a node.
+    /// Specify `node_types` when expecting the node to be of a particular type,
+    /// or use an empty slice for `node_types` if you want a node of any type.
+    /// Nodes acquired by this function must be released in reverse acquiring order.
     fn acquire(
-        &self,
+        &mut self,
         fragment: &Fragment,
         node_types: &[NodeType],
     ) -> Result<AcquiredNode, AcquireError> {
+        self.num_elements_acquired += 1;
+
+        // This is a mitigation for SVG files that try to instance a huge number of
+        // elements via <use>, recursive patterns, etc.  See limits.rs for details.
+        if self.num_elements_acquired > limits::MAX_REFERENCED_ELEMENTS {
+            return Err(AcquireError::MaxReferencesExceeded);
+        }
+
         let node = self.lookup_node(fragment, node_types)?;
 
         if node_is_accessed_by_reference(&node) {
-            self.push_node_ref(&node)
+            self.acquire_ref(&node)
         } else {
             Ok(AcquiredNode {
                 stack: None,
@@ -1156,7 +1419,7 @@ impl AcquiredNodes {
         }
     }
 
-    fn push_node_ref(&self, node: &RsvgNode) -> Result<AcquiredNode, AcquireError> {
+    fn acquire_ref(&self, node: &RsvgNode) -> Result<AcquiredNode, AcquireError> {
         if self.node_stack.borrow().contains(&node) {
             Err(AcquireError::CircularReference(node.clone()))
         } else {
