@@ -1,12 +1,10 @@
 //! Filter primitive subregion computation.
-use cairo;
-
 use crate::bbox::BoundingBox;
 use crate::drawing_ctx::DrawingCtx;
 use crate::length::*;
-use crate::rect::IRect;
+use crate::rect::{IRect, Rect};
 
-use super::context::{FilterContext, FilterInput, FilterOutput};
+use super::context::{FilterContext, FilterInput};
 
 /// A helper type for filter primitive subregion computation.
 #[derive(Clone, Copy)]
@@ -40,7 +38,7 @@ impl<'a> BoundsBuilder<'a> {
         Self {
             ctx,
             // The matrix is paffine because we're using that fact in apply_properties().
-            bbox: BoundingBox::new(&ctx.paffine()),
+            bbox: BoundingBox::new().with_affine(ctx.paffine()),
             standard_input_was_referenced: false,
             x,
             y,
@@ -63,18 +61,8 @@ impl<'a> BoundsBuilder<'a> {
             FilterInput::StandardInput(_) => {
                 self.standard_input_was_referenced = true;
             }
-            FilterInput::PrimitiveOutput(FilterOutput {
-                bounds: IRect { x0, y0, x1, y1 },
-                ..
-            }) => {
-                let rect = cairo::Rectangle {
-                    x: f64::from(x0),
-                    y: f64::from(y0),
-                    width: f64::from(x1 - x0),
-                    height: f64::from(y1 - y0),
-                };
-
-                let input_bbox = BoundingBox::new(&cairo::Matrix::identity()).with_rect(Some(rect));
+            FilterInput::PrimitiveOutput(ref output) => {
+                let input_bbox = BoundingBox::new().with_rect(output.bounds.into());
                 self.bbox.insert(&input_bbox);
             }
         }
@@ -82,23 +70,26 @@ impl<'a> BoundsBuilder<'a> {
         self
     }
 
-    /// Returns the final pixel bounds.
-    #[inline]
-    pub fn into_irect(self, draw_ctx: &mut DrawingCtx) -> IRect {
+    /// Returns the final exact bounds.
+    pub fn into_rect(self, draw_ctx: &mut DrawingCtx) -> Rect {
         let mut bbox = self.apply_properties(draw_ctx);
 
         let effects_region = self.ctx.effects_region();
         bbox.clip(&effects_region);
 
-        bbox.rect.unwrap().into()
+        bbox.rect.unwrap()
+    }
+
+    /// Returns the final pixel bounds.
+    pub fn into_irect(self, draw_ctx: &mut DrawingCtx) -> IRect {
+        self.into_rect(draw_ctx).into()
     }
 
     /// Returns the final pixel bounds without clipping to the filter effects region.
     ///
     /// Used by feImage.
-    #[inline]
-    pub fn into_irect_without_clipping(self, draw_ctx: &mut DrawingCtx) -> IRect {
-        self.apply_properties(draw_ctx).rect.unwrap().into()
+    pub fn into_rect_without_clipping(self, draw_ctx: &mut DrawingCtx) -> Rect {
+        self.apply_properties(draw_ctx).rect.unwrap()
     }
 
     /// Applies the filter primitive properties.
@@ -108,7 +99,8 @@ impl<'a> BoundsBuilder<'a> {
             let effects_region = self.ctx.effects_region();
 
             // Clear out the rect.
-            self.bbox = self.bbox.with_rect(None);
+            self.bbox.clear();
+
             // Convert into the paffine coordinate system.
             self.bbox.insert(&effects_region);
         }
@@ -123,21 +115,25 @@ impl<'a> BoundsBuilder<'a> {
             let rect = self.bbox.rect.as_mut().unwrap();
 
             if let Some(x) = self.x {
-                rect.x = x.normalize(values, &params);
+                let w = rect.width();
+                rect.x0 = x.normalize(values, &params);
+                rect.x1 = rect.x0 + w;
             }
             if let Some(y) = self.y {
-                rect.y = y.normalize(values, &params);
+                let h = rect.height();
+                rect.y0 = y.normalize(values, &params);
+                rect.y1 = rect.y0 + h;
             }
             if let Some(width) = self.width {
-                rect.width = width.normalize(values, &params);
+                rect.x1 = rect.x0 + width.normalize(values, &params);
             }
             if let Some(height) = self.height {
-                rect.height = height.normalize(values, &params);
+                rect.y1 = rect.y0 + height.normalize(values, &params);
             }
         }
 
         // Convert into the surface coordinate system.
-        let mut bbox = BoundingBox::new(&cairo::Matrix::identity());
+        let mut bbox = BoundingBox::new();
         bbox.insert(&self.bbox);
         bbox
     }
