@@ -1,15 +1,15 @@
-use cairo::{self, ImageSurface};
-use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
+use cssparser::Parser;
+use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use nalgebra::{Matrix3, Matrix4x5, Matrix5, Vector5};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
-use crate::number_list::{NumberList, NumberListError, NumberListLength};
-use crate::parsers::{self, ParseError};
+use crate::number_list::{NumberList, NumberListLength};
+use crate::parsers::{Parse, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::surface_utils::{
-    iterators::Pixels, shared_surface::SharedImageSurface, ImageSurfaceDataExt, Pixel,
+    iterators::Pixels, shared_surface::ExclusiveImageSurface, ImageSurfaceDataExt, Pixel,
 };
 use crate::util::clamp;
 
@@ -25,6 +25,12 @@ enum OperationType {
     LuminanceToAlpha,
 }
 
+impl Default for OperationType {
+    fn default() -> Self {
+        OperationType::Matrix
+    }
+}
+
 /// The `feColorMatrix` filter primitive.
 pub struct FeColorMatrix {
     base: PrimitiveWithInput,
@@ -49,12 +55,12 @@ impl NodeTrait for FeColorMatrix {
         self.base.set_atts(parent, pbag)?;
 
         // First, determine the operation type.
-        let mut operation_type = OperationType::Matrix;
+        let mut operation_type = Default::default();
         for (attr, value) in pbag
             .iter()
-            .filter(|(attr, _)| attr.expanded() == expanded_name!(svg "type"))
+            .filter(|(attr, _)| attr.expanded() == expanded_name!("", "type"))
         {
-            operation_type = OperationType::parse(attr, value)?;
+            operation_type = attr.parse(value)?;
         }
 
         // Now read the matrix correspondingly.
@@ -73,36 +79,27 @@ impl NodeTrait for FeColorMatrix {
         } else {
             for (attr, value) in pbag
                 .iter()
-                .filter(|(attr, _)| attr.expanded() == expanded_name!(svg "values"))
+                .filter(|(attr, _)| attr.expanded() == expanded_name!("", "values"))
             {
                 let new_matrix = match operation_type {
                     OperationType::LuminanceToAlpha => unreachable!(),
                     OperationType::Matrix => {
-                        let NumberList(v) = NumberList::parse_str(
-                            value,
-                            NumberListLength::Exact(20),
-                        )
-                        .map_err(|err| {
-                            NodeError::parse_error(
-                                attr,
-                                match err {
-                                    NumberListError::IncorrectNumberOfElements => {
-                                        ParseError::new("incorrect number of elements: expected 20")
-                                    }
-                                    NumberListError::Parse(err) => err,
-                                },
-                            )
-                        })?;
+                        let NumberList(v) =
+                            NumberList::parse_str(value, NumberListLength::Exact(20))
+                                .attribute(attr)?;
                         let matrix = Matrix4x5::from_row_slice(&v);
                         let mut matrix = matrix.fixed_resize(0.0);
                         matrix[(4, 4)] = 1.0;
                         matrix
                     }
                     OperationType::Saturate => {
-                        let s = parsers::number(value).attribute(attr.clone())?;
-                        if s < 0.0 || s > 1.0 {
-                            return Err(NodeError::value_error(attr, "expected value from 0 to 1"));
-                        }
+                        let s: f64 = attr.parse_and_validate(value, |s| {
+                            if s < 0.0 || s > 1.0 {
+                                Err(ValueErrorKind::value_error("expected value from 0 to 1"))
+                            } else {
+                                Ok(s)
+                            }
+                        })?;
 
                         #[cfg_attr(rustfmt, rustfmt_skip)]
                         Matrix5::new(
@@ -114,7 +111,7 @@ impl NodeTrait for FeColorMatrix {
                         )
                     }
                     OperationType::HueRotate => {
-                        let degrees = parsers::number(value).attribute(attr.clone())?;
+                        let degrees: f64 = attr.parse(value)?;
                         let (sin, cos) = degrees.to_radians().sin_cos();
 
                         #[cfg_attr(rustfmt, rustfmt_skip)]
@@ -169,17 +166,14 @@ impl FilterEffect for FeColorMatrix {
             .add_input(&input)
             .into_irect(draw_ctx);
 
-        let mut output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
+        let mut surface = ExclusiveImageSurface::new(
             ctx.source_graphic().width(),
             ctx.source_graphic().height(),
+            input.surface().surface_type(),
         )?;
 
-        let output_stride = output_surface.get_stride() as usize;
-        {
-            let mut output_data = output_surface.get_data().unwrap();
-
-            for (x, y, pixel) in Pixels::new(input.surface(), bounds) {
+        surface.modify(&mut |data, stride| {
+            for (x, y, pixel) in Pixels::within(input.surface(), bounds) {
                 let alpha = f64::from(pixel.a) / 255f64;
 
                 let pixel_vec = if alpha == 0.0 {
@@ -207,14 +201,14 @@ impl FilterEffect for FeColorMatrix {
                     a: ((new_alpha * 255f64) + 0.5) as u8,
                 };
 
-                output_data.set_pixel(output_stride, output_pixel, x, y);
+                data.set_pixel(stride, output_pixel, x, y);
             }
-        }
+        });
 
         Ok(FilterResult {
             name: self.base.result.clone(),
             output: FilterOutput {
-                surface: SharedImageSurface::new(output_surface, input.surface().surface_type())?,
+                surface: surface.share()?,
                 bounds,
             },
         })
@@ -226,17 +220,14 @@ impl FilterEffect for FeColorMatrix {
     }
 }
 
-impl OperationType {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "matrix" => Ok(OperationType::Matrix),
-            "saturate" => Ok(OperationType::Saturate),
-            "hueRotate" => Ok(OperationType::HueRotate),
-            "luminanceToAlpha" => Ok(OperationType::LuminanceToAlpha),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for OperationType {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "matrix" => OperationType::Matrix,
+            "saturate" => OperationType::Saturate,
+            "hueRotate" => OperationType::HueRotate,
+            "luminanceToAlpha" => OperationType::LuminanceToAlpha,
+        )?)
     }
 }
