@@ -1,19 +1,16 @@
 use std::cmp::min;
 
-use cairo::{self, ImageSurface};
-use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
+use cssparser::Parser;
+use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{NodeResult, NodeTrait, NodeType, RsvgNode};
-use crate::number_list::{NumberList, NumberListError, NumberListLength};
-use crate::parsers::{self, ParseError};
+use crate::number_list::{NumberList, NumberListLength};
+use crate::parsers::{Parse, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::surface_utils::{
-    iterators::Pixels,
-    shared_surface::SharedImageSurface,
-    ImageSurfaceDataExt,
-    Pixel,
+    iterators::Pixels, shared_surface::ExclusiveImageSurface, ImageSurfaceDataExt, Pixel,
 };
 use crate::util::clamp;
 
@@ -54,7 +51,6 @@ enum Channel {
 }
 
 /// Component transfer function types.
-#[derive(Debug, Clone, Copy)]
 enum FunctionType {
     Identity,
     Table,
@@ -63,19 +59,16 @@ enum FunctionType {
     Gamma,
 }
 
-impl FunctionType {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "identity" => Ok(FunctionType::Identity),
-            "table" => Ok(FunctionType::Table),
-            "discrete" => Ok(FunctionType::Discrete),
-            "linear" => Ok(FunctionType::Linear),
-            "gamma" => Ok(FunctionType::Gamma),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for FunctionType {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "identity" => FunctionType::Identity,
+            "table" => FunctionType::Table,
+            "discrete" => FunctionType::Discrete,
+            "linear" => FunctionType::Linear,
+            "gamma" => FunctionType::Gamma,
+        )?)
     }
 }
 
@@ -213,37 +206,19 @@ macro_rules! func_x {
             ) -> NodeResult {
                 for (attr, value) in pbag.iter() {
                     match attr.expanded() {
-                        expanded_name!(svg "type") => {
-                            self.function_type = FunctionType::parse(attr, value)?
-                        }
-                        expanded_name!(svg "tableValues") => {
+                        expanded_name!("", "type") => self.function_type = attr.parse(value)?,
+                        expanded_name!("", "tableValues") => {
                             let NumberList(v) =
-                                NumberList::parse_str(value, NumberListLength::Unbounded).map_err(
-                                    |err| {
-                                        if let NumberListError::Parse(err) = err {
-                                            NodeError::parse_error(attr, err)
-                                        } else {
-                                            panic!("unexpected number list error");
-                                        }
-                                    },
-                                )?;
+                                NumberList::parse_str(value, NumberListLength::Unbounded)
+                                    .attribute(attr)?;
                             self.table_values = v;
                         }
-                        expanded_name!(svg "slope") => {
-                            self.slope = parsers::number(value).attribute(attr)?
-                        }
-                        expanded_name!(svg "intercept") => {
-                            self.intercept = parsers::number(value).attribute(attr)?
-                        }
-                        expanded_name!(svg "amplitude") => {
-                            self.amplitude = parsers::number(value).attribute(attr)?
-                        }
-                        expanded_name!(svg "exponent") => {
-                            self.exponent = parsers::number(value).attribute(attr)?
-                        }
-                        expanded_name!(svg "offset") => {
-                            self.offset = parsers::number(value).attribute(attr)?
-                        }
+                        expanded_name!("", "slope") => self.slope = attr.parse(value)?,
+                        expanded_name!("", "intercept") => self.intercept = attr.parse(value)?,
+                        expanded_name!("", "amplitude") => self.amplitude = attr.parse(value)?,
+                        expanded_name!("", "exponent") => self.exponent = attr.parse(value)?,
+                        expanded_name!("", "offset") => self.offset = attr.parse(value)?,
+
                         _ => (),
                     }
                 }
@@ -304,10 +279,10 @@ impl FilterEffect for FeComponentTransfer {
             .into_irect(draw_ctx);
 
         // Create the output surface.
-        let mut output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
+        let mut surface = ExclusiveImageSurface::new(
             ctx.source_graphic().width(),
             ctx.source_graphic().height(),
+            input.surface().surface_type(),
         )?;
 
         // Get a node for every pixel component.
@@ -382,11 +357,8 @@ impl FilterEffect for FeComponentTransfer {
         let compute_a = |alpha| compute_a(&params_a, alpha);
 
         // Do the actual processing.
-        let output_stride = output_surface.get_stride() as usize;
-        {
-            let mut output_data = output_surface.get_data().unwrap();
-
-            for (x, y, pixel) in Pixels::new(input.surface(), bounds) {
+        surface.modify(&mut |data, stride| {
+            for (x, y, pixel) in Pixels::within(input.surface(), bounds) {
                 let alpha = f64::from(pixel.a) / 255f64;
                 let new_alpha = compute_a(alpha);
 
@@ -397,14 +369,14 @@ impl FilterEffect for FeComponentTransfer {
                     a: ((new_alpha * 255f64) + 0.5) as u8,
                 };
 
-                output_data.set_pixel(output_stride, output_pixel, x, y);
+                data.set_pixel(stride, output_pixel, x, y);
             }
-        }
+        });
 
         Ok(FilterResult {
             name: self.base.result.clone(),
             output: FilterOutput {
-                surface: SharedImageSurface::new(output_surface, input.surface().surface_type())?,
+                surface: surface.share()?,
                 bounds,
             },
         })
