@@ -2,8 +2,6 @@ use std::cell::RefCell;
 use std::collections::HashMap;
 use std::f64;
 
-use cairo;
-
 use crate::bbox::BoundingBox;
 use crate::coord_units::CoordUnits;
 use crate::drawing_ctx::{DrawingCtx, ViewParams};
@@ -12,11 +10,13 @@ use crate::node::RsvgNode;
 use crate::paint_server::PaintServer;
 use crate::properties::ComputedValues;
 use crate::rect::IRect;
-use crate::surface_utils::shared_surface::{SharedImageSurface, SurfaceType};
+use crate::surface_utils::shared_surface::{
+    ExclusiveImageSurface, SharedImageSurface, SurfaceType,
+};
 use crate::unit_interval::UnitInterval;
 
 use super::error::FilterError;
-use super::input::Input;
+use super::input::{CustomIdent, Input};
 
 /// A filter primitive output.
 #[derive(Debug, Clone)]
@@ -32,7 +32,7 @@ pub struct FilterOutput {
 #[derive(Debug, Clone)]
 pub struct FilterResult {
     /// The name of this result: the value of the `result` attribute.
-    pub name: Option<String>,
+    pub name: Option<CustomIdent>,
 
     /// The output.
     pub output: FilterOutput,
@@ -60,7 +60,7 @@ pub struct FilterContext {
     /// Output of the last filter primitive.
     last_result: Option<FilterOutput>,
     /// Surfaces of the previous filter primitives by name.
-    previous_results: HashMap<String, FilterOutput>,
+    previous_results: HashMap<CustomIdent, FilterOutput>,
     /// The background surface. Computed lazily.
     background_surface: RefCell<Option<Result<SharedImageSurface, FilterError>>>,
     /// The filter effects region.
@@ -107,12 +107,7 @@ impl FilterContext {
 
         // The rect can be empty (for example, if the filter is applied to an empty group).
         // However, with userSpaceOnUse it's still possible to create images with a filter.
-        let bbox_rect = node_bbox.rect.unwrap_or(cairo::Rectangle {
-            x: 0.0,
-            y: 0.0,
-            width: 0.0,
-            height: 0.0,
-        });
+        let bbox_rect = node_bbox.rect.unwrap_or_default();
 
         let node_data = filter_node.borrow();
         let filter = node_data.get_impl::<Filter>();
@@ -121,12 +116,12 @@ impl FilterContext {
             CoordUnits::UserSpaceOnUse => cr_affine,
             CoordUnits::ObjectBoundingBox => {
                 let affine = cairo::Matrix::new(
-                    bbox_rect.width,
-                    0f64,
-                    0f64,
-                    bbox_rect.height,
-                    bbox_rect.x,
-                    bbox_rect.y,
+                    bbox_rect.width(),
+                    0.0,
+                    0.0,
+                    bbox_rect.height(),
+                    bbox_rect.x0,
+                    bbox_rect.y0,
                 );
                 cairo::Matrix::multiply(&affine, &cr_affine)
             }
@@ -136,12 +131,12 @@ impl FilterContext {
             CoordUnits::UserSpaceOnUse => cr_affine,
             CoordUnits::ObjectBoundingBox => {
                 let affine = cairo::Matrix::new(
-                    bbox_rect.width,
-                    0f64,
-                    0f64,
-                    bbox_rect.height,
-                    bbox_rect.x,
-                    bbox_rect.y,
+                    bbox_rect.width(),
+                    0.0,
+                    0.0,
+                    bbox_rect.height(),
+                    bbox_rect.x0,
+                    bbox_rect.y0,
                 );
                 cairo::Matrix::multiply(&affine, &cr_affine)
             }
@@ -177,26 +172,12 @@ impl FilterContext {
         &self.computed_from_node_being_filtered
     }
 
-    /// Returns the surface corresponding to the last filter primitive's result.
-    #[inline]
-    pub fn last_result(&self) -> Option<&FilterOutput> {
-        self.last_result.as_ref()
-    }
-
     /// Returns the surface corresponding to the source graphic.
     #[inline]
     pub fn source_graphic(&self) -> &SharedImageSurface {
         &self.source_surface
     }
 
-    /// Returns the surface containing the source graphic alpha.
-    #[inline]
-    pub fn source_alpha(&self, bounds: IRect) -> Result<SharedImageSurface, FilterError> {
-        self.source_surface
-            .extract_alpha(bounds)
-            .map_err(FilterError::CairoError)
-    }
-
     /// Returns the surface corresponding to the background image snapshot.
     pub fn background_image(
         &self,
@@ -219,38 +200,17 @@ impl FilterContext {
         let mut bg = self.background_surface.borrow_mut();
 
         *bg = Some(
-            cairo::ImageSurface::create(
-                cairo::Format::ARgb32,
-                self.source_surface.width(),
-                self.source_surface.height(),
-            )
-            .map_err(FilterError::CairoError)
-            .and_then(|s| {
-                draw_ctx.get_snapshot(&s);
-                SharedImageSurface::new(s, SurfaceType::SRgb).map_err(FilterError::CairoError)
-            }),
+            draw_ctx
+                .get_snapshot(self.source_surface.width(), self.source_surface.height())
+                .map_err(FilterError::CairoError),
         );
 
         // Return the only existing reference as immutable.
-        bg.as_ref().unwrap().as_ref().map(|surf| surf.clone()).map_err(|&s| s)
-    }
-
-    /// Returns the surface containing the background image snapshot alpha.
-    #[inline]
-    pub fn background_alpha(
-        &self,
-        draw_ctx: &DrawingCtx,
-        bounds: IRect,
-    ) -> Result<SharedImageSurface, FilterError> {
-        self.background_image(draw_ctx)?
-            .extract_alpha(bounds)
-            .map_err(FilterError::CairoError)
-    }
-
-    /// Returns the output of the filter primitive by its result name.
-    #[inline]
-    pub fn filter_output(&self, name: &str) -> Option<&FilterOutput> {
-        self.previous_results.get(name)
+        bg.as_ref()
+            .unwrap()
+            .as_ref()
+            .map(|surf| surf.clone())
+            .map_err(|&s| s)
     }
 
     /// Converts this `FilterContext` into the surface corresponding to the output of the filter
@@ -262,18 +222,11 @@ impl FilterContext {
     pub fn into_output(self) -> Result<SharedImageSurface, cairo::Status> {
         match self.last_result {
             Some(FilterOutput { surface, bounds }) => surface.to_srgb(bounds),
-            None => {
-                let empty_surface = cairo::ImageSurface::create(
-                    cairo::Format::ARgb32,
-                    self.source_surface.width(),
-                    self.source_surface.height(),
-                )?;
-
-                Ok(SharedImageSurface::new(
-                    empty_surface,
-                    SurfaceType::AlphaOnly,
-                )?)
-            }
+            None => SharedImageSurface::empty(
+                self.source_surface.width(),
+                self.source_surface.height(),
+                SurfaceType::AlphaOnly,
+            ),
         }
     }
 
@@ -323,34 +276,38 @@ impl FilterContext {
         draw_ctx: &mut DrawingCtx,
         paint_server: &PaintServer,
         opacity: UnitInterval,
-    ) -> Result<cairo::ImageSurface, cairo::Status> {
-        let surface = cairo::ImageSurface::create(
-            cairo::Format::ARgb32,
+    ) -> Result<SharedImageSurface, cairo::Status> {
+        let mut surface = ExclusiveImageSurface::new(
             self.source_surface.width(),
             self.source_surface.height(),
+            SurfaceType::SRgb,
         )?;
 
-        let cr_save = draw_ctx.get_cairo_context();
-        let cr = cairo::Context::new(&surface);
-        draw_ctx.set_cairo_context(&cr);
+        surface.draw(&mut |cr| {
+            let cr_save = draw_ctx.get_cairo_context();
+            draw_ctx.set_cairo_context(&cr);
 
-        // FIXME: we are ignoring the following error; propagate it upstream
-        let _ = draw_ctx
-            .set_source_paint_server(
-                paint_server,
-                opacity,
-                &self.node_bbox,
-                self.computed_from_node_being_filtered.color.0,
-            )
-            .and_then(|had_paint_server| {
-                if had_paint_server {
-                    cr.paint();
-                }
-                Ok(())
-            });
+            // FIXME: we are ignoring the following error; propagate it upstream
+            let _ = draw_ctx
+                .set_source_paint_server(
+                    paint_server,
+                    opacity,
+                    &self.node_bbox,
+                    self.computed_from_node_being_filtered.color.0,
+                )
+                .and_then(|had_paint_server| {
+                    if had_paint_server {
+                        cr.paint();
+                    }
+                    Ok(())
+                });
 
-        draw_ctx.set_cairo_context(&cr_save);
-        Ok(surface)
+            draw_ctx.set_cairo_context(&cr_save);
+
+            Ok(())
+        })?;
+
+        surface.share()
     }
 
     /// Retrieves the filter input surface according to the SVG rules.
@@ -365,8 +322,8 @@ impl FilterContext {
             // No value => use the last result.
             // As per the SVG spec, if the filter primitive is the first in the chain, return the
             // source graphic.
-            if let Some(output) = self.last_result().cloned() {
-                return Ok(FilterInput::PrimitiveOutput(output));
+            if let Some(output) = self.last_result.as_ref() {
+                return Ok(FilterInput::PrimitiveOutput(output.clone()));
             } else {
                 return Ok(FilterInput::StandardInput(self.source_graphic().clone()));
             }
@@ -376,35 +333,39 @@ impl FilterContext {
 
         match *in_.unwrap() {
             Input::SourceGraphic => Ok(FilterInput::StandardInput(self.source_graphic().clone())),
+
             Input::SourceAlpha => self
-                .source_alpha(self.effects_region().rect.unwrap().into())
+                .source_graphic()
+                .extract_alpha(self.effects_region().rect.unwrap().into())
+                .map_err(FilterError::CairoError)
                 .map(FilterInput::StandardInput),
+
             Input::BackgroundImage => self
                 .background_image(draw_ctx)
                 .map(FilterInput::StandardInput),
+
             Input::BackgroundAlpha => self
-                .background_alpha(draw_ctx, self.effects_region().rect.unwrap().into())
+                .background_image(draw_ctx)
+                .and_then(|surface| {
+                    surface
+                        .extract_alpha(self.effects_region().rect.unwrap().into())
+                        .map_err(FilterError::CairoError)
+                })
                 .map(FilterInput::StandardInput),
 
             Input::FillPaint => self
                 .get_paint_server_surface(draw_ctx, &values.fill.0, values.fill_opacity.0)
                 .map_err(FilterError::CairoError)
-                .and_then(|surface| {
-                    SharedImageSurface::new(surface, SurfaceType::SRgb)
-                        .map_err(FilterError::CairoError)
-                })
                 .map(FilterInput::StandardInput),
+
             Input::StrokePaint => self
                 .get_paint_server_surface(draw_ctx, &values.stroke.0, values.stroke_opacity.0)
                 .map_err(FilterError::CairoError)
-                .and_then(|surface| {
-                    SharedImageSurface::new(surface, SurfaceType::SRgb)
-                        .map_err(FilterError::CairoError)
-                })
                 .map(FilterInput::StandardInput),
 
             Input::FilterOutput(ref name) => self
-                .filter_output(name)
+                .previous_results
+                .get(name)
                 .cloned()
                 .map(FilterInput::PrimitiveOutput)
                 .ok_or(FilterError::InvalidInput),
@@ -471,58 +432,3 @@ impl FilterInput {
         }
     }
 }
-
-#[cfg(test)]
-mod tests {
-    use super::*;
-    use crate::surface_utils::iterators::Pixels;
-
-    #[test]
-    fn test_extract_alpha() {
-        const WIDTH: i32 = 32;
-        const HEIGHT: i32 = 64;
-        const BOUNDS: IRect = IRect {
-            x0: 8,
-            x1: 24,
-            y0: 16,
-            y1: 48,
-        };
-        const FULL_BOUNDS: IRect = IRect {
-            x0: 0,
-            x1: WIDTH,
-            y0: 0,
-            y1: HEIGHT,
-        };
-
-        let mut surface =
-            cairo::ImageSurface::create(cairo::Format::ARgb32, WIDTH, HEIGHT).unwrap();
-
-        // Fill the surface with some data.
-        {
-            let mut data = surface.get_data().unwrap();
-
-            let mut counter = 0u16;
-            for x in data.iter_mut() {
-                *x = counter as u8;
-                counter = (counter + 1) % 256;
-            }
-        }
-
-        let surface = SharedImageSurface::new(surface, SurfaceType::SRgb).unwrap();
-        let alpha = surface.extract_alpha(BOUNDS).unwrap();
-
-        for (x, y, p, pa) in
-            Pixels::new(&surface, FULL_BOUNDS).map(|(x, y, p)| (x, y, p, alpha.get_pixel(x, y)))
-        {
-            assert_eq!(pa.r, 0);
-            assert_eq!(pa.g, 0);
-            assert_eq!(pa.b, 0);
-
-            if !BOUNDS.contains(x as i32, y as i32) {
-                assert_eq!(pa.a, 0);
-            } else {
-                assert_eq!(pa.a, p.a);
-            }
-        }
-    }
-}
