@@ -1,20 +1,18 @@
-use cairo::{self, ImageSurface};
+use cssparser::Parser;
 use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
 use nalgebra::{DMatrix, Dynamic, VecStorage};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
-use crate::number_list::{NumberList, NumberListError, NumberListLength};
-use crate::parsers::{self, ParseError};
+use crate::number_list::{NumberList, NumberListLength};
+use crate::parsers::{NumberOptionalNumber, Parse, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::rect::IRect;
 use crate::surface_utils::{
     iterators::{PixelRectangle, Pixels},
-    shared_surface::SharedImageSurface,
-    EdgeMode,
-    ImageSurfaceDataExt,
-    Pixel,
+    shared_surface::ExclusiveImageSurface,
+    EdgeMode, ImageSurfaceDataExt, Pixel,
 };
 use crate::util::clamp;
 
@@ -62,61 +60,44 @@ impl NodeTrait for FeConvolveMatrix {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "order") => {
-                    self.order = parsers::integer_optional_integer(value)
-                        .attribute(attr.clone())
-                        .and_then(|(x, y)| {
-                            if x > 0 && y > 0 {
-                                Ok((x as u32, y as u32))
+                expanded_name!("", "order") => {
+                    let NumberOptionalNumber(x, y) =
+                        attr.parse_and_validate(value, |v: NumberOptionalNumber<i32>| {
+                            if v.0 > 0 && v.1 > 0 {
+                                Ok(v)
                             } else {
-                                Err(NodeError::value_error(
-                                    attr,
-                                    "values must be greater than 0",
+                                Err(ValueErrorKind::value_error("values must be greater than 0"))?
+                            }
+                        })?;
+                    self.order = (x as u32, y as u32);
+                }
+                expanded_name!("", "divisor") => {
+                    self.divisor = Some(attr.parse_and_validate(value, |x| {
+                        if x != 0.0 {
+                            Ok(x)
+                        } else {
+                            Err(ValueErrorKind::value_error("divisor cannot be equal to 0"))
+                        }
+                    })?)
+                }
+                expanded_name!("", "bias") => self.bias = attr.parse(value)?,
+                expanded_name!("", "edgeMode") => self.edge_mode = attr.parse(value)?,
+                expanded_name!("", "kernelUnitLength") => {
+                    let NumberOptionalNumber(x, y) =
+                        attr.parse_and_validate(value, |v: NumberOptionalNumber<f64>| {
+                            if v.0 > 0.0 && v.1 > 0.0 {
+                                Ok(v)
+                            } else {
+                                Err(ValueErrorKind::value_error(
+                                    "kernelUnitLength can't be less or equal to zero",
                                 ))
                             }
-                        })?
-                }
-                expanded_name!(svg "divisor") => {
-                    self.divisor = Some(parsers::number(value).attribute(attr.clone()).and_then(
-                        |x| {
-                            if x != 0.0 {
-                                Ok(x)
-                            } else {
-                                Err(NodeError::value_error(attr, "divisor cannot be equal to 0"))
-                            }
-                        },
-                    )?)
-                }
-                expanded_name!(svg "bias") => self.bias = parsers::number(value).attribute(attr)?,
-                expanded_name!(svg "edgeMode") => self.edge_mode = EdgeMode::parse(attr, value)?,
-                expanded_name!(svg "kernelUnitLength") => {
-                    self.kernel_unit_length = Some(
-                        parsers::number_optional_number(value)
-                            .attribute(attr.clone())
-                            .and_then(|(x, y)| {
-                                if x > 0.0 && y > 0.0 {
-                                    Ok((x, y))
-                                } else {
-                                    Err(NodeError::value_error(
-                                        attr,
-                                        "kernelUnitLength can't be less or equal to zero",
-                                    ))
-                                }
-                            })?,
-                    )
-                }
-                expanded_name!(svg "preserveAlpha") => {
-                    self.preserve_alpha = match value {
-                        "false" => false,
-                        "true" => true,
-                        _ => {
-                            return Err(NodeError::parse_error(
-                                attr,
-                                ParseError::new("expected false or true"),
-                            ));
-                        }
-                    }
+                        })?;
+
+                    self.kernel_unit_length = Some((x, y))
                 }
+                expanded_name!("", "preserveAlpha") => self.preserve_alpha = attr.parse(value)?,
+
                 _ => (),
             }
         }
@@ -124,33 +105,33 @@ impl NodeTrait for FeConvolveMatrix {
         // target_x and target_y depend on order.
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "targetX") => {
-                    self.target_x = Some(parsers::integer(value).attribute(attr.clone()).and_then(
-                        |x| {
-                            if x >= 0 && x < self.order.0 as i32 {
-                                Ok(x as u32)
+                expanded_name!("", "targetX") => {
+                    self.target_x = {
+                        let v = attr.parse_and_validate(value, |v: i32| {
+                            if v >= 0 && v < self.order.0 as i32 {
+                                Ok(v)
                             } else {
-                                Err(NodeError::value_error(
-                                    attr,
+                                Err(ValueErrorKind::value_error(
                                     "targetX must be greater or equal to zero and less than orderX",
                                 ))
                             }
-                        },
-                    )?)
+                        })?;
+                        Some(v as u32)
+                    }
                 }
-                expanded_name!(svg "targetY") => {
-                    self.target_y = Some(parsers::integer(value).attribute(attr.clone()).and_then(
-                        |x| {
-                            if x >= 0 && x < self.order.1 as i32 {
-                                Ok(x as u32)
+                expanded_name!("", "targetY") => {
+                    self.target_y = {
+                        let v = attr.parse_and_validate(value, |v: i32| {
+                            if v >= 0 && v < self.order.1 as i32 {
+                                Ok(v)
                             } else {
-                                Err(NodeError::value_error(
-                                    attr,
+                                Err(ValueErrorKind::value_error(
                                     "targetY must be greater or equal to zero and less than orderY",
                                 ))
                             }
-                        },
-                    )?)
+                        })?;
+                        Some(v as u32)
+                    }
                 }
                 _ => (),
             }
@@ -167,7 +148,7 @@ impl NodeTrait for FeConvolveMatrix {
         // Finally, parse the kernel matrix.
         for (attr, value) in pbag
             .iter()
-            .filter(|(attr, _)| attr.expanded() == expanded_name!(svg "kernelMatrix"))
+            .filter(|(attr, _)| attr.expanded() == expanded_name!("", "kernelMatrix"))
         {
             self.kernel_matrix = Some({
                 let number_of_elements = self.order.0 as usize * self.order.1 as usize;
@@ -175,24 +156,14 @@ impl NodeTrait for FeConvolveMatrix {
                 // #352: Parse as an unbounded list rather than exact length to prevent aborts due
                 //       to huge allocation attempts by underlying Vec::with_capacity().
                 let NumberList(v) = NumberList::parse_str(value, NumberListLength::Unbounded)
-                    .map_err(|err| {
-                        NodeError::parse_error(
-                            attr.clone(),
-                            match err {
-                                NumberListError::IncorrectNumberOfElements => unreachable!(),
-                                NumberListError::Parse(err) => err,
-                            },
-                        )
-                    })?;
+                    .attribute(attr.clone())?;
 
                 if v.len() != number_of_elements {
-                    return Err(NodeError::value_error(
-                        attr.clone(),
-                        &format!(
-                            "incorrect number of elements: expected {}",
-                            number_of_elements
-                        ),
-                    ));
+                    return Err(ValueErrorKind::value_error(&format!(
+                        "incorrect number of elements: expected {}",
+                        number_of_elements
+                    )))
+                    .attribute(attr);
                 }
 
                 DMatrix::from_data(VecStorage::new(
@@ -205,10 +176,8 @@ impl NodeTrait for FeConvolveMatrix {
 
         // kernel_matrix must have been specified.
         if self.kernel_matrix.is_none() {
-            return Err(NodeError::value_error(
-                QualName::new(None, ns!(svg), local_name!("kernelMatrix")),
-                "the value must be set",
-            ));
+            return Err(ValueErrorKind::value_error("the value must be set"))
+                .attribute(QualName::new(None, ns!(svg), local_name!("kernelMatrix")));
         }
 
         // Default value for the divisor.
@@ -260,24 +229,21 @@ impl FilterEffect for FeConvolveMatrix {
 
         let matrix = self.kernel_matrix.as_ref().unwrap();
 
-        let mut output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
+        let mut surface = ExclusiveImageSurface::new(
             input_surface.width(),
             input_surface.height(),
+            input.surface().surface_type(),
         )?;
 
-        let output_stride = output_surface.get_stride() as usize;
-        {
-            let mut output_data = output_surface.get_data().unwrap();
-
-            for (x, y, pixel) in Pixels::new(&input_surface, bounds) {
+        surface.modify(&mut |data, stride| {
+            for (x, y, pixel) in Pixels::within(&input_surface, bounds) {
                 // Compute the convolution rectangle bounds.
-                let kernel_bounds = IRect {
-                    x0: x as i32 - self.target_x.unwrap() as i32,
-                    y0: y as i32 - self.target_y.unwrap() as i32,
-                    x1: x as i32 - self.target_x.unwrap() as i32 + self.order.0 as i32,
-                    y1: y as i32 - self.target_y.unwrap() as i32 + self.order.1 as i32,
-                };
+                let kernel_bounds = IRect::new(
+                    x as i32 - self.target_x.unwrap() as i32,
+                    y as i32 - self.target_y.unwrap() as i32,
+                    x as i32 - self.target_x.unwrap() as i32 + self.order.0 as i32,
+                    y as i32 - self.target_y.unwrap() as i32 + self.order.1 as i32,
+                );
 
                 // Do the convolution.
                 let mut r = 0.0;
@@ -286,7 +252,7 @@ impl FilterEffect for FeConvolveMatrix {
                 let mut a = 0.0;
 
                 for (x, y, pixel) in
-                    PixelRectangle::new(&input_surface, bounds, kernel_bounds, self.edge_mode)
+                    PixelRectangle::within(&input_surface, bounds, kernel_bounds, self.edge_mode)
                 {
                     let kernel_x = (kernel_bounds.x1 - x - 1) as usize;
                     let kernel_y = (kernel_bounds.y1 - y - 1) as usize;
@@ -329,16 +295,15 @@ impl FilterEffect for FeConvolveMatrix {
                     a: ((clamped_a * 255.0) + 0.5) as u8,
                 };
 
-                output_data.set_pixel(output_stride, output_pixel, x, y);
+                data.set_pixel(stride, output_pixel, x, y);
             }
-        }
+        });
 
-        let mut output_surface =
-            SharedImageSurface::new(output_surface, input.surface().surface_type())?;
+        let mut surface = surface.share()?;
 
         if let Some((ox, oy)) = scale {
             // Scale the output surface back.
-            output_surface = output_surface.scale_to(
+            surface = surface.scale_to(
                 ctx.source_graphic().width(),
                 ctx.source_graphic().height(),
                 original_bounds,
@@ -351,10 +316,7 @@ impl FilterEffect for FeConvolveMatrix {
 
         Ok(FilterResult {
             name: self.base.result.clone(),
-            output: FilterOutput {
-                surface: output_surface,
-                bounds,
-            },
+            output: FilterOutput { surface, bounds },
         })
     }
 
@@ -364,16 +326,24 @@ impl FilterEffect for FeConvolveMatrix {
     }
 }
 
-impl EdgeMode {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "duplicate" => Ok(EdgeMode::Duplicate),
-            "wrap" => Ok(EdgeMode::Wrap),
-            "none" => Ok(EdgeMode::None),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for EdgeMode {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "duplicate" => EdgeMode::Duplicate,
+            "wrap" => EdgeMode::Wrap,
+            "none" => EdgeMode::None,
+        )?)
+    }
+}
+
+// Used for the preserveAlpha attribute
+impl Parse for bool {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "false" => false,
+            "true" => true,
+        )?)
     }
 }
