@@ -1,18 +1,18 @@
-use cairo::{self, ImageSurface};
-use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
+use cssparser::Parser;
+use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
-use crate::parsers::{self, ParseError};
+use crate::parsers::{Parse, ParseValue};
 use crate::property_bag::PropertyBag;
-use crate::surface_utils::{iterators::Pixels, shared_surface::SharedImageSurface};
+use crate::surface_utils::{iterators::Pixels, shared_surface::ExclusiveImageSurface};
 
 use super::context::{FilterContext, FilterOutput, FilterResult};
 use super::{FilterEffect, FilterError, Input, PrimitiveWithInput};
 
 /// Enumeration of the color channels the displacement map can source.
-#[derive(Debug, Clone, Copy, Eq, PartialEq, Hash)]
+#[derive(Clone, Copy)]
 enum ColorChannel {
     R,
     G,
@@ -51,13 +51,15 @@ impl NodeTrait for FeDisplacementMap {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "in2") => self.in2 = Some(Input::parse(attr, value)?),
-                expanded_name!(svg "scale") => self.scale = parsers::number(value).attribute(attr)?,
-                expanded_name!(svg "xChannelSelector") => {
-                    self.x_channel_selector = ColorChannel::parse(attr, value)?
+                expanded_name!("", "in2") => self.in2 = Some(attr.parse(value)?),
+                expanded_name!("", "scale") => self.scale = attr.parse(value)?,
+
+                expanded_name!("", "xChannelSelector") => {
+                    self.x_channel_selector = attr.parse(value)?
                 }
-                expanded_name!(svg "yChannelSelector") => {
-                    self.y_channel_selector = ColorChannel::parse(attr, value)?
+
+                expanded_name!("", "yChannelSelector") => {
+                    self.y_channel_selector = attr.parse(value)?
                 }
                 _ => (),
             }
@@ -88,16 +90,14 @@ impl FilterEffect for FeDisplacementMap {
 
         let (sx, sy) = ctx.paffine().transform_distance(self.scale, self.scale);
 
-        let output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
+        let mut surface = ExclusiveImageSurface::new(
             ctx.source_graphic().width(),
             ctx.source_graphic().height(),
+            input.surface().surface_type(),
         )?;
 
-        {
-            let cr = cairo::Context::new(&output_surface);
-
-            for (x, y, displacement_pixel) in Pixels::new(&displacement_surface, bounds) {
+        surface.draw(&mut |cr| {
+            for (x, y, displacement_pixel) in Pixels::within(&displacement_surface, bounds) {
                 let get_value = |channel| match channel {
                     ColorChannel::R => displacement_pixel.r,
                     ColorChannel::G => displacement_pixel.g,
@@ -124,12 +124,14 @@ impl FilterEffect for FeDisplacementMap {
                 input.surface().set_as_source_surface(&cr, -ox, -oy);
                 cr.paint();
             }
-        }
+
+            Ok(())
+        })?;
 
         Ok(FilterResult {
             name: self.base.result.clone(),
             output: FilterOutput {
-                surface: SharedImageSurface::new(output_surface, input.surface().surface_type())?,
+                surface: surface.share()?,
                 bounds,
             },
         })
@@ -143,17 +145,14 @@ impl FilterEffect for FeDisplacementMap {
     }
 }
 
-impl ColorChannel {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "R" => Ok(ColorChannel::R),
-            "G" => Ok(ColorChannel::G),
-            "B" => Ok(ColorChannel::B),
-            "A" => Ok(ColorChannel::A),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for ColorChannel {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "R" => ColorChannel::R,
+            "G" => ColorChannel::G,
+            "B" => ColorChannel::B,
+            "A" => ColorChannel::A,
+        )?)
     }
 }
