@@ -1,31 +1,27 @@
-use cairo::{self, ImageSurface};
-use cssparser;
-
 use crate::drawing_ctx::DrawingCtx;
 use crate::node::{CascadedValues, NodeResult, NodeTrait, RsvgNode};
 use crate::property_bag::PropertyBag;
-use crate::surface_utils::shared_surface::{SharedImageSurface, SurfaceType};
 
 use super::context::{FilterContext, FilterOutput, FilterResult};
-use super::{Filter, FilterError, Primitive};
+use super::{FilterEffect, FilterError, Primitive};
 
 /// The `feFlood` filter primitive.
-pub struct Flood {
+pub struct FeFlood {
     base: Primitive,
 }
 
-impl Default for Flood {
+impl Default for FeFlood {
     /// Constructs a new `Flood` with empty properties.
     #[inline]
-    fn default() -> Flood {
-        Flood {
+    fn default() -> FeFlood {
+        FeFlood {
             base: Primitive::new::<Self>(),
         }
     }
 }
 
-impl NodeTrait for Flood {
-    impl_node_as_filter!();
+impl NodeTrait for FeFlood {
+    impl_node_as_filter_effect!();
 
     #[inline]
     fn set_atts(&mut self, parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
@@ -33,7 +29,7 @@ impl NodeTrait for Flood {
     }
 }
 
-impl Filter for Flood {
+impl FilterEffect for FeFlood {
     fn render(
         &self,
         node: &RsvgNode,
@@ -42,12 +38,6 @@ impl Filter for Flood {
     ) -> Result<FilterResult, FilterError> {
         let bounds = self.base.get_bounds(ctx).into_irect(draw_ctx);
 
-        let output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
-            ctx.source_graphic().width(),
-            ctx.source_graphic().height(),
-        )?;
-
         let cascaded = CascadedValues::new_from_node(node);
         let values = cascaded.get();
 
@@ -55,29 +45,13 @@ impl Filter for Flood {
             cssparser::Color::CurrentColor => values.color.0,
             cssparser::Color::RGBA(rgba) => rgba,
         };
-        let opacity = (values.flood_opacity.0).0;
+        let opacity = values.flood_opacity.0;
 
-        if opacity > 0f64 {
-            let cr = cairo::Context::new(&output_surface);
-            let r = cairo::Rectangle::from(bounds);
-            cr.rectangle(r.x, r.y, r.width, r.height);
-            cr.clip();
-
-            cr.set_source_rgba(
-                f64::from(color.red) / 255f64,
-                f64::from(color.green) / 255f64,
-                f64::from(color.blue) / 255f64,
-                opacity,
-            );
-            cr.paint();
-        }
+        let surface = ctx.source_graphic().flood(bounds, color, opacity)?;
 
         Ok(FilterResult {
             name: self.base.result.clone(),
-            output: FilterOutput {
-                surface: SharedImageSurface::new(output_surface, SurfaceType::SRgb)?,
-                bounds,
-            },
+            output: FilterOutput { surface, bounds },
         })
     }
 
