@@ -5,9 +5,9 @@ use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use nalgebra::{DMatrix, Dynamic, VecStorage};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
-use crate::parsers;
+use crate::parsers::{NumberOptionalNumber, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::rect::IRect;
 use crate::surface_utils::{
@@ -48,16 +48,17 @@ impl NodeTrait for FeGaussianBlur {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "stdDeviation") => {
-                    self.std_deviation = parsers::number_optional_number(value)
-                        .attribute(attr.clone())
-                        .and_then(|(x, y)| {
-                            if x >= 0.0 && y >= 0.0 {
-                                Ok((x, y))
+                expanded_name!("", "stdDeviation") => {
+                    let NumberOptionalNumber(x, y) =
+                        attr.parse_and_validate(value, |v: NumberOptionalNumber<f64>| {
+                            if v.0 >= 0.0 && v.1 >= 0.0 {
+                                Ok(v)
                             } else {
-                                Err(NodeError::value_error(attr, "values can't be negative"))
+                                Err(ValueErrorKind::value_error("values can't be negative"))
                             }
-                        })?
+                        })?;
+
+                    self.std_deviation = (x, y);
                 }
                 _ => (),
             }
@@ -222,25 +223,21 @@ impl FilterEffect for FeGaussianBlur {
         // channels.
 
         // Horizontal convolution.
-        let horiz_result_surface = if std_x != 0.0 {
+        let horiz_result_surface = if std_x >= 2.0 {
             // The spec says for deviation >= 2.0 three box blurs can be used as an optimization.
-            if std_x >= 2.0 {
-                three_box_blurs::<Horizontal>(input.surface(), bounds, std_x)?
-            } else {
-                gaussian_blur(input.surface(), bounds, std_x, false)?
-            }
+            three_box_blurs::<Horizontal>(input.surface(), bounds, std_x)?
+        } else if std_x != 0.0 {
+            gaussian_blur(input.surface(), bounds, std_x, false)?
         } else {
             input.surface().clone()
         };
 
         // Vertical convolution.
-        let output_surface = if std_y != 0.0 {
+        let output_surface = if std_y >= 2.0 {
             // The spec says for deviation >= 2.0 three box blurs can be used as an optimization.
-            if std_y >= 2.0 {
-                three_box_blurs::<Vertical>(&horiz_result_surface, bounds, std_y)?
-            } else {
-                gaussian_blur(&horiz_result_surface, bounds, std_y, true)?
-            }
+            three_box_blurs::<Vertical>(&horiz_result_surface, bounds, std_y)?
+        } else if std_y != 0.0 {
+            gaussian_blur(&horiz_result_surface, bounds, std_y, true)?
         } else {
             horiz_result_surface
         };
