@@ -1,44 +1,28 @@
 use std::cmp::max;
 
-use cairo::{self, ImageSurface};
-use cssparser;
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use nalgebra::Vector3;
 use num_traits::identities::Zero;
 use rayon::prelude::*;
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::filters::{
     context::{FilterContext, FilterOutput, FilterResult},
     light::{
-        bottom_left_normal,
-        bottom_right_normal,
-        bottom_row_normal,
-        interior_normal,
-        left_column_normal,
-        light_source::FeDistantLight,
-        light_source::FePointLight,
-        light_source::FeSpotLight,
-        light_source::LightSource,
-        right_column_normal,
-        top_left_normal,
-        top_right_normal,
-        top_row_normal,
-        Normal,
+        bottom_left_normal, bottom_right_normal, bottom_row_normal, interior_normal,
+        left_column_normal, light_source::FeDistantLight, light_source::FePointLight,
+        light_source::FeSpotLight, light_source::LightSource, right_column_normal, top_left_normal,
+        top_right_normal, top_row_normal, Normal,
     },
-    FilterEffect,
-    FilterError,
-    PrimitiveWithInput,
+    FilterEffect, FilterError, PrimitiveWithInput,
 };
 use crate::node::{CascadedValues, NodeResult, NodeTrait, NodeType, RsvgNode};
-use crate::parsers;
+use crate::parsers::{NumberOptionalNumber, ParseValue};
 use crate::property_bag::PropertyBag;
-use crate::property_defs::ColorInterpolationFilters;
 use crate::surface_utils::{
-    shared_surface::{SharedImageSurface, SurfaceType},
-    ImageSurfaceDataExt,
-    Pixel,
+    shared_surface::{ExclusiveImageSurface, SurfaceType},
+    ImageSurfaceDataExt, Pixel,
 };
 use crate::util::clamp;
 
@@ -68,24 +52,21 @@ impl Common {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "surfaceScale") => {
-                    self.surface_scale = parsers::number(value).attribute(attr)?
-                }
-                expanded_name!(svg "kernelUnitLength") => {
-                    self.kernel_unit_length = Some(
-                        parsers::number_optional_number(value)
-                            .attribute(attr.clone())
-                            .and_then(|(x, y)| {
-                                if x > 0.0 && y > 0.0 {
-                                    Ok((x, y))
-                                } else {
-                                    Err(NodeError::value_error(
-                                        attr,
-                                        "kernelUnitLength can't be less or equal to zero",
-                                    ))
-                                }
-                            })?,
-                    )
+                expanded_name!("", "surfaceScale") => self.surface_scale = attr.parse(value)?,
+
+                expanded_name!("", "kernelUnitLength") => {
+                    let NumberOptionalNumber(x, y) =
+                        attr.parse_and_validate(value, |v: NumberOptionalNumber<f64>| {
+                            if v.0 > 0.0 && v.1 > 0.0 {
+                                Ok(v)
+                            } else {
+                                Err(ValueErrorKind::value_error(
+                                    "kernelUnitLength can't be less or equal to zero",
+                                ))
+                            }
+                        })?;
+
+                    self.kernel_unit_length = Some((x, y));
                 }
                 _ => (),
             }
@@ -118,19 +99,16 @@ impl NodeTrait for FeDiffuseLighting {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "diffuseConstant") => {
-                    self.diffuse_constant = parsers::number(value)
-                        .attribute(attr.clone())
-                        .and_then(|x| {
-                            if x >= 0.0 {
-                                Ok(x)
-                            } else {
-                                Err(NodeError::value_error(
-                                    attr,
-                                    "diffuseConstant can't be negative",
-                                ))
-                            }
-                        })?;
+                expanded_name!("", "diffuseConstant") => {
+                    self.diffuse_constant = attr.parse_and_validate(value, |x| {
+                        if x >= 0.0 {
+                            Ok(x)
+                        } else {
+                            Err(ValueErrorKind::value_error(
+                                "diffuseConstant can't be negative",
+                            ))
+                        }
+                    })?;
                 }
                 _ => (),
             }
@@ -190,33 +168,27 @@ impl NodeTrait for FeSpecularLighting {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "specularConstant") => {
-                    self.specular_constant = parsers::number(value)
-                        .attribute(attr.clone())
-                        .and_then(|x| {
-                            if x >= 0.0 {
-                                Ok(x)
-                            } else {
-                                Err(NodeError::value_error(
-                                    attr,
-                                    "specularConstant can't be negative",
-                                ))
-                            }
-                        })?;
+                expanded_name!("", "specularConstant") => {
+                    self.specular_constant = attr.parse_and_validate(value, |x| {
+                        if x >= 0.0 {
+                            Ok(x)
+                        } else {
+                            Err(ValueErrorKind::value_error(
+                                "specularConstant can't be negative",
+                            ))
+                        }
+                    })?;
                 }
-                expanded_name!(svg "specularExponent") => {
-                    self.specular_exponent = parsers::number(value)
-                        .attribute(attr.clone())
-                        .and_then(|x| {
-                            if x >= 1.0 && x <= 128.0 {
-                                Ok(x)
-                            } else {
-                                Err(NodeError::value_error(
-                                    attr,
-                                    "specularExponent should be between 1.0 and 128.0",
-                                ))
-                            }
-                        })?;
+                expanded_name!("", "specularExponent") => {
+                    self.specular_exponent = attr.parse_and_validate(value, |x| {
+                        if x >= 1.0 && x <= 128.0 {
+                            Ok(x)
+                        } else {
+                            Err(ValueErrorKind::value_error(
+                                "specularExponent should be between 1.0 and 128.0",
+                            ))
+                        }
+                    })?;
                 }
                 _ => (),
             }
@@ -305,29 +277,38 @@ macro_rules! impl_lighting_filter {
 
                 if let Some((ox, oy)) = scale {
                     // Scale the input surface to match kernel_unit_length.
-                    let (new_surface, new_bounds) = input_surface.scale(bounds, 1.0 / ox, 1.0 / oy)?;
+                    let (new_surface, new_bounds) =
+                        input_surface.scale(bounds, 1.0 / ox, 1.0 / oy)?;
 
                     input_surface = new_surface;
                     bounds = new_bounds;
                 }
 
-                // Check if the surface is too small for normal computation. This case is unspecified;
-                // WebKit doesn't render anything in this case.
-                if bounds.x1 < bounds.x0 + 2 || bounds.y1 < bounds.y0 + 2 {
+                let (bounds_w, bounds_h) = bounds.size();
+
+                // Check if the surface is too small for normal computation. This case is
+                // unspecified; WebKit doesn't render anything in this case.
+                if bounds_w < 2 || bounds_h < 2 {
                     return Err(FilterError::LightingInputTooSmall);
                 }
 
                 let (ox, oy) = scale.unwrap_or((1.0, 1.0));
 
-                let mut output_surface = ImageSurface::create(
-                    cairo::Format::ARgb32,
+                let cascaded = CascadedValues::new_from_node(node);
+                let values = cascaded.get();
+                // The generated color values are in the color space determined by
+                // color-interpolation-filters.
+                let surface_type = SurfaceType::from(values.color_interpolation_filters);
+
+                let mut surface = ExclusiveImageSurface::new(
                     input_surface.width(),
                     input_surface.height(),
+                    surface_type,
                 )?;
 
-                let output_stride = output_surface.get_stride() as usize;
                 {
-                    let mut output_data = output_surface.get_data().unwrap();
+                    let output_stride = surface.stride() as usize;
+                    let mut output_data = surface.get_data();
                     let output_slice = &mut *output_data;
 
                     let compute_output_pixel =
@@ -342,19 +323,15 @@ macro_rules! impl_lighting_filter {
 
                             // compute the factor just once for the three colors
                             let factor = self.compute_factor(normal, light_vector);
-                            let compute = |x| (clamp(factor * f64::from(x), 0.0, 255.0) + 0.5) as u8;
+                            let compute =
+                                |x| (clamp(factor * f64::from(x), 0.0, 255.0) + 0.5) as u8;
 
                             let r = compute(light_color.red);
                             let g = compute(light_color.green);
                             let b = compute(light_color.blue);
                             let a = $alpha_func(r, g, b);
 
-                            let output_pixel = Pixel {
-                                r,
-                                g,
-                                b,
-                                a,
-                            };
+                            let output_pixel = Pixel { r, g, b, a };
 
                             output_slice.set_pixel(output_stride, output_pixel, x, y - base_y);
                         };
@@ -395,7 +372,7 @@ macro_rules! impl_lighting_filter {
                         bottom_right_normal(&input_surface, bounds),
                     );
 
-                    if bounds.x1 - bounds.x0 >= 3 {
+                    if bounds_w >= 3 {
                         // Top row.
                         for x in bounds.x0 as u32 + 1..bounds.x1 as u32 - 1 {
                             compute_output_pixel(
@@ -419,7 +396,7 @@ macro_rules! impl_lighting_filter {
                         }
                     }
 
-                    if bounds.y1 - bounds.y0 >= 3 {
+                    if bounds_h >= 3 {
                         // Left column.
                         for y in bounds.y0 as u32 + 1..bounds.y1 as u32 - 1 {
                             compute_output_pixel(
@@ -443,7 +420,7 @@ macro_rules! impl_lighting_filter {
                         }
                     }
 
-                    if bounds.x1 - bounds.x0 >= 3 && bounds.y1 - bounds.y0 >= 3 {
+                    if bounds_w >= 3 && bounds_h >= 3 {
                         // Interior pixels.
                         let first_row = bounds.y0 as u32 + 1;
                         let one_past_last_row = bounds.y1 as u32 - 1;
@@ -467,21 +444,11 @@ macro_rules! impl_lighting_filter {
                     }
                 }
 
-                let cascaded = CascadedValues::new_from_node(node);
-                let values = cascaded.get();
-                // The generated color values are in the color space determined by
-                // color-interpolation-filters.
-                let surface_type =
-                    if values.color_interpolation_filters == ColorInterpolationFilters::LinearRgb {
-                        SurfaceType::LinearRgb
-                    } else {
-                        SurfaceType::SRgb
-                    };
-                let mut output_surface = SharedImageSurface::new(output_surface, surface_type)?;
+                let mut surface = surface.share()?;
 
                 if let Some((ox, oy)) = scale {
                     // Scale the output surface back.
-                    output_surface = output_surface.scale_to(
+                    surface = surface.scale_to(
                         ctx.source_graphic().width(),
                         ctx.source_graphic().height(),
                         original_bounds,
@@ -494,10 +461,7 @@ macro_rules! impl_lighting_filter {
 
                 Ok(FilterResult {
                     name: self.common().base.result.clone(),
-                    output: FilterOutput {
-                        surface: output_surface,
-                        bounds,
-                    },
+                    output: FilterOutput { surface, bounds },
                 })
             }
 
@@ -506,7 +470,7 @@ macro_rules! impl_lighting_filter {
                 true
             }
         }
-    }
+    };
 }
 
 const fn diffuse_alpha(_r: u8, _g: u8, _b: u8) -> u8 {
