@@ -1,8 +1,8 @@
-use cairo::{self, ImageSurface};
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::drawing_ctx::DrawingCtx;
 use crate::node::{NodeResult, NodeTrait, NodeType, RsvgNode};
+use crate::parsers::ParseValue;
 use crate::property_bag::PropertyBag;
 use crate::rect::IRect;
 use crate::surface_utils::shared_surface::{SharedImageSurface, SurfaceType};
@@ -46,7 +46,7 @@ impl NodeTrait for FeMergeNode {
     fn set_atts(&mut self, _parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "in") => self.in_ = Some(Input::parse(attr, value)?),
+                expanded_name!("", "in") => self.in_ = Some(attr.parse(value)?),
                 _ => (),
             }
         }
@@ -68,39 +68,11 @@ impl FeMergeNode {
         if output_surface.is_none() {
             return Ok(input.surface().clone());
         }
-        let output_surface = output_surface.unwrap();
 
-        // If we're combining two alpha-only surfaces, the result is alpha-only. Otherwise the
-        // result is whatever the non-alpha-only type we're working on (which can be either sRGB or
-        // linear sRGB depending on color-interpolation-filters).
-        let surface_type = if input.surface().is_alpha_only() {
-            output_surface.surface_type()
-        } else {
-            if !output_surface.is_alpha_only() {
-                // All surface types should match (this is enforced by get_input()).
-                assert_eq!(
-                    output_surface.surface_type(),
-                    input.surface().surface_type()
-                );
-            }
-
-            input.surface().surface_type()
-        };
-
-        let output_surface = output_surface.into_image_surface()?;
-
-        {
-            let cr = cairo::Context::new(&output_surface);
-            let r = cairo::Rectangle::from(bounds);
-            cr.rectangle(r.x, r.y, r.width, r.height);
-            cr.clip();
-
-            input.surface().set_as_source_surface(&cr, 0f64, 0f64);
-            cr.set_operator(cairo::Operator::Over);
-            cr.paint();
-        }
-
-        Ok(SharedImageSurface::new(output_surface, surface_type)?)
+        input
+            .surface()
+            .compose(&output_surface.unwrap(), bounds, cairo::Operator::Over)
+            .map_err(FilterError::CairoError)
     }
 }
 
@@ -143,24 +115,18 @@ impl FilterEffect for FeMerge {
             )?);
         }
 
-        let output_surface = match output_surface {
-            Some(surface) => surface,
-            None => SharedImageSurface::new(
-                ImageSurface::create(
-                    cairo::Format::ARgb32,
-                    ctx.source_graphic().width(),
-                    ctx.source_graphic().height(),
-                )?,
+        let surface = match output_surface {
+            Some(s) => s,
+            None => SharedImageSurface::empty(
+                ctx.source_graphic().width(),
+                ctx.source_graphic().height(),
                 SurfaceType::AlphaOnly,
             )?,
         };
 
         Ok(FilterResult {
             name: self.base.result.clone(),
-            output: FilterOutput {
-                surface: output_surface,
-                bounds,
-            },
+            output: FilterOutput { surface, bounds },
         })
     }
 
