@@ -1,7 +1,8 @@
+//! Entry point for the CSS filters infrastructure.
+
 use std::ops::Deref;
 use std::time::Instant;
 
-use cairo;
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::bbox::BoundingBox;
@@ -11,7 +12,7 @@ use crate::error::{RenderingError, ValueErrorKind};
 use crate::filter::Filter;
 use crate::length::*;
 use crate::node::{CascadedValues, NodeResult, NodeTrait, NodeType, RsvgNode};
-use crate::parsers::{ParseError, ParseValue};
+use crate::parsers::ParseValue;
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
 use crate::property_defs::ColorInterpolationFilters;
@@ -27,7 +28,7 @@ mod error;
 use self::error::FilterError;
 
 mod input;
-use self::input::Input;
+use self::input::{CustomIdent, Input};
 
 /// A filter primitive interface.
 pub trait FilterEffect: NodeTrait {
@@ -80,7 +81,7 @@ struct Primitive {
     y: Option<Length<Vertical>>,
     width: Option<Length<Horizontal>>,
     height: Option<Length<Vertical>>,
-    result: Option<String>,
+    result: Option<CustomIdent>,
 }
 
 /// The base node for filter primitives which accept input.
@@ -131,9 +132,9 @@ impl NodeTrait for Primitive {
 
             match length.unit {
                 LengthUnit::Px | LengthUnit::Percent => Ok(length),
-                _ => Err(ValueErrorKind::Parse(ParseError::new(
+                _ => Err(ValueErrorKind::parse_error(
                     "unit identifiers are not allowed with primitiveUnits set to objectBoundingBox",
-                ))),
+                )),
             }
         };
 
@@ -144,9 +145,9 @@ impl NodeTrait for Primitive {
 
             match length.unit {
                 LengthUnit::Px | LengthUnit::Percent => Ok(length),
-                _ => Err(ValueErrorKind::Parse(ParseError::new(
+                _ => Err(ValueErrorKind::parse_error(
                     "unit identifiers are not allowed with primitiveUnits set to objectBoundingBox",
-                ))),
+                )),
             }
         };
 
@@ -160,27 +161,27 @@ impl NodeTrait for Primitive {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "x") => {
+                expanded_name!("", "x") => {
                     self.x = Some(attr.parse_and_validate(value, check_units_horizontal)?)
                 }
-                expanded_name!(svg "y") => {
+                expanded_name!("", "y") => {
                     self.y = Some(attr.parse_and_validate(value, check_units_vertical)?)
                 }
-                expanded_name!(svg "width") => {
+                expanded_name!("", "width") => {
                     self.width =
                         Some(attr.parse_and_validate(
                             value,
                             check_units_horizontal_and_ensure_nonnegative,
                         )?)
                 }
-                expanded_name!(svg "height") => {
+                expanded_name!("", "height") => {
                     self.height =
                         Some(attr.parse_and_validate(
                             value,
                             check_units_vertical_and_ensure_nonnegative,
                         )?)
                 }
-                expanded_name!(svg "result") => self.result = Some(value.to_string()),
+                expanded_name!("", "result") => self.result = Some(attr.parse(value)?),
                 _ => (),
             }
         }
@@ -216,7 +217,7 @@ impl NodeTrait for PrimitiveWithInput {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "in") => drop(self.in_ = Some(Input::parse(attr, value)?)),
+                expanded_name!("", "in") => self.in_ = Some(attr.parse(value)?),
                 _ => (),
             }
         }
@@ -234,40 +235,18 @@ impl Deref for PrimitiveWithInput {
     }
 }
 
-/// Creates a `SharedImageSurface` from an `ImageSurface`, even if the former
-/// does not have a reference count of 1.
-fn copy_to_shared_surface(
-    surface: &cairo::ImageSurface,
-) -> Result<SharedImageSurface, cairo::Status> {
-    let copy = cairo::ImageSurface::create(
-        cairo::Format::ARgb32,
-        surface.get_width(),
-        surface.get_height(),
-    )?;
-    {
-        let cr = cairo::Context::new(&copy);
-        cr.set_source_surface(surface, 0f64, 0f64);
-        cr.paint();
-    }
-    SharedImageSurface::new(copy, SurfaceType::SRgb)
-}
-
 /// Applies a filter and returns the resulting surface.
 pub fn render(
     filter_node: &RsvgNode,
     computed_from_node_being_filtered: &ComputedValues,
-    source: &cairo::ImageSurface,
+    source_surface: SharedImageSurface,
     draw_ctx: &mut DrawingCtx,
     node_bbox: BoundingBox,
-) -> Result<cairo::ImageSurface, RenderingError> {
+) -> Result<SharedImageSurface, RenderingError> {
     let filter_node = &*filter_node;
     assert_eq!(filter_node.borrow().get_type(), NodeType::Filter);
     assert!(!filter_node.borrow().is_in_error());
 
-    // The source surface has multiple references. We need to copy it to a new surface to have a
-    // unique reference to be able to safely access the pixel data.
-    let source_surface = copy_to_shared_surface(source)?;
-
     let mut filter_ctx = FilterContext::new(
         filter_node,
         computed_from_node_being_filtered,
@@ -279,7 +258,7 @@ pub fn render(
     // If paffine is non-invertible, we won't draw anything. Also bbox combining in bounds
     // computations will panic due to non-invertible martrix.
     if filter_ctx.paffine().try_invert().is_err() {
-        return Ok(filter_ctx.into_output()?.into_image_surface()?);
+        return Ok(filter_ctx.into_output()?);
     }
 
     let primitives = filter_node
@@ -344,5 +323,14 @@ pub fn render(
         );
     }
 
-    Ok(filter_ctx.into_output()?.into_image_surface()?)
+    Ok(filter_ctx.into_output()?)
+}
+
+impl From<ColorInterpolationFilters> for SurfaceType {
+    fn from(c: ColorInterpolationFilters) -> Self {
+        match c {
+            ColorInterpolationFilters::LinearRgb => SurfaceType::LinearRgb,
+            _ => SurfaceType::SRgb,
+        }
+    }
 }
