@@ -1,27 +1,24 @@
 use std::cmp::{max, min};
 
-use cairo::{self, ImageSurface};
-use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
+use cssparser::Parser;
+use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{NodeResult, NodeTrait, RsvgNode};
-use crate::parsers::{self, ParseError};
+use crate::parsers::{NumberOptionalNumber, Parse, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::rect::IRect;
 use crate::surface_utils::{
     iterators::{PixelRectangle, Pixels},
-    shared_surface::SharedImageSurface,
-    EdgeMode,
-    ImageSurfaceDataExt,
-    Pixel,
+    shared_surface::ExclusiveImageSurface,
+    EdgeMode, ImageSurfaceDataExt, Pixel,
 };
 
 use super::context::{FilterContext, FilterOutput, FilterResult};
 use super::{FilterEffect, FilterError, PrimitiveWithInput};
 
 /// Enumeration of the possible morphology operations.
-#[derive(Debug, Clone, Copy, Eq, PartialEq, Hash)]
 enum Operator {
     Erode,
     Dilate,
@@ -54,17 +51,18 @@ impl NodeTrait for FeMorphology {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "operator") => self.operator = Operator::parse(attr, value)?,
-                expanded_name!(svg "radius") => {
-                    self.radius = parsers::number_optional_number(value)
-                        .attribute(attr.clone())
-                        .and_then(|(x, y)| {
-                            if x >= 0.0 && y >= 0.0 {
-                                Ok((x, y))
+                expanded_name!("", "operator") => self.operator = attr.parse(value)?,
+                expanded_name!("", "radius") => {
+                    let NumberOptionalNumber(x, y) =
+                        attr.parse_and_validate(value, |v: NumberOptionalNumber<f64>| {
+                            if v.0 >= 0.0 && v.1 >= 0.0 {
+                                Ok(v)
                             } else {
-                                Err(NodeError::value_error(attr, "radius cannot be negative"))
+                                Err(ValueErrorKind::value_error("radius cannot be negative"))
                             }
-                        })?
+                        })?;
+
+                    self.radius = (x, y);
                 }
                 _ => (),
             }
@@ -92,27 +90,23 @@ impl FilterEffect for FeMorphology {
         let (rx, ry) = ctx.paffine().transform_distance(rx, ry);
 
         // The radii can become negative here due to the transform.
-        let rx = rx.abs();
-        let ry = ry.abs();
+        let (rx, ry) = (rx.abs(), ry.abs());
 
-        let mut output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
+        let mut surface = ExclusiveImageSurface::new(
             ctx.source_graphic().width(),
             ctx.source_graphic().height(),
+            input.surface().surface_type(),
         )?;
 
-        let output_stride = output_surface.get_stride() as usize;
-        {
-            let mut output_data = output_surface.get_data().unwrap();
-
-            for (x, y, _pixel) in Pixels::new(input.surface(), bounds) {
+        surface.modify(&mut |data, stride| {
+            for (x, y, _pixel) in Pixels::within(input.surface(), bounds) {
                 // Compute the kernel rectangle bounds.
-                let kernel_bounds = IRect {
-                    x0: (f64::from(x) - rx).floor() as i32,
-                    y0: (f64::from(y) - ry).floor() as i32,
-                    x1: (f64::from(x) + rx).ceil() as i32 + 1,
-                    y1: (f64::from(y) + ry).ceil() as i32 + 1,
-                };
+                let kernel_bounds = IRect::new(
+                    (f64::from(x) - rx).floor() as i32,
+                    (f64::from(y) - ry).floor() as i32,
+                    (f64::from(x) + rx).ceil() as i32 + 1,
+                    (f64::from(y) + ry).ceil() as i32 + 1,
+                );
 
                 // Compute the new pixel values.
                 let initial = match self.operator {
@@ -128,7 +122,7 @@ impl FilterEffect for FeMorphology {
                 };
 
                 for (_x, _y, pixel) in
-                    PixelRectangle::new(&input.surface(), bounds, kernel_bounds, EdgeMode::None)
+                    PixelRectangle::within(&input.surface(), bounds, kernel_bounds, EdgeMode::None)
                 {
                     let op = match self.operator {
                         Operator::Erode => min,
@@ -141,14 +135,14 @@ impl FilterEffect for FeMorphology {
                     output_pixel.a = op(output_pixel.a, pixel.a);
                 }
 
-                output_data.set_pixel(output_stride, output_pixel, x, y);
+                data.set_pixel(stride, output_pixel, x, y);
             }
-        }
+        });
 
         Ok(FilterResult {
             name: self.base.result.clone(),
             output: FilterOutput {
-                surface: SharedImageSurface::new(output_surface, input.surface().surface_type())?,
+                surface: surface.share()?,
                 bounds,
             },
         })
@@ -160,15 +154,12 @@ impl FilterEffect for FeMorphology {
     }
 }
 
-impl Operator {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "erode" => Ok(Operator::Erode),
-            "dilate" => Ok(Operator::Dilate),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for Operator {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "erode" => Operator::Erode,
+            "dilate" => Operator::Dilate,
+        )?)
     }
 }
