@@ -1,16 +1,14 @@
-use cairo::{self, ImageSurface};
-use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
+use cssparser::Parser;
+use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, NodeError};
+use crate::error::*;
 use crate::node::{CascadedValues, NodeResult, NodeTrait, RsvgNode};
-use crate::parsers::{self, ParseError};
+use crate::parsers::{NumberOptionalNumber, Parse, ParseValue};
 use crate::property_bag::PropertyBag;
-use crate::property_defs::ColorInterpolationFilters;
 use crate::surface_utils::{
-    shared_surface::{SharedImageSurface, SurfaceType},
-    ImageSurfaceDataExt,
-    Pixel,
+    shared_surface::{ExclusiveImageSurface, SurfaceType},
+    ImageSurfaceDataExt, Pixel,
 };
 use crate::util::clamp;
 
@@ -65,34 +63,32 @@ impl NodeTrait for FeTurbulence {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "baseFrequency") => {
-                    self.base_frequency = parsers::number_optional_number(value)
-                        .attribute(attr.clone())
-                        .and_then(|(x, y)| {
-                            if x >= 0.0 && y >= 0.0 {
-                                Ok((x, y))
+                expanded_name!("", "baseFrequency") => {
+                    let NumberOptionalNumber(x, y) =
+                        attr.parse_and_validate(value, |v: NumberOptionalNumber<f64>| {
+                            if v.0 >= 0.0 && v.1 >= 0.0 {
+                                Ok(v)
                             } else {
-                                Err(NodeError::value_error(attr, "values can't be negative"))
+                                Err(ValueErrorKind::value_error("values can't be negative"))
                             }
-                        })?
+                        })?;
+
+                    self.base_frequency = (x, y);
                 }
-                expanded_name!(svg "numOctaves") => {
-                    self.num_octaves = parsers::integer(value).attribute(attr)?
+                expanded_name!("", "numOctaves") => {
+                    self.num_octaves = attr.parse(value)?;
                 }
                 // Yes, seed needs to be parsed as a number and then truncated.
-                expanded_name!(svg "seed") => {
-                    self.seed = parsers::number(value)
-                        .map(|x| {
-                            clamp(
-                                x.trunc(),
-                                f64::from(i32::min_value()),
-                                f64::from(i32::max_value()),
-                            ) as i32
-                        })
-                        .attribute(attr)?
+                expanded_name!("", "seed") => {
+                    let v: f64 = attr.parse(value)?;
+                    self.seed = clamp(
+                        v.trunc(),
+                        f64::from(i32::min_value()),
+                        f64::from(i32::max_value()),
+                    ) as i32;
                 }
-                expanded_name!(svg "stitchTiles") => self.stitch_tiles = StitchTiles::parse(attr, value)?,
-                expanded_name!(svg "type") => self.type_ = NoiseType::parse(attr, value)?,
+                expanded_name!("", "stitchTiles") => self.stitch_tiles = attr.parse(value)?,
+                expanded_name!("", "type") => self.type_ = attr.parse(value)?,
                 _ => (),
             }
         }
@@ -352,22 +348,25 @@ impl FilterEffect for FeTurbulence {
             self.num_octaves,
             self.type_,
             self.stitch_tiles,
-            f64::from(bounds.x1 - bounds.x0),
-            f64::from(bounds.y1 - bounds.y0),
+            f64::from(bounds.width()),
+            f64::from(bounds.height()),
         );
 
-        let mut output_surface = ImageSurface::create(
-            cairo::Format::ARgb32,
+        let cascaded = CascadedValues::new_from_node(node);
+        let values = cascaded.get();
+        // The generated color values are in the color space determined by
+        // color-interpolation-filters.
+        let surface_type = SurfaceType::from(values.color_interpolation_filters);
+
+        let mut surface = ExclusiveImageSurface::new(
             ctx.source_graphic().width(),
             ctx.source_graphic().height(),
+            surface_type,
         )?;
 
-        let output_stride = output_surface.get_stride() as usize;
-        {
-            let mut output_data = output_surface.get_data().unwrap();
-
-            for y in bounds.y0..bounds.y1 {
-                for x in bounds.x0..bounds.x1 {
+        surface.modify(&mut |data, stride| {
+            for y in bounds.y_range() {
+                for x in bounds.x_range() {
                     let point = affine.transform_point(f64::from(x), f64::from(y));
                     let point = [point.0, point.1];
 
@@ -395,26 +394,15 @@ impl FilterEffect for FeTurbulence {
                     }
                     .premultiply();
 
-                    output_data.set_pixel(output_stride, pixel, x as u32, y as u32);
+                    data.set_pixel(stride, pixel, x as u32, y as u32);
                 }
             }
-        }
-
-        let cascaded = CascadedValues::new_from_node(node);
-        let values = cascaded.get();
-        // The generated color values are in the color space determined by
-        // color-interpolation-filters.
-        let surface_type =
-            if values.color_interpolation_filters == ColorInterpolationFilters::LinearRgb {
-                SurfaceType::LinearRgb
-            } else {
-                SurfaceType::SRgb
-            };
+        });
 
         Ok(FilterResult {
             name: self.base.result.clone(),
             output: FilterOutput {
-                surface: SharedImageSurface::new(output_surface, surface_type)?,
+                surface: surface.share()?,
                 bounds,
             },
         })
@@ -426,29 +414,23 @@ impl FilterEffect for FeTurbulence {
     }
 }
 
-impl StitchTiles {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "stitch" => Ok(StitchTiles::Stitch),
-            "noStitch" => Ok(StitchTiles::NoStitch),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for StitchTiles {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "stitch" => StitchTiles::Stitch,
+            "noStitch" => StitchTiles::NoStitch,
+        )?)
     }
 }
 
-impl NoiseType {
-    fn parse(attr: QualName, s: &str) -> Result<Self, NodeError> {
-        match s {
-            "fractalNoise" => Ok(NoiseType::FractalNoise),
-            "turbulence" => Ok(NoiseType::Turbulence),
-            _ => Err(NodeError::parse_error(
-                attr,
-                ParseError::new("invalid value"),
-            )),
-        }
+impl Parse for NoiseType {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Self, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "fractalNoise" => NoiseType::FractalNoise,
+            "turbulence" => NoiseType::Turbulence,
+        )?)
     }
 }
 
