@@ -1,5 +1,6 @@
-use cairo;
-use cssparser::{self, CowRcStr, Parser, Token};
+//! Gradient paint servers; the `linearGradient` and `radialGradient` elements.
+
+use cssparser::Parser;
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use std::cell::RefCell;
 
@@ -11,11 +12,10 @@ use crate::error::*;
 use crate::length::*;
 use crate::node::{CascadedValues, NodeResult, NodeTrait, NodeType, RsvgNode};
 use crate::paint_server::{AsPaintSource, PaintSource};
-use crate::parsers::{Parse, ParseError, ParseValue};
+use crate::parsers::{Parse, ParseValue};
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
 use crate::property_defs::StopColor;
-use crate::rect::RectangleExt;
 use crate::unit_interval::UnitInterval;
 
 /// Contents of a <stop> element for gradient color stops
@@ -43,26 +43,13 @@ enum SpreadMethod {
 }
 
 impl Parse for SpreadMethod {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<SpreadMethod, ValueErrorKind> {
-        let loc = parser.current_source_location();
-
-        parser
-            .expect_ident()
-            .and_then(|cow| match cow.as_ref() {
-                "pad" => Ok(SpreadMethod::Pad),
-                "reflect" => Ok(SpreadMethod::Reflect),
-                "repeat" => Ok(SpreadMethod::Repeat),
-                _ => Err(
-                    loc.new_basic_unexpected_token_error(Token::Ident(CowRcStr::from(
-                        cow.as_ref().to_string(),
-                    ))),
-                ),
-            })
-            .map_err(|_| {
-                ValueErrorKind::Parse(ParseError::new("expected 'pad' | 'reflect' | 'repeat'"))
-            })
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<SpreadMethod, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "pad" => SpreadMethod::Pad,
+            "reflect" => SpreadMethod::Reflect,
+            "repeat" => SpreadMethod::Repeat,
+        )?)
     }
 }
 
@@ -126,9 +113,8 @@ fn fix_focus_point(fx: f64, fy: f64, cx: f64, cy: f64, radius: f64) -> (f64, f64
 pub struct Stop {
     /// <stop offset="..."/>
     offset: UnitInterval,
-
-    // stop-color and stop-opacity are not attributes; they are properties, so
-    // they go into property_defs.rs
+    /* stop-color and stop-opacity are not attributes; they are properties, so
+     * they go into property_defs.rs */
 }
 
 fn validate_offset(length: Length<Both>) -> Result<Length<Both>, ValueErrorKind> {
@@ -144,7 +130,7 @@ impl NodeTrait for Stop {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "offset") => {
+                expanded_name!("", "offset") => {
                     self.offset = attr
                         .parse_and_validate(value, validate_offset)
                         .map(|l| UnitInterval::clamp(l.length))?
@@ -234,16 +220,31 @@ impl UnresolvedVariant {
 
     fn resolve_from_fallback(&self, fallback: &UnresolvedVariant) -> UnresolvedVariant {
         match (*self, *fallback) {
-            (UnresolvedVariant::Linear { x1, y1, x2, y2 },
-             UnresolvedVariant::Linear { x1: fx1, y1: fy1, x2: fx2, y2: fy2 }) => UnresolvedVariant::Linear {
+            (
+                UnresolvedVariant::Linear { x1, y1, x2, y2 },
+                UnresolvedVariant::Linear {
+                    x1: fx1,
+                    y1: fy1,
+                    x2: fx2,
+                    y2: fy2,
+                },
+            ) => UnresolvedVariant::Linear {
                 x1: x1.or(fx1),
                 y1: y1.or(fy1),
                 x2: x2.or(fx2),
                 y2: y2.or(fy2),
             },
 
-            (UnresolvedVariant::Radial { cx, cy, r, fx, fy },
-             UnresolvedVariant::Radial { cx: fcx, cy: fcy, r: fr, fx: ffx, fy: ffy }) => UnresolvedVariant::Radial {
+            (
+                UnresolvedVariant::Radial { cx, cy, r, fx, fy },
+                UnresolvedVariant::Radial {
+                    cx: fcx,
+                    cy: fcy,
+                    r: fr,
+                    fx: ffx,
+                    fy: ffy,
+                },
+            ) => UnresolvedVariant::Radial {
                 cx: cx.or(fcx),
                 cy: cy.or(fcy),
                 r: r.or(fr),
@@ -276,7 +277,7 @@ impl UnresolvedVariant {
                 let fy = fy.or(cy);
 
                 UnresolvedVariant::Radial { cx, cy, r, fx, fy }
-            },
+            }
         }
     }
 }
@@ -484,7 +485,13 @@ impl UnresolvedGradient {
         let stops = self.stops.clone().or_else(|| fallback.stops.clone());
         let variant = self.variant.resolve_from_fallback(&fallback.variant);
 
-        UnresolvedGradient { units, affine, spread, stops, variant }
+        UnresolvedGradient {
+            units,
+            affine,
+            spread,
+            stops,
+            variant,
+        }
     }
 
     fn resolve_from_defaults(&self) -> UnresolvedGradient {
@@ -494,7 +501,13 @@ impl UnresolvedGradient {
         let stops = self.stops.clone().or_else(|| Some(Vec::<ColorStop>::new()));
         let variant = self.variant.resolve_from_defaults();
 
-        UnresolvedGradient { units, affine, spread, stops, variant }
+        UnresolvedGradient {
+            units,
+            affine,
+            spread,
+            stops,
+            variant,
+        }
     }
 }
 
@@ -551,7 +564,7 @@ macro_rules! impl_get_unresolved {
                 }
             }
         }
-    }
+    };
 }
 impl_get_unresolved!(LinearGradient);
 impl_get_unresolved!(RadialGradient);
@@ -560,9 +573,9 @@ impl Common {
     fn set_atts(&mut self, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "gradientUnits") => self.units = Some(attr.parse(value)?),
-                expanded_name!(svg "gradientTransform") => self.affine = Some(attr.parse(value)?),
-                expanded_name!(svg "spreadMethod") => self.spread = Some(attr.parse(value)?),
+                expanded_name!("", "gradientUnits") => self.units = Some(attr.parse(value)?),
+                expanded_name!("", "gradientTransform") => self.affine = Some(attr.parse(value)?),
+                expanded_name!("", "spreadMethod") => self.spread = Some(attr.parse(value)?),
                 expanded_name!(xlink "href") => {
                     self.fallback = Some(Fragment::parse(value).attribute(attr)?)
                 }
@@ -580,10 +593,10 @@ impl NodeTrait for LinearGradient {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "x1") => self.x1 = Some(attr.parse(value)?),
-                expanded_name!(svg "y1") => self.y1 = Some(attr.parse(value)?),
-                expanded_name!(svg "x2") => self.x2 = Some(attr.parse(value)?),
-                expanded_name!(svg "y2") => self.y2 = Some(attr.parse(value)?),
+                expanded_name!("", "x1") => self.x1 = Some(attr.parse(value)?),
+                expanded_name!("", "y1") => self.y1 = Some(attr.parse(value)?),
+                expanded_name!("", "x2") => self.x2 = Some(attr.parse(value)?),
+                expanded_name!("", "y2") => self.y2 = Some(attr.parse(value)?),
 
                 _ => (),
             }
@@ -599,11 +612,11 @@ impl NodeTrait for RadialGradient {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "cx") => self.cx = Some(attr.parse(value)?),
-                expanded_name!(svg "cy") => self.cy = Some(attr.parse(value)?),
-                expanded_name!(svg "r") => self.r = Some(attr.parse(value)?),
-                expanded_name!(svg "fx") => self.fx = Some(attr.parse(value)?),
-                expanded_name!(svg "fy") => self.fy = Some(attr.parse(value)?),
+                expanded_name!("", "cx") => self.cx = Some(attr.parse(value)?),
+                expanded_name!("", "cy") => self.cy = Some(attr.parse(value)?),
+                expanded_name!("", "r") => self.r = Some(attr.parse(value)?),
+                expanded_name!("", "fx") => self.fx = Some(attr.parse(value)?),
+                expanded_name!("", "fy") => self.fy = Some(attr.parse(value)?),
 
                 _ => (),
             }
@@ -628,7 +641,10 @@ macro_rules! impl_paint_source {
                     return Ok(gradient.clone());
                 }
 
-                let Unresolved { mut gradient, mut fallback } = self.get_unresolved(node);
+                let Unresolved {
+                    mut gradient,
+                    mut fallback,
+                } = self.get_unresolved(node);
 
                 let mut stack = NodeStack::new();
 
@@ -653,7 +669,7 @@ macro_rules! impl_paint_source {
                                 a_gradient.get_unresolved(&acquired_node)
                             }
 
-                            _ => unreachable!()
+                            _ => unreachable!(),
                         };
 
                         gradient = gradient.resolve_from_fallback(&unresolved.gradient);
@@ -673,7 +689,7 @@ macro_rules! impl_paint_source {
                 Ok(gradient)
             }
         }
-    }
+    };
 }
 
 impl_paint_source!(
@@ -742,12 +758,12 @@ impl Gradient {
         if self.units == GradientUnits(CoordUnits::ObjectBoundingBox) {
             let bbox_rect = bbox.rect.unwrap();
             let bbox_matrix = cairo::Matrix::new(
-                bbox_rect.width,
+                bbox_rect.width(),
                 0.0,
                 0.0,
-                bbox_rect.height,
-                bbox_rect.x,
-                bbox_rect.y,
+                bbox_rect.height(),
+                bbox_rect.x0,
+                bbox_rect.y0,
             );
             affine = cairo::Matrix::multiply(&affine, &bbox_matrix);
         }
@@ -781,15 +797,18 @@ fn acquire_gradient<'a>(
     draw_ctx: &'a mut DrawingCtx,
     fragment: &Fragment,
 ) -> Result<AcquiredNode, AcquireError> {
-    draw_ctx.acquire_node(fragment, &[NodeType::LinearGradient, NodeType::RadialGradient])
+    draw_ctx.acquire_node(
+        fragment,
+        &[NodeType::LinearGradient, NodeType::RadialGradient],
+    )
 }
 
 #[cfg(test)]
 mod tests {
     use super::*;
-    use markup5ever::{namespace_url, ns, QualName};
     use crate::float_eq_cairo::ApproxEqCairo;
     use crate::node::{NodeData, NodeType, RsvgNode};
+    use markup5ever::{namespace_url, ns, QualName};
 
     #[test]
     fn parses_spread_method() {
@@ -826,7 +845,7 @@ mod tests {
             &QualName::new(None, ns!(svg), local_name!("linearGradient")),
             None,
             None,
-            Box::new(LinearGradient::default())
+            Box::new(LinearGradient::default()),
         ));
 
         let borrow = node.borrow();
@@ -840,7 +859,7 @@ mod tests {
             &QualName::new(None, ns!(svg), local_name!("radialGradient")),
             None,
             None,
-            Box::new(RadialGradient::default())
+            Box::new(RadialGradient::default()),
         ));
 
         let borrow = node.borrow();
