@@ -1,38 +1,43 @@
+//! Toplevel handle for a loaded SVG document.
+//!
+//! This module provides the primitives on which the public APIs are implemented.
+
 use std::cell::Cell;
 use std::ptr;
 use std::rc::Rc;
 
-use cairo::{self, ImageSurface, Status};
-use gio;
-use glib;
-use libc;
 use locale_config::{LanguageRange, Locale};
 
 use crate::allowed_url::{AllowedUrl, Href};
 use crate::bbox::BoundingBox;
 use crate::document::Document;
 use crate::dpi::Dpi;
-use crate::drawing_ctx::{DrawingCtx, RsvgRectangle};
+use crate::drawing_ctx::DrawingCtx;
 use crate::error::{DefsLookupErrorKind, LoadingError, RenderingError};
 use crate::node::{CascadedValues, RsvgNode};
+use crate::rect::{IRect, Rect};
 use crate::structure::{IntrinsicDimensions, Svg};
 use url::Url;
 
+/// Loading options for SVG documents.
 #[derive(Clone)]
 pub struct LoadOptions {
-    /// Base URL
+    /// Base URL; all relative references will be resolved with respect to this.
     pub base_url: Option<Url>,
 
-    /// Whether to turn off size limits in libxml2
+    /// Whether to turn off size limits in libxml2.
     pub unlimited_size: bool,
 
-    /// Whether to keep original (undecoded) image data to embed in Cairo PDF surfaces
+    /// Whether to keep original (undecoded) image data to embed in Cairo PDF surfaces.
     pub keep_image_data: bool,
 
+    /// The environment's locale, used for the `<switch>` element and the `systemLanguage`
+    /// attribute.
     locale: Locale,
 }
 
 impl LoadOptions {
+    /// Creates a `LoadOptions` with defaults, and sets the `base_url`.
     pub fn new(base_url: Option<Url>) -> Self {
         LoadOptions {
             base_url,
@@ -42,19 +47,31 @@ impl LoadOptions {
         }
     }
 
+    /// Sets whether libxml2's limits on memory usage should be turned off.
+    ///
+    /// This should only be done for trusted data.
     pub fn with_unlimited_size(mut self, unlimited: bool) -> Self {
         self.unlimited_size = unlimited;
         self
     }
 
+    /// Sets whether to keep the original compressed image data from referenced JPEG/PNG images.
+    ///
+    /// This is only useful for rendering to Cairo PDF
+    /// surfaces, which can embed the original, compressed image data instead of uncompressed
+    /// RGB buffers.
     pub fn keep_image_data(mut self, keep: bool) -> Self {
         self.keep_image_data = keep;
         self
     }
 
+    /// Creates a new `LoadOptions` with a different `base_url`.
+    ///
+    /// This is used when loading a referenced file that may in turn cause other files
+    /// to be loaded, for example `<image xlink:href="subimage.svg"/>`
     pub fn copy_with_base_url(&self, base_url: &AllowedUrl) -> Self {
         LoadOptions {
-            base_url: Some((*base_url).clone()),
+            base_url: Some((**base_url).clone()),
             unlimited_size: self.unlimited_size,
             keep_image_data: self.keep_image_data,
             locale: self.locale.clone(),
@@ -186,7 +203,11 @@ impl Handle {
         cancellable: Option<&gio::Cancellable>,
     ) -> Result<Handle, LoadingError> {
         Ok(Handle {
-            document: Rc::new(Document::load_from_stream(load_options, stream, cancellable)?),
+            document: Rc::new(Document::load_from_stream(
+                load_options,
+                stream,
+                cancellable,
+            )?),
         })
     }
 
@@ -275,11 +296,11 @@ impl Handle {
     fn get_node_geometry_with_viewport(
         &self,
         node: &RsvgNode,
-        viewport: &cairo::Rectangle,
+        viewport: Rect,
         dpi: Dpi,
         is_testing: bool,
-    ) -> Result<(RsvgRectangle, RsvgRectangle), RenderingError> {
-        let target = ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
+    ) -> Result<(cairo::Rectangle, cairo::Rectangle), RenderingError> {
+        let target = cairo::ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
         let cr = cairo::Context::new(&target);
         let mut draw_ctx = DrawingCtx::new(
             self.document.clone(),
@@ -292,12 +313,16 @@ impl Handle {
         );
         let root = self.document.root();
 
-        let bbox = draw_ctx.draw_node_from_stack(&CascadedValues::new_from_node(&root), &root, false)?;
+        let bbox =
+            draw_ctx.draw_node_from_stack(&CascadedValues::new_from_node(&root), &root, false)?;
 
-        let ink_rect = bbox.ink_rect.map(RsvgRectangle::from).unwrap_or_default();
-        let logical_rect = bbox.rect.map(RsvgRectangle::from).unwrap_or_default();
+        let ink_rect = bbox.ink_rect.unwrap_or_default();
+        let logical_rect = bbox.rect.unwrap_or_default();
 
-        Ok((ink_rect, logical_rect))
+        Ok((
+            cairo::Rectangle::from(ink_rect),
+            cairo::Rectangle::from(logical_rect),
+        ))
     }
 
     /// Returns (ink_rect, logical_rect)
@@ -306,7 +331,7 @@ impl Handle {
         id: Option<&str>,
         dpi: Dpi,
         is_testing: bool,
-    ) -> Result<(RsvgRectangle, RsvgRectangle), RenderingError> {
+    ) -> Result<(cairo::Rectangle, cairo::Rectangle), RenderingError> {
         let node = self.get_node_or_root(id)?;
 
         let root = self.document.root();
@@ -319,20 +344,13 @@ impl Handle {
             if let Some((root_width, root_height)) =
                 node.borrow().get_impl::<Svg>().get_size(&values, dpi)
             {
-                let ink_r = RsvgRectangle {
-                    x: 0.0,
-                    y: 0.0,
-                    width: f64::from(root_width),
-                    height: f64::from(root_height),
-                };
+                let rect = IRect::from_size(root_width, root_height);
 
-                let logical_r = ink_r;
-
-                return Ok((ink_r, logical_r));
+                return Ok((cairo::Rectangle::from(rect), cairo::Rectangle::from(rect)));
             }
         }
 
-        self.get_node_geometry_with_viewport(&node, &unit_rectangle(), dpi, is_testing)
+        self.get_node_geometry_with_viewport(&node, unit_rectangle(), dpi, is_testing)
     }
 
     fn get_node_or_root(&self, id: Option<&str>) -> Result<RsvgNode, RenderingError> {
@@ -349,8 +367,9 @@ impl Handle {
         viewport: &cairo::Rectangle,
         dpi: Dpi,
         is_testing: bool,
-    ) -> Result<(RsvgRectangle, RsvgRectangle), RenderingError> {
+    ) -> Result<(cairo::Rectangle, cairo::Rectangle), RenderingError> {
         let node = self.get_node_or_root(id)?;
+        let viewport = Rect::from(*viewport);
         self.get_node_geometry_with_viewport(&node, viewport, dpi, is_testing)
     }
 
@@ -446,7 +465,7 @@ impl Handle {
             self.document.clone(),
             node.as_ref(),
             cr,
-            viewport,
+            Rect::from(*viewport),
             dpi,
             false,
             is_testing,
@@ -466,14 +485,14 @@ impl Handle {
         dpi: Dpi,
         is_testing: bool,
     ) -> Result<BoundingBox, RenderingError> {
-        let target = ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
+        let target = cairo::ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
         let cr = cairo::Context::new(&target);
 
         let mut draw_ctx = DrawingCtx::new(
             self.document.clone(),
             None,
             &cr,
-            &unit_rectangle(),
+            unit_rectangle(),
             dpi,
             true,
             is_testing,
@@ -488,26 +507,21 @@ impl Handle {
         id: Option<&str>,
         dpi: Dpi,
         is_testing: bool,
-    ) -> Result<(RsvgRectangle, RsvgRectangle), RenderingError> {
+    ) -> Result<(cairo::Rectangle, cairo::Rectangle), RenderingError> {
         let node = self.get_node_or_root(id)?;
 
         let bbox = self.get_bbox_for_element(&node, dpi, is_testing)?;
 
-        let mut ink_rect = bbox.ink_rect.map(RsvgRectangle::from).unwrap_or_default();
-        let mut logical_rect = bbox.rect.map(RsvgRectangle::from).unwrap_or_default();
+        let ink_rect = bbox.ink_rect.unwrap_or_default();
+        let logical_rect = bbox.rect.unwrap_or_default();
 
         // Translate so ink_rect is always at offset (0, 0)
+        let ofs = (-ink_rect.x0, -ink_rect.y0);
 
-        let xofs = ink_rect.x;
-        let yofs = ink_rect.y;
-
-        ink_rect.x -= xofs;
-        ink_rect.y -= yofs;
-
-        logical_rect.x -= xofs;
-        logical_rect.y -= yofs;
-
-        Ok((ink_rect, logical_rect))
+        Ok((
+            cairo::Rectangle::from(ink_rect.translate(ofs)),
+            cairo::Rectangle::from(logical_rect.translate(ofs)),
+        ))
     }
 
     pub fn render_element(
@@ -529,9 +543,9 @@ impl Handle {
             return Ok(());
         }
 
-        let ink_r = bbox.ink_rect.map(RsvgRectangle::from).unwrap_or_default();
+        let ink_r = bbox.ink_rect.unwrap_or_default();
 
-        if ink_r.width == 0.0 || ink_r.height == 0.0 {
+        if ink_r.is_empty() {
             return Ok(());
         }
 
@@ -540,17 +554,17 @@ impl Handle {
         cr.save();
 
         let factor =
-            (element_viewport.width / ink_r.width).min(element_viewport.height / ink_r.height);
+            (element_viewport.width / ink_r.width()).min(element_viewport.height / ink_r.height());
 
         cr.translate(element_viewport.x, element_viewport.y);
         cr.scale(factor, factor);
-        cr.translate(-ink_r.x, -ink_r.y);
+        cr.translate(-ink_r.x0, -ink_r.y0);
 
         let mut draw_ctx = DrawingCtx::new(
             self.document.clone(),
             None,
             &cr,
-            &unit_rectangle(),
+            unit_rectangle(),
             dpi,
             false,
             is_testing,
@@ -572,7 +586,7 @@ impl Handle {
 
 fn check_cairo_context(cr: &cairo::Context) -> Result<(), RenderingError> {
     let status = cr.status();
-    if status == Status::Success {
+    if status == cairo::Status::Success {
         Ok(())
     } else {
         Err(RenderingError::Cairo(status))
@@ -600,11 +614,6 @@ fn locale_from_environment() -> Locale {
     locale
 }
 
-fn unit_rectangle() -> cairo::Rectangle {
-    cairo::Rectangle {
-        x: 0.0,
-        y: 0.0,
-        width: 1.0,
-        height: 1.0,
-    }
+fn unit_rectangle() -> Rect {
+    Rect::from_size(1.0, 1.0)
 }
