@@ -1,8 +1,10 @@
+//! CSS funciri values.
+
 use cssparser::Parser;
 
 use crate::allowed_url::{Fragment, Href};
+use crate::error::*;
 use crate::parsers::Parse;
-use crate::parsers::ParseError;
 
 /// Used where style properties take a funciri or "none"
 ///
@@ -33,25 +35,25 @@ impl IRI {
 }
 
 impl Parse for IRI {
-    type Err = ParseError;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<IRI, ParseError> {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<IRI, ParseError<'i>> {
         if parser
             .try_parse(|i| i.expect_ident_matching("none"))
             .is_ok()
         {
             Ok(IRI::None)
         } else {
-            let url = parser
-                .expect_url()
-                .map_err(|_| ParseError::new("expected url"))?;
+            let loc = parser.current_source_location();
 
-            parser
-                .expect_exhausted()
-                .map_err(|_| ParseError::new("expected url"))?;
+            let url = parser.expect_url()?;
+            parser.expect_exhausted()?;
 
-            match Href::parse(&url).map_err(|_| ParseError::new("could not parse href"))? {
-                Href::PlainUrl(_) => Err(ParseError::new("href requires a fragment identifier")),
+            let href =
+                Href::parse(&url).map_err(|e| loc.new_custom_error(ValueErrorKind::from(e)))?;
+
+            match href {
+                Href::PlainUrl(_) => Err(loc.new_custom_error(ValueErrorKind::parse_error(
+                    "href requires a fragment identifier",
+                )))?,
                 Href::WithFragment(f) => Ok(IRI::Resource(f)),
             }
         }
