@@ -47,8 +47,7 @@ use std::marker::PhantomData;
 
 use crate::drawing_ctx::ViewParams;
 use crate::error::*;
-use crate::parsers::Parse;
-use crate::parsers::{finite_f32, ParseError};
+use crate::parsers::{finite_f32, Parse};
 use crate::properties::ComputedValues;
 
 /// Units for length values.
@@ -227,37 +226,27 @@ const CM_PER_INCH: f64 = 2.54;
 const MM_PER_INCH: f64 = 25.4;
 const PICA_PER_INCH: f64 = 6.0;
 
-fn make_err() -> ValueErrorKind {
-    ValueErrorKind::Parse(ParseError::new(
-        "expected length: number(\"em\" | \"ex\" | \"px\" | \"in\" | \"cm\" | \"mm\" | \"pt\" | \
-         \"pc\" | \"%\")?",
-    ))
-}
-
 impl<N: Normalize> Parse for Length<N> {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<Length<N>, ValueErrorKind> {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Length<N>, ParseError<'i>> {
         let length = {
-            let token = parser.next().map_err(|_| {
-                ValueErrorKind::Parse(ParseError::new(
-                    "expected number and optional symbol, or number and percentage",
-                ))
-            })?;
+            let token = parser.next()?.clone();
 
-            match *token {
-                Token::Number { value, .. } => {
-                    Length::new(f64::from(finite_f32(value)?), LengthUnit::Px)
-                }
+            match token {
+                Token::Number { value, .. } => Length::new(
+                    f64::from(finite_f32(value).map_err(|e| parser.new_custom_error(e))?),
+                    LengthUnit::Px,
+                ),
 
-                Token::Percentage { unit_value, .. } => {
-                    Length::new(f64::from(finite_f32(unit_value)?), LengthUnit::Percent)
-                }
+                Token::Percentage { unit_value, .. } => Length::new(
+                    f64::from(finite_f32(unit_value).map_err(|e| parser.new_custom_error(e))?),
+                    LengthUnit::Percent,
+                ),
 
                 Token::Dimension {
                     value, ref unit, ..
                 } => {
-                    let value = f64::from(finite_f32(value)?);
+                    let value =
+                        f64::from(finite_f32(value).map_err(|e| parser.new_custom_error(e))?);
 
                     match unit.as_ref() {
                         "px" => Length::new(value, LengthUnit::Px),
@@ -269,11 +258,11 @@ impl<N: Normalize> Parse for Length<N> {
                         "pt" => Length::new(value, LengthUnit::Pt),
                         "pc" => Length::new(value, LengthUnit::Pc),
 
-                        _ => return Err(make_err()),
+                        _ => return Err(parser.new_unexpected_token_error(token.clone()))?,
                     }
                 }
 
-                _ => return Err(make_err()),
+                _ => return Err(parser.new_unexpected_token_error(token.clone()))?,
             }
         };
 
@@ -340,18 +329,14 @@ impl<N: Normalize> Length<N> {
 
             LengthUnit::Ex => self.length * font_size_from_values(values, params) / 2.0,
 
-            LengthUnit::In => {
-                self.length * <N as Normalize>::normalize(params.dpi_x, params.dpi_y)
-            }
+            LengthUnit::In => self.length * <N as Normalize>::normalize(params.dpi_x, params.dpi_y),
 
             LengthUnit::Cm => {
-                self.length * <N as Normalize>::normalize(params.dpi_x, params.dpi_y)
-                    / CM_PER_INCH
+                self.length * <N as Normalize>::normalize(params.dpi_x, params.dpi_y) / CM_PER_INCH
             }
 
             LengthUnit::Mm => {
-                self.length * <N as Normalize>::normalize(params.dpi_x, params.dpi_y)
-                    / MM_PER_INCH
+                self.length * <N as Normalize>::normalize(params.dpi_x, params.dpi_y) / MM_PER_INCH
             }
 
             LengthUnit::Pt => {
@@ -398,50 +383,6 @@ fn viewport_percentage(x: f64, y: f64) -> f64 {
     (x * x + y * y).sqrt() / SQRT_2
 }
 
-#[derive(Debug, PartialEq, Clone)]
-pub enum Dasharray {
-    None,
-    Array(Vec<Length<Both>>),
-}
-
-impl Default for Dasharray {
-    fn default() -> Dasharray {
-        Dasharray::None
-    }
-}
-
-impl Parse for Dasharray {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<Dasharray, ValueErrorKind> {
-        if parser
-            .try_parse(|p| p.expect_ident_matching("none"))
-            .is_ok()
-        {
-            Ok(Dasharray::None)
-        } else {
-            Ok(Dasharray::Array(parse_dash_array(parser)?))
-        }
-    }
-}
-
-// This does not handle "inherit" or "none" state, the caller is responsible for that.
-fn parse_dash_array(parser: &mut Parser<'_, '_>) -> Result<Vec<Length<Both>>, ValueErrorKind> {
-    let mut dasharray = Vec::new();
-
-    loop {
-        dasharray.push(Length::<Both>::parse(parser).and_then(Length::<Both>::check_nonnegative)?);
-
-        if parser.is_exhausted() {
-            break;
-        } else if parser.try_parse(|p| p.expect_comma()).is_ok() {
-            continue;
-        }
-    }
-
-    Ok(dasharray)
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -515,24 +456,26 @@ mod tests {
 
     #[test]
     fn empty_length_yields_error() {
-        assert!(is_parse_error(&Length::<Both>::parse_str("")));
+        assert!(Length::<Both>::parse_str("").is_err());
     }
 
     #[test]
     fn invalid_unit_yields_error() {
-        assert!(is_parse_error(&Length::<Both>::parse_str("8furlong")));
+        assert!(Length::<Both>::parse_str("8furlong").is_err());
     }
 
     #[test]
     fn check_nonnegative_works() {
         // and_then with anonymous function
         assert!(Length::<Both>::parse_str("0")
-            .and_then(|l| l.check_nonnegative())
+            .unwrap()
+            .check_nonnegative()
             .is_ok());
 
         // and_then with named function
         assert!(Length::<Both>::parse_str("-10")
-            .and_then(Length::check_nonnegative)
+            .unwrap()
+            .check_nonnegative()
             .is_err());
     }
 
@@ -616,74 +559,4 @@ mod tests {
             6.0
         );
     }
-
-    fn parse_dash_array_str(s: &str) -> Result<Dasharray, ValueErrorKind> {
-        Dasharray::parse_str(s)
-    }
-
-    #[test]
-    fn parses_dash_array() {
-        // helper to cut down boilderplate
-        let length_parse = |s| Length::<Both>::parse_str(s).unwrap();
-
-        let expected = Dasharray::Array(vec![
-            length_parse("1"),
-            length_parse("2in"),
-            length_parse("3"),
-            length_parse("4%"),
-        ]);
-
-        let sample_1 = Dasharray::Array(vec![length_parse("10"), length_parse("6")]);
-
-        let sample_2 = Dasharray::Array(vec![
-            length_parse("5"),
-            length_parse("5"),
-            length_parse("20"),
-        ]);
-
-        let sample_3 = Dasharray::Array(vec![
-            length_parse("10px"),
-            length_parse("20px"),
-            length_parse("20px"),
-        ]);
-
-        let sample_4 = Dasharray::Array(vec![
-            length_parse("25"),
-            length_parse("5"),
-            length_parse("5"),
-            length_parse("5"),
-        ]);
-
-        let sample_5 = Dasharray::Array(vec![length_parse("3.1415926"), length_parse("8")]);
-        let sample_6 = Dasharray::Array(vec![length_parse("5"), length_parse("3.14")]);
-        let sample_7 = Dasharray::Array(vec![length_parse("2")]);
-
-        assert_eq!(parse_dash_array_str("none").unwrap(), Dasharray::None);
-        assert_eq!(parse_dash_array_str("1 2in,3 4%").unwrap(), expected);
-        assert_eq!(parse_dash_array_str("10,6").unwrap(), sample_1);
-        assert_eq!(parse_dash_array_str("5,5,20").unwrap(), sample_2);
-        assert_eq!(parse_dash_array_str("10px 20px 20px").unwrap(), sample_3);
-        assert_eq!(parse_dash_array_str("25  5 , 5 5").unwrap(), sample_4);
-        assert_eq!(parse_dash_array_str("3.1415926,8").unwrap(), sample_5);
-        assert_eq!(parse_dash_array_str("5, 3.14").unwrap(), sample_6);
-        assert_eq!(parse_dash_array_str("2").unwrap(), sample_7);
-
-        // Negative numbers
-        assert_eq!(
-            parse_dash_array_str("20,40,-20"),
-            Err(ValueErrorKind::Value(String::from(
-                "value must be non-negative"
-            )))
-        );
-
-        // Empty dash_array
-        assert!(parse_dash_array_str("").is_err());
-        assert!(parse_dash_array_str("\t  \n     ").is_err());
-        assert!(parse_dash_array_str(",,,").is_err());
-        assert!(parse_dash_array_str("10,  \t, 20 \n").is_err());
-        // No trailing commas allowed, parse error
-        assert!(parse_dash_array_str("10,").is_err());
-        // A comma should be followed by a number
-        assert!(parse_dash_array_str("20,,10").is_err());
-    }
 }
