@@ -1,8 +1,9 @@
+//! The `marker` element, and geometry computations for markers.
+
 use std::f64::consts::*;
 use std::ops::Deref;
 
-use cairo::Rectangle;
-use cssparser::{CowRcStr, Parser, Token};
+use cssparser::Parser;
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::allowed_url::Fragment;
@@ -15,11 +16,11 @@ use crate::float_eq_cairo::ApproxEqCairo;
 use crate::iri::IRI;
 use crate::length::*;
 use crate::node::*;
-use crate::parsers::{Parse, ParseError, ParseValue};
+use crate::parsers::{Parse, ParseValue};
 use crate::path_builder::*;
 use crate::properties::{ComputedValues, SpecifiedValue, SpecifiedValues};
 use crate::property_bag::PropertyBag;
-use crate::rect::RectangleExt;
+use crate::rect::Rect;
 use crate::viewbox::*;
 
 // markerUnits attribute: https://www.w3.org/TR/SVG/painting.html#MarkerElement
@@ -36,27 +37,12 @@ impl Default for MarkerUnits {
 }
 
 impl Parse for MarkerUnits {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<MarkerUnits, ValueErrorKind> {
-        let loc = parser.current_source_location();
-
-        parser
-            .expect_ident()
-            .and_then(|cow| match cow.as_ref() {
-                "userSpaceOnUse" => Ok(MarkerUnits::UserSpaceOnUse),
-                "strokeWidth" => Ok(MarkerUnits::StrokeWidth),
-                _ => Err(
-                    loc.new_basic_unexpected_token_error(Token::Ident(CowRcStr::from(
-                        cow.as_ref().to_string(),
-                    ))),
-                ),
-            })
-            .map_err(|_| {
-                ValueErrorKind::Parse(ParseError::new(
-                    "expected \"userSpaceOnUse\" or \"strokeWidth\"",
-                ))
-            })
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<MarkerUnits, ParseError<'i>> {
+        Ok(parse_identifiers!(
+            parser,
+            "userSpaceOnUse" => MarkerUnits::UserSpaceOnUse,
+            "strokeWidth" => MarkerUnits::StrokeWidth,
+        )?)
     }
 }
 
@@ -74,17 +60,10 @@ impl Default for MarkerOrient {
 }
 
 impl Parse for MarkerOrient {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<MarkerOrient, ValueErrorKind> {
-        if parser
-            .try_parse(|p| p.expect_ident_matching("auto"))
-            .is_ok()
-        {
-            Ok(MarkerOrient::Auto)
-        } else {
-            Angle::parse(parser).map(MarkerOrient::Angle)
-        }
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<MarkerOrient, ParseError<'i>> {
+        parser
+            .try_parse(|p| p.expect_ident_matching("auto").map(|_| MarkerOrient::Auto))
+            .or_else(|_| Angle::parse(parser).map(MarkerOrient::Angle))
     }
 }
 
@@ -157,18 +136,18 @@ impl Marker {
             }
 
             let params = if let Some(vbox) = self.vbox {
-                let (_, _, w, h) = self.aspect.compute(
-                    &vbox,
-                    &Rectangle::new(0.0, 0.0, marker_width, marker_height),
-                );
-
-                if vbox.width.approx_eq_cairo(0.0) || vbox.height.approx_eq_cairo(0.0) {
+                if vbox.0.is_empty() {
                     return Ok(dc.empty_bbox());
                 }
 
-                cr.scale(w / vbox.width, h / vbox.height);
+                let r = self
+                    .aspect
+                    .compute(&vbox, &Rect::from_size(marker_width, marker_height));
 
-                dc.push_view_box(vbox.width, vbox.height)
+                let (vb_width, vb_height) = vbox.0.size();
+                cr.scale(r.width() / vb_width, r.height() / vb_height);
+
+                dc.push_view_box(vb_width, vb_height)
             } else {
                 dc.push_view_box(marker_width, marker_height)
             };
@@ -179,11 +158,11 @@ impl Marker {
             );
 
             if !values.is_overflow() {
-                if let Some(vbox) = self.vbox {
-                    dc.clip(vbox.x, vbox.y, vbox.width, vbox.height);
-                } else {
-                    dc.clip(0.0, 0.0, marker_width, marker_height);
-                }
+                let clip_rect = self
+                    .vbox
+                    .map_or_else(|| Rect::from_size(marker_width, marker_height), |vb| vb.0);
+
+                dc.clip(clip_rect);
             }
 
             dc.with_discrete_layer(node, values, clipping, &mut |dc| {
@@ -197,20 +176,20 @@ impl NodeTrait for Marker {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "markerUnits") => self.units = attr.parse(value)?,
-                expanded_name!(svg "refX") => self.ref_x = attr.parse(value)?,
-                expanded_name!(svg "refY") => self.ref_y = attr.parse(value)?,
-                expanded_name!(svg "markerWidth") => {
+                expanded_name!("", "markerUnits") => self.units = attr.parse(value)?,
+                expanded_name!("", "refX") => self.ref_x = attr.parse(value)?,
+                expanded_name!("", "refY") => self.ref_y = attr.parse(value)?,
+                expanded_name!("", "markerWidth") => {
                     self.width =
                         attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?
                 }
-                expanded_name!(svg "markerHeight") => {
+                expanded_name!("", "markerHeight") => {
                     self.height =
                         attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?
                 }
-                expanded_name!(svg "orient") => self.orient = attr.parse(value)?,
-                expanded_name!(svg "preserveAspectRatio") => self.aspect = attr.parse(value)?,
-                expanded_name!(svg "viewBox") => self.vbox = Some(attr.parse(value)?),
+                expanded_name!("", "orient") => self.orient = attr.parse(value)?,
+                expanded_name!("", "preserveAspectRatio") => self.aspect = attr.parse(value)?,
+                expanded_name!("", "viewBox") => self.vbox = Some(attr.parse(value)?),
                 _ => (),
             }
         }
@@ -524,18 +503,18 @@ impl<'a> From<&'a PathBuilder> for Segments {
 // segment's start and end points to align with the positive x-axis
 // in user space.
 impl Segments {
-    fn find_incoming_directionality_backwards(&self, start_index: usize) -> (bool, f64, f64) {
+    fn find_incoming_angle_backwards(&self, start_index: usize) -> Option<Angle> {
         // "go backwards ... within the current subpath until ... segment which has directionality
         // at its end point"
         for segment in self[..=start_index].iter().rev() {
             match *segment {
                 Segment::Degenerate { .. } => {
-                    return (false, 0.0, 0.0); // reached the beginning of the subpath as we ran into a standalone point
+                    return None; // reached the beginning of the subpath as we ran into a standalone point
                 }
 
                 Segment::LineOrCurve { .. } => match segment.get_directionalities() {
                     Some((_, _, v2x, v2y)) => {
-                        return (true, v2x, v2y);
+                        return Some(Angle::from_vector(v2x, v2y));
                     }
                     None => {
                         continue;
@@ -544,21 +523,21 @@ impl Segments {
             }
         }
 
-        (false, 0.0, 0.0)
+        None
     }
 
-    fn find_outgoing_directionality_forwards(&self, start_index: usize) -> (bool, f64, f64) {
+    fn find_outgoing_angle_forwards(&self, start_index: usize) -> Option<Angle> {
         // "go forwards ... within the current subpath until ... segment which has directionality at
         // its start point"
         for segment in &self[start_index..] {
             match *segment {
                 Segment::Degenerate { .. } => {
-                    return (false, 0.0, 0.0); // reached the end of a subpath as we ran into a standalone point
+                    return None; // reached the end of a subpath as we ran into a standalone point
                 }
 
                 Segment::LineOrCurve { .. } => match segment.get_directionalities() {
                     Some((v1x, v1y, _, _)) => {
-                        return (true, v1x, v1y);
+                        return Some(Angle::from_vector(v1x, v1y));
                     }
                     None => {
                         continue;
@@ -567,7 +546,7 @@ impl Segments {
             }
         }
 
-        (false, 0.0, 0.0)
+        None
     }
 }
 
@@ -588,8 +567,7 @@ fn emit_marker_by_name(
     line_width: f64,
     clipping: bool,
 ) -> Result<BoundingBox, RenderingError> {
-    if let Ok(acquired) = draw_ctx.acquire_node(name, &[NodeType::Marker])
-    {
+    if let Ok(acquired) = draw_ctx.acquire_node(name, &[NodeType::Marker]) {
         let node = acquired.get();
 
         node.borrow().get_impl::<Marker>().render(
@@ -702,13 +680,14 @@ where
                     assert!(i > 0);
 
                     // Got a lone point after a subpath; render the subpath's end marker first
-                    let (_, incoming_vx, incoming_vy) =
-                        segments.find_incoming_directionality_backwards(i - 1);
+                    let angle = segments
+                        .find_incoming_angle_backwards(i - 1)
+                        .unwrap_or_else(|| Angle::new(0.0));
                     let marker_bbox = emit_marker(
                         &segments[i - 1],
                         MarkerEndpoint::End,
                         MarkerType::End,
-                        Angle::from_vector(incoming_vx, incoming_vy),
+                        angle,
                         emit_fn,
                     )?;
                     bbox.insert(&marker_bbox);
@@ -731,13 +710,14 @@ where
                 // Not a degenerate segment
                 match subpath_state {
                     SubpathState::NoSubpath => {
-                        let (_, outgoing_vx, outgoing_vy) =
-                            segments.find_outgoing_directionality_forwards(i);
+                        let angle = segments
+                            .find_outgoing_angle_forwards(i)
+                            .unwrap_or_else(|| Angle::new(0.0));
                         let marker_bbox = emit_marker(
                             segment,
                             MarkerEndpoint::Start,
                             MarkerType::Start,
-                            Angle::from_vector(outgoing_vx, outgoing_vy),
+                            angle,
                             emit_fn,
                         )?;
                         bbox.insert(&marker_bbox);
@@ -748,25 +728,15 @@ where
                     SubpathState::InSubpath => {
                         assert!(i > 0);
 
-                        let (has_incoming, incoming_vx, incoming_vy) =
-                            segments.find_incoming_directionality_backwards(i - 1);
-                        let (has_outgoing, outgoing_vx, outgoing_vy) =
-                            segments.find_outgoing_directionality_forwards(i);
+                        let incoming = segments.find_incoming_angle_backwards(i - 1);
+                        let outgoing = segments.find_outgoing_angle_forwards(i);
 
-                        let incoming = Angle::from_vector(incoming_vx, incoming_vy);
-                        let outgoing = Angle::from_vector(outgoing_vx, outgoing_vy);
-
-                        let angle: Angle;
-
-                        if has_incoming && has_outgoing {
-                            angle = incoming.bisect(outgoing);
-                        } else if has_incoming {
-                            angle = incoming;
-                        } else if has_outgoing {
-                            angle = outgoing;
-                        } else {
-                            angle = Angle::new(0.0);
-                        }
+                        let angle = match (incoming, outgoing) {
+                            (Some(incoming), Some(outgoing)) => incoming.bisect(outgoing),
+                            (Some(incoming), _) => incoming,
+                            (_, Some(outgoing)) => outgoing,
+                            _ => Angle::new(0.0),
+                        };
 
                         let marker_bbox = emit_marker(
                             segment,
@@ -786,18 +756,18 @@ where
     if !segments.is_empty() {
         let segment = &segments[segments.len() - 1];
         if let Segment::LineOrCurve { .. } = *segment {
-            let (_, incoming_vx, incoming_vy) =
-                segments.find_incoming_directionality_backwards(segments.len() - 1);
+            let incoming = segments
+                .find_incoming_angle_backwards(segments.len() - 1)
+                .unwrap_or_else(|| Angle::new(0.0));
 
             let angle = {
                 if let PathCommand::ClosePath = builder.get_path_commands()[segments.len()] {
-                    let (_, outgoing_vx, outgoing_vy) =
-                        segments.find_outgoing_directionality_forwards(0);
-                    let incoming = Angle::from_vector(incoming_vx, incoming_vy);
-                    let outgoing = Angle::from_vector(outgoing_vx, outgoing_vy);
+                    let outgoing = segments
+                        .find_outgoing_angle_forwards(0)
+                        .unwrap_or_else(|| Angle::new(0.0));
                     incoming.bisect(outgoing)
                 } else {
-                    Angle::from_vector(incoming_vx, incoming_vy)
+                    incoming
                 }
             };
 
@@ -821,12 +791,8 @@ mod parser_tests {
 
     #[test]
     fn parsing_invalid_marker_units_yields_error() {
-        assert!(is_parse_error(
-            &MarkerUnits::parse_str("").map_err(|e| ValueErrorKind::from(e))
-        ));
-        assert!(is_parse_error(
-            &MarkerUnits::parse_str("foo").map_err(|e| ValueErrorKind::from(e))
-        ));
+        assert!(MarkerUnits::parse_str("").is_err());
+        assert!(MarkerUnits::parse_str("foo").is_err());
     }
 
     #[test]
@@ -843,15 +809,9 @@ mod parser_tests {
 
     #[test]
     fn parsing_invalid_marker_orient_yields_error() {
-        assert!(is_parse_error(
-            &MarkerOrient::parse_str("").map_err(|e| ValueErrorKind::from(e))
-        ));
-        assert!(is_parse_error(
-            &MarkerOrient::parse_str("blah").map_err(|e| ValueErrorKind::from(e))
-        ));
-        assert!(is_parse_error(
-            &MarkerOrient::parse_str("45blah").map_err(|e| ValueErrorKind::from(e))
-        ));
+        assert!(MarkerOrient::parse_str("").is_err());
+        assert!(MarkerOrient::parse_str("blah").is_err());
+        assert!(MarkerOrient::parse_str("45blah").is_err());
     }
 
     #[test]
@@ -1145,14 +1105,14 @@ mod marker_tests {
 
         assert!(emit_markers_for_path_builder(
             &builder,
-            BoundingBox::new(&cairo::Matrix::identity()),
+            BoundingBox::new(),
             &mut |marker_type: MarkerType,
                   x: f64,
                   y: f64,
                   computed_angle: Angle|
              -> Result<BoundingBox, RenderingError> {
                 v.push((marker_type, x, y, computed_angle));
-                Ok(BoundingBox::new(&cairo::Matrix::identity()))
+                Ok(BoundingBox::new())
             }
         )
         .is_ok());
@@ -1181,14 +1141,14 @@ mod marker_tests {
 
         assert!(emit_markers_for_path_builder(
             &builder,
-            BoundingBox::new(&cairo::Matrix::identity()),
+            BoundingBox::new(),
             &mut |marker_type: MarkerType,
                   x: f64,
                   y: f64,
                   computed_angle: Angle|
              -> Result<BoundingBox, RenderingError> {
                 v.push((marker_type, x, y, computed_angle));
-                Ok(BoundingBox::new(&cairo::Matrix::identity()))
+                Ok(BoundingBox::new())
             }
         )
         .is_ok());
