@@ -1,5 +1,19 @@
-use cairo::Matrix;
+//! Tree nodes, the representation of SVG elements.
+//!
+//! Librsvg uses the [rctree crate][rctree] to represent the SVG tree of elements.
+//! Its [`Node`] struct provides a generic wrapper over nodes in a tree.
+//! Librsvg puts a [`NodeData`] as the type parameter of [`Node`].  For convenience,
+//! librsvg has a type alias `RsvgNode = Node<NodeData>`.
+//!
+//! Nodes are not constructed directly by callers; this is done in the [`create_node`] module.
+//!
+//! [rctree]: ../../rctree/index.html
+//! [`Node`]: ../../rctree/struct.Node.html
+//! [`NodeData`]: struct.NodeData.html
+//! [`create_node`]: ../create_node/index.html
+
 use downcast_rs::*;
+use locale_config::Locale;
 use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
 use std::cell::Ref;
 use std::collections::HashSet;
@@ -15,11 +29,15 @@ use crate::parsers::Parse;
 use crate::properties::{ComputedValues, SpecifiedValue, SpecifiedValues};
 use crate::property_bag::PropertyBag;
 use crate::property_defs::Overflow;
-use locale_config::Locale;
-use rctree;
 
-/// Tree node with specific data
+/// Strong reference to an element in the SVG tree.
+///
+/// See the [module documentation](index.html) for more information.
 pub type RsvgNode = rctree::Node<NodeData>;
+
+/// Weak reference to an element in the SVG tree.
+///
+/// See the [module documentation](index.html) for more information.
 pub type RsvgWeakNode = rctree::WeakNode<NodeData>;
 
 /// Contents of a tree node
@@ -31,7 +49,7 @@ pub struct NodeData {
     specified_values: SpecifiedValues,
     important_styles: HashSet<QualName>,
     result: NodeResult,
-    transform: Matrix,
+    transform: cairo::Matrix,
     values: ComputedValues,
     cond: bool,
     style_attr: String,
@@ -53,7 +71,7 @@ impl NodeData {
             class: class.map(str::to_string),
             specified_values: Default::default(),
             important_styles: Default::default(),
-            transform: Matrix::identity(),
+            transform: cairo::Matrix::identity(),
             result: Ok(()),
             values: ComputedValues::default(),
             cond: true,
@@ -94,7 +112,7 @@ impl NodeData {
         self.cond
     }
 
-    pub fn get_transform(&self) -> Matrix {
+    pub fn get_transform(&self) -> cairo::Matrix {
         self.transform
     }
 
@@ -125,7 +143,7 @@ impl NodeData {
     fn save_style_attribute(&mut self, pbag: &PropertyBag<'_>) {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "style") => self.style_attr.push_str(value),
+                expanded_name!("", "style") => self.style_attr.push_str(value),
                 _ => (),
             }
         }
@@ -134,11 +152,13 @@ impl NodeData {
     fn set_transform_attribute(&mut self, pbag: &PropertyBag<'_>) -> Result<(), NodeError> {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "transform") => {
-                    return Matrix::parse_str(value).attribute(attr).and_then(|affine| {
-                        self.transform = affine;
-                        Ok(())
-                    });
+                expanded_name!("", "transform") => {
+                    return cairo::Matrix::parse_str(value)
+                        .attribute(attr)
+                        .and_then(|affine| {
+                            self.transform = affine;
+                            Ok(())
+                        });
                 }
                 _ => (),
             }
@@ -155,20 +175,19 @@ impl NodeData {
         let mut cond = self.cond;
 
         for (attr, value) in pbag.iter() {
-            // FIXME: move this to "try {}" when we can bump the rustc version dependency
-            let mut parse = || {
+            let mut parse = || -> Result<_, ValueErrorKind> {
                 match attr.expanded() {
-                    expanded_name!(svg "requiredExtensions") if cond => {
+                    expanded_name!("", "requiredExtensions") if cond => {
                         cond = RequiredExtensions::from_attribute(value)
                             .map(|RequiredExtensions(res)| res)?;
                     }
 
-                    expanded_name!(svg "requiredFeatures") if cond => {
+                    expanded_name!("", "requiredFeatures") if cond => {
                         cond = RequiredFeatures::from_attribute(value)
                             .map(|RequiredFeatures(res)| res)?;
                     }
 
-                    expanded_name!(svg "systemLanguage") if cond => {
+                    expanded_name!("", "systemLanguage") if cond => {
                         cond = SystemLanguage::from_attribute(value, locale)
                             .map(|SystemLanguage(res)| res)?;
                     }
@@ -179,9 +198,7 @@ impl NodeData {
                 Ok(cond)
             };
 
-            parse()
-                .map(|c| self.cond = c)
-                .map_err(|e| NodeError::attribute_error(attr, e))?;
+            parse().map(|c| self.cond = c).attribute(attr)?;
         }
 
         Ok(())
@@ -485,10 +502,8 @@ impl NodeDraw for RsvgNode {
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
         if !self.borrow().is_in_error() {
-            draw_ctx.with_saved_matrix(&mut |dc| {
-                let cr = dc.get_cairo_context();
-                cr.transform(self.borrow().get_transform());
-
+            let transform = self.borrow().get_transform();
+            draw_ctx.with_saved_matrix(Some(transform), &mut |dc| {
                 self.borrow()
                     .get_node_trait()
                     .draw(self, cascaded, dc, clipping)
