@@ -1,6 +1,9 @@
+//! CSS number-list values.
+
 use cssparser::{Parser, ParserInput};
 
-use crate::parsers::{CssParserExt, ParseError};
+use crate::error::*;
+use crate::parsers::{optional_comma, Parse};
 
 #[derive(Eq, PartialEq)]
 pub enum NumberListLength {
@@ -8,23 +11,17 @@ pub enum NumberListLength {
     Unbounded,
 }
 
-#[derive(Debug, PartialEq)]
-pub enum NumberListError {
-    IncorrectNumberOfElements,
-    Parse(ParseError),
-}
-
 #[derive(Debug, PartialEq)]
 pub struct NumberList(pub Vec<f64>);
 
 impl NumberList {
-    pub fn parse(
-        parser: &mut Parser<'_, '_>,
+    pub fn parse<'i>(
+        parser: &mut Parser<'i, '_>,
         length: NumberListLength,
-    ) -> Result<NumberList, NumberListError> {
+    ) -> Result<Self, ParseError<'i>> {
         let mut v = match length {
             NumberListLength::Exact(l) if l > 0 => Vec::<f64>::with_capacity(l),
-            NumberListLength::Exact(_) => unreachable!(),
+            NumberListLength::Exact(_) => unreachable!("NumberListLength::Exact cannot be 0"),
             NumberListLength::Unbounded => Vec::<f64>::new(),
         };
 
@@ -34,12 +31,10 @@ impl NumberList {
 
         for i in 0.. {
             if i != 0 {
-                parser.optional_comma();
+                optional_comma(parser);
             }
 
-            v.push(f64::from(parser.expect_finite_number().map_err(|_| {
-                NumberListError::Parse(ParseError::new("expected number"))
-            })?));
+            v.push(f64::parse(parser)?);
 
             if let NumberListLength::Exact(l) = length {
                 if i + 1 == l {
@@ -59,14 +54,15 @@ impl NumberList {
             }
         }
 
-        parser
-            .expect_exhausted()
-            .map_err(|_| NumberListError::IncorrectNumberOfElements)?;
+        parser.expect_exhausted()?;
 
         Ok(NumberList(v))
     }
 
-    pub fn parse_str(s: &str, length: NumberListLength) -> Result<NumberList, NumberListError> {
+    pub fn parse_str<'i>(
+        s: &'i str,
+        length: NumberListLength,
+    ) -> Result<NumberList, ParseError<'i>> {
         let mut input = ParserInput::new(s);
         let mut parser = Parser::new(&mut input);
 
