@@ -1,10 +1,13 @@
-use crate::path_builder::*;
+//! Parser for SVG path data.
+
 use std::error::Error;
-use std::fmt::{self, Display, Formatter};
+use std::fmt;
 use std::iter::Enumerate;
 use std::str;
 use std::str::Chars;
 
+use crate::path_builder::*;
+
 struct PathParser<'b> {
     chars_enumerator: Enumerate<Chars<'b>>,
     lookahead: Option<char>,    // None if we are in EOF
@@ -918,8 +921,8 @@ impl Error for ParseError {
     }
 }
 
-impl Display for ParseError {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
+impl fmt::Display for ParseError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         write!(
             f,
             "error at position {}: {}",
