@@ -1,4 +1,5 @@
-use cairo;
+//! The `pattern` element.
+
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use std::cell::RefCell;
 use std::f64;
@@ -7,8 +8,8 @@ use crate::allowed_url::Fragment;
 use crate::aspect_ratio::*;
 use crate::bbox::*;
 use crate::coord_units::CoordUnits;
-use crate::drawing_ctx::{DrawingCtx, NodeStack};
-use crate::error::{AcquireError, AttributeResultExt, RenderingError};
+use crate::drawing_ctx::{DrawingCtx, NodeStack, ViewParams};
+use crate::error::*;
 use crate::float_eq_cairo::ApproxEqCairo;
 use crate::length::*;
 use crate::node::*;
@@ -16,7 +17,7 @@ use crate::paint_server::{AsPaintSource, PaintSource};
 use crate::parsers::ParseValue;
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
-use crate::rect::RectangleExt;
+use crate::rect::Rect;
 use crate::unit_interval::UnitInterval;
 use crate::viewbox::*;
 
@@ -118,25 +119,28 @@ impl NodeTrait for Pattern {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "patternUnits") => self.common.units = Some(attr.parse(value)?),
-                expanded_name!(svg "patternContentUnits") => {
+                expanded_name!("", "patternUnits") => self.common.units = Some(attr.parse(value)?),
+                expanded_name!("", "patternContentUnits") => {
                     self.common.content_units = Some(attr.parse(value)?)
                 }
-                expanded_name!(svg "viewBox") => self.common.vbox = Some(Some(attr.parse(value)?)),
-                expanded_name!(svg "preserveAspectRatio") => {
+                expanded_name!("", "viewBox") => self.common.vbox = Some(Some(attr.parse(value)?)),
+                expanded_name!("", "preserveAspectRatio") => {
                     self.common.preserve_aspect_ratio = Some(attr.parse(value)?)
                 }
-                expanded_name!(svg "patternTransform") => self.common.affine = Some(attr.parse(value)?),
+                expanded_name!("", "patternTransform") => {
+                    self.common.affine = Some(attr.parse(value)?)
+                }
                 expanded_name!(xlink "href") => {
                     self.fallback = Some(Fragment::parse(value).attribute(attr)?);
                 }
-                expanded_name!(svg "x") => self.common.x = Some(attr.parse(value)?),
-                expanded_name!(svg "y") => self.common.y = Some(attr.parse(value)?),
-                expanded_name!(svg "width") => {
-                    self.common.width =
-                        Some(attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?)
+                expanded_name!("", "x") => self.common.x = Some(attr.parse(value)?),
+                expanded_name!("", "y") => self.common.y = Some(attr.parse(value)?),
+                expanded_name!("", "width") => {
+                    self.common.width = Some(
+                        attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?,
+                    )
                 }
-                expanded_name!(svg "height") => {
+                expanded_name!("", "height") => {
                     self.common.height =
                         Some(attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?)
                 }
@@ -238,38 +242,20 @@ impl AsPaintSource for ResolvedPattern {
         let vbox = self.vbox;
         let preserve_aspect_ratio = self.preserve_aspect_ratio;
 
-        let (pattern_x, pattern_y, pattern_width, pattern_height) = {
-            let params = if units == PatternUnits(CoordUnits::ObjectBoundingBox) {
-                draw_ctx.push_view_box(1.0, 1.0)
-            } else {
-                draw_ctx.get_view_params()
-            };
-
-            let pattern_x = self.x.normalize(values, &params);
-            let pattern_y = self.y.normalize(values, &params);
-            let pattern_width = self.width.normalize(values, &params);
-            let pattern_height = self.height.normalize(values, &params);
-
-            (pattern_x, pattern_y, pattern_width, pattern_height)
+        let params = if units == PatternUnits(CoordUnits::ObjectBoundingBox) {
+            draw_ctx.push_view_box(1.0, 1.0)
+        } else {
+            draw_ctx.get_view_params()
         };
 
+        let pattern_rect = self.get_rect(values, &params);
+
         // Work out the size of the rectangle so it takes into account the object bounding box
 
-        let bbwscale: f64;
-        let bbhscale: f64;
-
-        match units {
-            PatternUnits(CoordUnits::ObjectBoundingBox) => {
-                let bbrect = bbox.rect.unwrap();
-                bbwscale = bbrect.width;
-                bbhscale = bbrect.height;
-            }
-
-            PatternUnits(CoordUnits::UserSpaceOnUse) => {
-                bbwscale = 1.0;
-                bbhscale = 1.0;
-            }
-        }
+        let (bbwscale, bbhscale) = match units {
+            PatternUnits(CoordUnits::ObjectBoundingBox) => bbox.rect.unwrap().size(),
+            PatternUnits(CoordUnits::UserSpaceOnUse) => (1.0, 1.0),
+        };
 
         let cr = draw_ctx.get_cairo_context();
         let affine = cr.get_matrix();
@@ -278,11 +264,11 @@ impl AsPaintSource for ResolvedPattern {
         let mut scwscale = (taffine.xx * taffine.xx + taffine.xy * taffine.xy).sqrt();
         let mut schscale = (taffine.yx * taffine.yx + taffine.yy * taffine.yy).sqrt();
 
-        let pw: i32 = (pattern_width * bbwscale * scwscale) as i32;
-        let ph: i32 = (pattern_height * bbhscale * schscale) as i32;
+        let scaled_width = pattern_rect.width() * bbwscale;
+        let scaled_height = pattern_rect.height() * bbhscale;
 
-        let scaled_width = pattern_width * bbwscale;
-        let scaled_height = pattern_height * bbhscale;
+        let pw: i32 = (scaled_width * scwscale) as i32;
+        let ph: i32 = (scaled_height * schscale) as i32;
 
         if scaled_width.abs() < f64::EPSILON
             || scaled_height.abs() < f64::EPSILON
@@ -295,20 +281,20 @@ impl AsPaintSource for ResolvedPattern {
         scwscale = f64::from(pw) / scaled_width;
         schscale = f64::from(ph) / scaled_height;
 
-        let mut affine: cairo::Matrix = cairo::Matrix::identity();
+        let mut affine = cairo::Matrix::identity();
 
         // Create the pattern coordinate system
         match units {
             PatternUnits(CoordUnits::ObjectBoundingBox) => {
                 let bbrect = bbox.rect.unwrap();
                 affine.translate(
-                    bbrect.x + pattern_x * bbrect.width,
-                    bbrect.y + pattern_y * bbrect.height,
+                    bbrect.x0 + pattern_rect.x0 * bbrect.width(),
+                    bbrect.y0 + pattern_rect.y0 * bbrect.height(),
                 );
             }
 
             PatternUnits(CoordUnits::UserSpaceOnUse) => {
-                affine.translate(pattern_x, pattern_y);
+                affine.translate(pattern_rect.x0, pattern_rect.y0);
             }
         }
 
@@ -320,28 +306,23 @@ impl AsPaintSource for ResolvedPattern {
         // Create the pattern contents coordinate system
         let _params = if let Some(vbox) = vbox {
             // If there is a vbox, use that
-            let (mut x, mut y, w, h) = preserve_aspect_ratio.compute(
-                &vbox,
-                &cairo::Rectangle::new(
-                    0.0,
-                    0.0,
-                    pattern_width * bbwscale,
-                    pattern_height * bbhscale,
-                ),
-            );
+            let r =
+                preserve_aspect_ratio.compute(&vbox, &Rect::from_size(scaled_width, scaled_height));
 
-            x -= vbox.x * w / vbox.width;
-            y -= vbox.y * h / vbox.height;
+            let sw = r.width() / vbox.0.width();
+            let sh = r.height() / vbox.0.height();
+            let x = r.x0 - vbox.0.x0 * sw;
+            let y = r.y0 - vbox.0.y0 * sh;
 
-            caffine = cairo::Matrix::new(w / vbox.width, 0.0, 0.0, h / vbox.height, x, y);
+            caffine = cairo::Matrix::new(sw, 0.0, 0.0, sh, x, y);
 
-            draw_ctx.push_view_box(vbox.width, vbox.height)
+            draw_ctx.push_view_box(vbox.0.width(), vbox.0.height())
         } else if content_units == PatternContentUnits(CoordUnits::ObjectBoundingBox) {
             // If coords are in terms of the bounding box, use them
-            let bbrect = bbox.rect.unwrap();
+            let (bbw, bbh) = bbox.rect.unwrap().size();
 
             caffine = cairo::Matrix::identity();
-            caffine.scale(bbrect.width, bbrect.height);
+            caffine.scale(bbw, bbh);
 
             draw_ctx.push_view_box(1.0, 1.0)
         } else {
@@ -368,6 +349,12 @@ impl AsPaintSource for ResolvedPattern {
             .get_target()
             .create_similar(cairo::Content::ColorAlpha, pw, ph);
 
+        // FIXME: remove this once we update cairo-rs
+        let status = surface.status();
+        if status != cairo::Status::Success {
+            return Err(RenderingError::Cairo(status));
+        }
+
         let cr_pattern = cairo::Context::new(&surface);
 
         draw_ctx.set_cairo_context(&cr_pattern);
@@ -570,6 +557,17 @@ impl Children {
     }
 }
 
+impl ResolvedPattern {
+    fn get_rect(&self, values: &ComputedValues, params: &ViewParams) -> Rect {
+        let x = self.x.normalize(&values, &params);
+        let y = self.y.normalize(&values, &params);
+        let w = self.width.normalize(&values, &params);
+        let h = self.height.normalize(&values, &params);
+
+        Rect::new(x, y, x + w, y + h)
+    }
+}
+
 impl Pattern {
     fn get_unresolved(&self, node: &RsvgNode) -> Unresolved {
         let pattern = UnresolvedPattern {
@@ -587,8 +585,8 @@ impl Pattern {
 #[cfg(test)]
 mod tests {
     use super::*;
-    use markup5ever::{namespace_url, ns, QualName};
     use crate::node::{NodeData, NodeType, RsvgNode};
+    use markup5ever::{namespace_url, ns, QualName};
 
     #[test]
     fn pattern_resolved_from_defaults_is_really_resolved() {
