@@ -1,10 +1,14 @@
-use cssparser::{self, DeclarationListParser, Parser, ParserInput};
+//! CSS properties, specified values, computed values.
+
+use cssparser::{
+    self, BasicParseErrorKind, DeclarationListParser, ParseErrorKind, Parser, ParserInput, ToCss,
+};
 use markup5ever::{expanded_name, local_name, namespace_url, ns, QualName};
 use std::collections::HashSet;
 
 use crate::css::{DeclParser, Declaration};
 use crate::error::*;
-use crate::parsers::Parse;
+use crate::parsers::{Parse, ParseValue};
 use crate::property_bag::PropertyBag;
 use crate::property_defs::*;
 use crate::property_macros::Property;
@@ -223,153 +227,157 @@ pub struct ComputedValues {
 }
 
 #[cfg_attr(rustfmt, rustfmt_skip)]
-pub fn parse_property(prop_name: &QualName, input: &mut Parser, accept_shorthands: bool) -> Result<ParsedProperty, ValueErrorKind> {
+pub fn parse_property<'i>(prop_name: &QualName, input: &mut Parser<'i, '_>, accept_shorthands: bool) -> Result<ParsedProperty, ParseError<'i>> {
     // please keep these sorted
     match prop_name.expanded() {
-        expanded_name!(svg "baseline-shift") =>
+        expanded_name!("", "baseline-shift") =>
             Ok(ParsedProperty::BaselineShift(parse_input(input)?)),
 
-        expanded_name!(svg "clip-path") =>
+        expanded_name!("", "clip-path") =>
             Ok(ParsedProperty::ClipPath(parse_input(input)?)),
 
-        expanded_name!(svg "clip-rule") =>
+        expanded_name!("", "clip-rule") =>
             Ok(ParsedProperty::ClipRule(parse_input(input)?)),
 
-        expanded_name!(svg "color") =>
+        expanded_name!("", "color") =>
             Ok(ParsedProperty::Color(parse_input(input)?)),
 
-        expanded_name!(svg "color-interpolation-filters") =>
+        expanded_name!("", "color-interpolation-filters") =>
             Ok(ParsedProperty::ColorInterpolationFilters(parse_input(input)?)),
 
-        expanded_name!(svg "direction") =>
+        expanded_name!("", "direction") =>
             Ok(ParsedProperty::Direction(parse_input(input)?)),
 
-        expanded_name!(svg "display") =>
+        expanded_name!("", "display") =>
             Ok(ParsedProperty::Display(parse_input(input)?)),
 
-        expanded_name!(svg "enable-background") =>
+        expanded_name!("", "enable-background") =>
             Ok(ParsedProperty::EnableBackground(parse_input(input)?)),
 
-        expanded_name!(svg "fill") =>
+        expanded_name!("", "fill") =>
             Ok(ParsedProperty::Fill(parse_input(input)?)),
 
-        expanded_name!(svg "fill-opacity") =>
+        expanded_name!("", "fill-opacity") =>
             Ok(ParsedProperty::FillOpacity(parse_input(input)?)),
 
-        expanded_name!(svg "fill-rule") =>
+        expanded_name!("", "fill-rule") =>
             Ok(ParsedProperty::FillRule(parse_input(input)?)),
 
-        expanded_name!(svg "filter") =>
+        expanded_name!("", "filter") =>
             Ok(ParsedProperty::Filter(parse_input(input)?)),
 
-        expanded_name!(svg "flood-color") =>
+        expanded_name!("", "flood-color") =>
             Ok(ParsedProperty::FloodColor(parse_input(input)?)),
 
-        expanded_name!(svg "flood-opacity") =>
+        expanded_name!("", "flood-opacity") =>
             Ok(ParsedProperty::FloodOpacity(parse_input(input)?)),
 
-        expanded_name!(svg "font-family") =>
+        expanded_name!("", "font-family") =>
             Ok(ParsedProperty::FontFamily(parse_input(input)?)),
 
-        expanded_name!(svg "font-size") =>
+        expanded_name!("", "font-size") =>
             Ok(ParsedProperty::FontSize(parse_input(input)?)),
 
-        expanded_name!(svg "font-stretch") =>
+        expanded_name!("", "font-stretch") =>
             Ok(ParsedProperty::FontStretch(parse_input(input)?)),
 
-        expanded_name!(svg "font-style") =>
+        expanded_name!("", "font-style") =>
             Ok(ParsedProperty::FontStyle(parse_input(input)?)),
 
-        expanded_name!(svg "font-variant") =>
+        expanded_name!("", "font-variant") =>
             Ok(ParsedProperty::FontVariant(parse_input(input)?)),
 
-        expanded_name!(svg "font-weight") =>
+        expanded_name!("", "font-weight") =>
             Ok(ParsedProperty::FontWeight(parse_input(input)?)),
 
-        expanded_name!(svg "letter-spacing") =>
+        expanded_name!("", "letter-spacing") =>
             Ok(ParsedProperty::LetterSpacing(parse_input(input)?)),
 
-        expanded_name!(svg "lighting-color") =>
+        expanded_name!("", "lighting-color") =>
             Ok(ParsedProperty::LightingColor(parse_input(input)?)),
 
-        expanded_name!(svg "marker") => {
+        expanded_name!("", "marker") => {
             if accept_shorthands {
                 Ok(ParsedProperty::Marker(parse_input(input)?))
             } else {
-                Err(ValueErrorKind::UnknownProperty)
+                let loc = input.current_source_location();
+                Err(loc.new_custom_error(ValueErrorKind::UnknownProperty))
             }
         }
 
-        expanded_name!(svg "marker-end") =>
+        expanded_name!("", "marker-end") =>
             Ok(ParsedProperty::MarkerEnd(parse_input(input)?)),
 
-        expanded_name!(svg "marker-mid") =>
+        expanded_name!("", "marker-mid") =>
             Ok(ParsedProperty::MarkerMid(parse_input(input)?)),
 
-        expanded_name!(svg "marker-start") =>
+        expanded_name!("", "marker-start") =>
             Ok(ParsedProperty::MarkerStart(parse_input(input)?)),
 
-        expanded_name!(svg "mask") =>
+        expanded_name!("", "mask") =>
             Ok(ParsedProperty::Mask(parse_input(input)?)),
 
-        expanded_name!(svg "opacity") =>
+        expanded_name!("", "opacity") =>
             Ok(ParsedProperty::Opacity(parse_input(input)?)),
 
-        expanded_name!(svg "overflow") =>
+        expanded_name!("", "overflow") =>
             Ok(ParsedProperty::Overflow(parse_input(input)?)),
 
-        expanded_name!(svg "shape-rendering") =>
+        expanded_name!("", "shape-rendering") =>
             Ok(ParsedProperty::ShapeRendering(parse_input(input)?)),
 
-        expanded_name!(svg "stop-color") =>
+        expanded_name!("", "stop-color") =>
             Ok(ParsedProperty::StopColor(parse_input(input)?)),
 
-        expanded_name!(svg "stop-opacity") =>
+        expanded_name!("", "stop-opacity") =>
             Ok(ParsedProperty::StopOpacity(parse_input(input)?)),
 
-        expanded_name!(svg "stroke") =>
+        expanded_name!("", "stroke") =>
             Ok(ParsedProperty::Stroke(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-dasharray") =>
+        expanded_name!("", "stroke-dasharray") =>
             Ok(ParsedProperty::StrokeDasharray(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-dashoffset") =>
+        expanded_name!("", "stroke-dashoffset") =>
             Ok(ParsedProperty::StrokeDashoffset(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-linecap") =>
+        expanded_name!("", "stroke-linecap") =>
             Ok(ParsedProperty::StrokeLinecap(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-linejoin") =>
+        expanded_name!("", "stroke-linejoin") =>
             Ok(ParsedProperty::StrokeLinejoin(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-miterlimit") =>
+        expanded_name!("", "stroke-miterlimit") =>
             Ok(ParsedProperty::StrokeMiterlimit(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-opacity") =>
+        expanded_name!("", "stroke-opacity") =>
             Ok(ParsedProperty::StrokeOpacity(parse_input(input)?)),
 
-        expanded_name!(svg "stroke-width") =>
+        expanded_name!("", "stroke-width") =>
             Ok(ParsedProperty::StrokeWidth(parse_input(input)?)),
 
-        expanded_name!(svg "text-anchor") =>
+        expanded_name!("", "text-anchor") =>
             Ok(ParsedProperty::TextAnchor(parse_input(input)?)),
 
-        expanded_name!(svg "text-decoration") =>
+        expanded_name!("", "text-decoration") =>
             Ok(ParsedProperty::TextDecoration(parse_input(input)?)),
 
-        expanded_name!(svg "text-rendering") =>
+        expanded_name!("", "text-rendering") =>
             Ok(ParsedProperty::TextRendering(parse_input(input)?)),
 
-        expanded_name!(svg "unicode-bidi") =>
+        expanded_name!("", "unicode-bidi") =>
             Ok(ParsedProperty::UnicodeBidi(parse_input(input)?)),
 
-        expanded_name!(svg "visibility") =>
+        expanded_name!("", "visibility") =>
             Ok(ParsedProperty::Visibility(parse_input(input)?)),
 
-        expanded_name!(svg "writing-mode") =>
+        expanded_name!("", "writing-mode") =>
             Ok(ParsedProperty::WritingMode(parse_input(input)?)),
 
-        _ => Err(ValueErrorKind::UnknownProperty)
+        _ => {
+            let loc = input.current_source_location();
+            Err(loc.new_custom_error(ValueErrorKind::UnknownProperty))
+        }
     }
 }
 
@@ -518,34 +526,82 @@ impl SpecifiedValues {
         }
     }
 
-    fn parse_attribute_pair(
+    fn parse_one_presentation_attribute(
         &mut self,
         attr: QualName,
         value: &str,
-        accept_shorthands: bool,
     ) -> Result<(), NodeError> {
         let mut input = ParserInput::new(value);
         let mut parser = Parser::new(&mut input);
 
-        match parse_property(&attr, &mut parser, accept_shorthands).attribute(attr.clone()) {
+        // Presentation attributes don't accept shorthands, e.g. there is no
+        // attribute like marker="#foo" and it needs to be set in the style attribute
+        // like style="marker: #foo;".  So, pass false for accept_shorthands here.
+        match parse_property(&attr, &mut parser, false) {
             Ok(prop) => self.set_parsed_property(&prop),
-            Err(e) => {
-                // https://www.w3.org/TR/CSS2/syndata.html#unsupported-values
-                // Ignore unsupported / illegal values; don't set the whole
-                // node to be in error in that case.
 
+            // not a presentation attribute; just ignore it
+            Err(ParseError {
+                kind: ParseErrorKind::Custom(ValueErrorKind::UnknownProperty),
+                ..
+            }) => (),
+
+            // https://www.w3.org/TR/CSS2/syndata.html#unsupported-values
+            // For all the following cases, ignore illegal values; don't set the whole node to
+            // be in error in that case.
+            Err(ParseError {
+                kind: ParseErrorKind::Basic(BasicParseErrorKind::UnexpectedToken(ref t)),
+                ..
+            }) => {
+                let mut tok = String::new();
+
+                t.to_css(&mut tok).unwrap(); // FIXME: what do we do with a fmt::Error?
                 rsvg_log!(
-                    "(style property error for attribute {:?}\n    value=\"{}\"\n    {}\n    \
-                     property will be ignored)",
-                    attr,
+                    "(ignoring invalid presentation attribute {:?}\n    value=\"{}\"\n    \
+                     unexpected token '{}')",
+                    attr.expanded(),
                     value,
-                    e
+                    tok,
+                );
+            }
+
+            Err(ParseError {
+                kind: ParseErrorKind::Basic(BasicParseErrorKind::EndOfInput),
+                ..
+            }) => {
+                rsvg_log!(
+                    "(ignoring invalid presentation attribute {:?}\n    value=\"{}\"\n    \
+                     unexpected end of input)",
+                    attr.expanded(),
+                    value,
+                );
+            }
+
+            Err(ParseError {
+                kind: ParseErrorKind::Basic(_),
+                ..
+            }) => {
+                rsvg_log!(
+                    "(ignoring invalid presentation attribute {:?}\n    value=\"{}\"\n    \
+                     unexpected error)",
+                    attr.expanded(),
+                    value,
+                );
+            }
+
+            Err(ParseError {
+                kind: ParseErrorKind::Custom(ref v),
+                ..
+            }) => {
+                rsvg_log!(
+                    "(ignoring invalid presentation attribute {:?}\n    value=\"{}\"\n    {})",
+                    attr.expanded(),
+                    value,
+                    v
                 );
             }
         }
 
-        // If we didn't ignore property errors, we could just return this:
-        // ParsedProperty::parse().attribute(attr)
         Ok(())
     }
 
@@ -555,23 +611,21 @@ impl SpecifiedValues {
     ) -> Result<(), NodeError> {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "xml:lang") => {
+                expanded_name!(xml "lang") => {
                     // xml:lang is a non-presentation attribute and as such cannot have the
-                    // "inherit" value.  So, we don't call parse_attribute_pair() for it, but
-                    // rather call its parser directly.
-                    self.xml_lang =
-                        SpecifiedValue::Specified(XmlLang::parse_str(value).attribute(attr)?);
+                    // "inherit" value.  So, we don't call parse_one_presentation_attribute()
+                    // for it, but rather call its parser directly.
+                    self.xml_lang = SpecifiedValue::Specified(attr.parse(value)?);
                 }
 
-                expanded_name!(svg "xml:space") => {
+                expanded_name!(xml "space") => {
                     // xml:space is a non-presentation attribute and as such cannot have the
-                    // "inherit" value.  So, we don't call parse_attribute_pair() for it, but
-                    // rather call its parser directly.
-                    self.xml_space =
-                        SpecifiedValue::Specified(XmlSpace::parse_str(value).attribute(attr)?);
+                    // "inherit" value.  So, we don't call parse_one_presentation_attribute()
+                    // for it, but rather call its parser directly.
+                    self.xml_space = SpecifiedValue::Specified(attr.parse(value)?);
                 }
 
-                _ => self.parse_attribute_pair(attr, value, false)?,
+                _ => self.parse_one_presentation_attribute(attr, value)?,
             }
         }
 
@@ -603,7 +657,13 @@ impl SpecifiedValues {
         let mut parser = Parser::new(&mut input);
 
         DeclarationListParser::new(&mut parser, DeclParser)
-            .filter_map(Result::ok) // ignore invalid property name or value
+            .filter_map(|r| match r {
+                Ok(decl) => Some(decl),
+                Err(e) => {
+                    rsvg_log!("Invalid declaration; ignoring: {:?}", e);
+                    None
+                }
+            })
             .for_each(|decl| self.set_property_from_declaration(&decl, important_styles));
 
         Ok(())
@@ -611,7 +671,7 @@ impl SpecifiedValues {
 }
 
 // Parses the value for the type `T` of the property out of the Parser, including `inherit` values.
-fn parse_input<T>(input: &mut Parser) -> Result<SpecifiedValue<T>, <T as Parse>::Err>
+fn parse_input<'i, T>(input: &mut Parser<'i, '_>) -> Result<SpecifiedValue<T>, ParseError<'i>>
 where
     T: Property<ComputedValues> + Clone + Default + Parse,
 {
