@@ -1,73 +1,61 @@
-use cairo;
+//! Basic SVG shapes: the `path`, `polygon`, `polyline`, `line`,
+//! `rect`, `circle`, `ellipse` elements.
+
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 use std::ops::Deref;
+use std::rc::Rc;
 
 use crate::bbox::BoundingBox;
 use crate::drawing_ctx::DrawingCtx;
 use crate::error::*;
 use crate::length::*;
-use crate::marker;
 use crate::node::*;
-use crate::parsers::{CssParserExt, Parse, ParseValue};
+use crate::parsers::{optional_comma, Parse, ParseValue};
 use crate::path_builder::*;
 use crate::path_parser;
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
 use cssparser::{Parser, Token};
 
-fn render_path_builder(
-    builder: &PathBuilder,
-    draw_ctx: &mut DrawingCtx,
-    node: &RsvgNode,
-    values: &ComputedValues,
-    render_markers: bool,
-    clipping: bool,
-) -> Result<BoundingBox, RenderingError> {
-    if !builder.is_empty() {
-        let bbox = draw_ctx.with_discrete_layer(node, values, clipping, &mut |dc| {
-            let cr = dc.get_cairo_context();
+#[derive(Copy, Clone, PartialEq)]
+pub enum Markers {
+    No,
+    Yes,
+}
 
-            builder.to_cairo(&cr)?;
+pub struct Shape {
+    builder: Rc<PathBuilder>,
+    markers: Markers,
+}
 
-            if clipping {
-                cr.set_fill_rule(cairo::FillRule::from(values.clip_rule));
-                Ok(dc.empty_bbox())
-            } else {
-                cr.set_fill_rule(cairo::FillRule::from(values.fill_rule));
-                dc.stroke_and_fill(&cr, values)
-            }
-        })?;
+impl Shape {
+    fn new(builder: Rc<PathBuilder>, markers: Markers) -> Shape {
+        Shape { builder, markers }
+    }
 
-        if render_markers {
-            marker::render_markers_for_path_builder(builder, draw_ctx, values, clipping)?;
-        }
-
-        Ok(bbox)
-    } else {
-        Ok(draw_ctx.empty_bbox())
+    fn draw(
+        &self,
+        node: &RsvgNode,
+        values: &ComputedValues,
+        draw_ctx: &mut DrawingCtx,
+        clipping: bool,
+    ) -> Result<BoundingBox, RenderingError> {
+        draw_ctx.draw_path(&self.builder, node, values, self.markers, clipping)
     }
 }
 
-fn render_ellipse(
-    cx: f64,
-    cy: f64,
-    rx: f64,
-    ry: f64,
-    draw_ctx: &mut DrawingCtx,
-    node: &RsvgNode,
-    values: &ComputedValues,
-    clipping: bool,
-) -> Result<BoundingBox, RenderingError> {
+fn make_ellipse(cx: f64, cy: f64, rx: f64, ry: f64) -> PathBuilder {
+    let mut builder = PathBuilder::new();
+
     // Per the spec, rx and ry must be nonnegative
     if rx <= 0.0 || ry <= 0.0 {
-        return Ok(draw_ctx.empty_bbox());
+        return builder;
     }
 
     // 4/3 * (1-cos 45°)/sin 45° = 4/3 * sqrt(2) - 1
     let arc_magic: f64 = 0.5522847498;
 
     // approximate an ellipse using 4 Bézier curves
-    let mut builder = PathBuilder::new();
 
     builder.move_to(cx + rx, cy);
 
@@ -109,24 +97,26 @@ fn render_ellipse(
 
     builder.close_path();
 
-    render_path_builder(&builder, draw_ctx, node, values, false, clipping)
+    builder
 }
 
 #[derive(Default)]
 pub struct Path {
-    builder: PathBuilder,
+    builder: Option<Rc<PathBuilder>>,
 }
 
 impl NodeTrait for Path {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
-            if attr.expanded() == expanded_name!(svg "d") {
-                if let Err(e) = path_parser::parse_path_into_builder(value, &mut self.builder) {
+            if attr.expanded() == expanded_name!("", "d") {
+                let mut builder = PathBuilder::new();
+                if let Err(e) = path_parser::parse_path_into_builder(value, &mut builder) {
                     // FIXME: we don't propagate errors upstream, but creating a partial
                     // path is OK per the spec
 
                     rsvg_log!("could not parse path: {}", e);
                 }
+                self.builder = Some(Rc::new(builder));
             }
         }
 
@@ -140,8 +130,12 @@ impl NodeTrait for Path {
         draw_ctx: &mut DrawingCtx,
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
-        let values = cascaded.get();
-        render_path_builder(&self.builder, draw_ctx, node, values, true, clipping)
+        if let Some(builder) = self.builder.as_ref() {
+            let values = cascaded.get();
+            Shape::new(builder.clone(), Markers::Yes).draw(node, values, draw_ctx, clipping)
+        } else {
+            Ok(draw_ctx.empty_bbox())
+        }
     }
 }
 
@@ -159,15 +153,13 @@ impl Deref for Points {
 // Parse a list-of-points as for polyline and polygon elements
 // https://www.w3.org/TR/SVG/shapes.html#PointsBNF
 impl Parse for Points {
-    type Err = ValueErrorKind;
-
-    fn parse(parser: &mut Parser<'_, '_>) -> Result<Points, ValueErrorKind> {
+    fn parse<'i>(parser: &mut Parser<'i, '_>) -> Result<Points, ParseError<'i>> {
         let mut v = Vec::new();
 
         loop {
-            let x = f64::from(parser.expect_finite_number()?);
-            parser.optional_comma();
-            let y = f64::from(parser.expect_finite_number()?);
+            let x = f64::parse(parser)?;
+            optional_comma(parser);
+            let y = f64::parse(parser)?;
 
             v.push((x, y));
 
@@ -177,7 +169,7 @@ impl Parse for Points {
 
             match parser.next_including_whitespace() {
                 Ok(&Token::WhiteSpace(_)) => (),
-                _ => parser.optional_comma(),
+                _ => optional_comma(parser),
             }
         }
 
@@ -185,19 +177,10 @@ impl Parse for Points {
     }
 }
 
-fn render_poly(
-    points: Option<&Points>,
-    closed: bool,
-    node: &RsvgNode,
-    cascaded: &CascadedValues<'_>,
-    draw_ctx: &mut DrawingCtx,
-    clipping: bool,
-) -> Result<BoundingBox, RenderingError> {
-    let values = cascaded.get();
+fn make_poly(points: Option<&Points>, closed: bool) -> PathBuilder {
+    let mut builder = PathBuilder::new();
 
     if let Some(points) = points {
-        let mut builder = PathBuilder::new();
-
         for (i, &(x, y)) in points.iter().enumerate() {
             if i == 0 {
                 builder.move_to(x, y);
@@ -209,11 +192,9 @@ fn render_poly(
         if closed {
             builder.close_path();
         }
-
-        render_path_builder(&builder, draw_ctx, node, values, true, clipping)
-    } else {
-        Ok(draw_ctx.empty_bbox())
     }
+
+    builder
 }
 
 #[derive(Default)]
@@ -224,8 +205,8 @@ pub struct Polygon {
 impl NodeTrait for Polygon {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
-            if attr.expanded() == expanded_name!(svg "points") {
-                self.points = attr.parse(value.trim()).map(Some)?;
+            if attr.expanded() == expanded_name!("", "points") {
+                self.points = attr.parse(value).map(Some)?;
             }
         }
 
@@ -239,14 +220,9 @@ impl NodeTrait for Polygon {
         draw_ctx: &mut DrawingCtx,
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
-        render_poly(
-            self.points.as_ref(),
-            true,
-            node,
-            cascaded,
-            draw_ctx,
-            clipping,
-        )
+        let values = cascaded.get();
+        Shape::new(Rc::new(make_poly(self.points.as_ref(), true)), Markers::Yes)
+            .draw(node, values, draw_ctx, clipping)
     }
 }
 
@@ -258,8 +234,8 @@ pub struct Polyline {
 impl NodeTrait for Polyline {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
-            if attr.expanded() == expanded_name!(svg "points") {
-                self.points = attr.parse(value.trim()).map(Some)?;
+            if attr.expanded() == expanded_name!("", "points") {
+                self.points = attr.parse(value).map(Some)?;
             }
         }
 
@@ -273,14 +249,12 @@ impl NodeTrait for Polyline {
         draw_ctx: &mut DrawingCtx,
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
-        render_poly(
-            self.points.as_ref(),
-            false,
-            node,
-            cascaded,
-            draw_ctx,
-            clipping,
+        let values = cascaded.get();
+        Shape::new(
+            Rc::new(make_poly(self.points.as_ref(), false)),
+            Markers::Yes,
         )
+        .draw(node, values, draw_ctx, clipping)
     }
 }
 
@@ -296,10 +270,10 @@ impl NodeTrait for Line {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "x1") => self.x1 = attr.parse(value)?,
-                expanded_name!(svg "y1") => self.y1 = attr.parse(value)?,
-                expanded_name!(svg "x2") => self.x2 = attr.parse(value)?,
-                expanded_name!(svg "y2") => self.y2 = attr.parse(value)?,
+                expanded_name!("", "x1") => self.x1 = attr.parse(value)?,
+                expanded_name!("", "y1") => self.y1 = attr.parse(value)?,
+                expanded_name!("", "x2") => self.x2 = attr.parse(value)?,
+                expanded_name!("", "y2") => self.y2 = attr.parse(value)?,
                 _ => (),
             }
         }
@@ -315,7 +289,16 @@ impl NodeTrait for Line {
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
         let values = cascaded.get();
+        Shape::new(
+            Rc::new(self.make_path_builder(values, draw_ctx)),
+            Markers::Yes,
+        )
+        .draw(node, values, draw_ctx, clipping)
+    }
+}
 
+impl Line {
+    fn make_path_builder(&self, values: &ComputedValues, draw_ctx: &mut DrawingCtx) -> PathBuilder {
         let mut builder = PathBuilder::new();
 
         let params = draw_ctx.get_view_params();
@@ -328,7 +311,7 @@ impl NodeTrait for Line {
         builder.move_to(x1, y1);
         builder.line_to(x2, y2);
 
-        render_path_builder(&builder, draw_ctx, node, values, true, clipping)
+        builder
     }
 }
 
@@ -348,20 +331,22 @@ impl NodeTrait for Rect {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "x") => self.x = attr.parse(value)?,
-                expanded_name!(svg "y") => self.y = attr.parse(value)?,
-                expanded_name!(svg "width") => {
-                    self.w = attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?
+                expanded_name!("", "x") => self.x = attr.parse(value)?,
+                expanded_name!("", "y") => self.y = attr.parse(value)?,
+                expanded_name!("", "width") => {
+                    self.w =
+                        attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?
                 }
-                expanded_name!(svg "height") => {
-                    self.h = attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?
+                expanded_name!("", "height") => {
+                    self.h =
+                        attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?
                 }
-                expanded_name!(svg "rx") => {
+                expanded_name!("", "rx") => {
                     self.rx = attr
                         .parse_and_validate(value, Length::<Horizontal>::check_nonnegative)
                         .map(Some)?
                 }
-                expanded_name!(svg "ry") => {
+                expanded_name!("", "ry") => {
                     self.ry = attr
                         .parse_and_validate(value, Length::<Vertical>::check_nonnegative)
                         .map(Some)?
@@ -381,7 +366,16 @@ impl NodeTrait for Rect {
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
         let values = cascaded.get();
+        Shape::new(
+            Rc::new(self.make_path_builder(values, draw_ctx)),
+            Markers::No,
+        )
+        .draw(node, values, draw_ctx, clipping)
+    }
+}
 
+impl Rect {
+    fn make_path_builder(&self, values: &ComputedValues, draw_ctx: &mut DrawingCtx) -> PathBuilder {
         let params = draw_ctx.get_view_params();
 
         let x = self.x.normalize(values, &params);
@@ -414,14 +408,16 @@ impl NodeTrait for Rect {
             }
         }
 
+        let mut builder = PathBuilder::new();
+
         // Per the spec, w,h must be >= 0
         if w <= 0.0 || h <= 0.0 {
-            return Ok(draw_ctx.empty_bbox());
+            return builder;
         }
 
         // ... and rx,ry must be nonnegative
         if rx < 0.0 || ry < 0.0 {
-            return Ok(draw_ctx.empty_bbox());
+            return builder;
         }
 
         let half_w = w / 2.0;
@@ -441,8 +437,6 @@ impl NodeTrait for Rect {
             rx = 0.0;
         }
 
-        let mut builder = PathBuilder::new();
-
         if rx == 0.0 {
             // Easy case, no rounded corners
             builder.move_to(x, y);
@@ -549,7 +543,7 @@ impl NodeTrait for Rect {
             builder.close_path();
         }
 
-        render_path_builder(&builder, draw_ctx, node, values, false, clipping)
+        builder
     }
 }
 
@@ -564,9 +558,9 @@ impl NodeTrait for Circle {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "cx") => self.cx = attr.parse(value)?,
-                expanded_name!(svg "cy") => self.cy = attr.parse(value)?,
-                expanded_name!(svg "r") => {
+                expanded_name!("", "cx") => self.cx = attr.parse(value)?,
+                expanded_name!("", "cy") => self.cy = attr.parse(value)?,
+                expanded_name!("", "r") => {
                     self.r = attr.parse_and_validate(value, Length::<Both>::check_nonnegative)?
                 }
                 _ => (),
@@ -584,14 +578,23 @@ impl NodeTrait for Circle {
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
         let values = cascaded.get();
+        Shape::new(
+            Rc::new(self.make_path_builder(values, draw_ctx)),
+            Markers::No,
+        )
+        .draw(node, values, draw_ctx, clipping)
+    }
+}
 
+impl Circle {
+    fn make_path_builder(&self, values: &ComputedValues, draw_ctx: &mut DrawingCtx) -> PathBuilder {
         let params = draw_ctx.get_view_params();
 
         let cx = self.cx.normalize(values, &params);
         let cy = self.cy.normalize(values, &params);
         let r = self.r.normalize(values, &params);
 
-        render_ellipse(cx, cy, r, r, draw_ctx, node, values, clipping)
+        make_ellipse(cx, cy, r, r)
     }
 }
 
@@ -607,13 +610,15 @@ impl NodeTrait for Ellipse {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "cx") => self.cx = attr.parse(value)?,
-                expanded_name!(svg "cy") => self.cy = attr.parse(value)?,
-                expanded_name!(svg "rx") => {
-                    self.rx = attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?
+                expanded_name!("", "cx") => self.cx = attr.parse(value)?,
+                expanded_name!("", "cy") => self.cy = attr.parse(value)?,
+                expanded_name!("", "rx") => {
+                    self.rx =
+                        attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?
                 }
-                expanded_name!(svg "ry") => {
-                    self.ry = attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?
+                expanded_name!("", "ry") => {
+                    self.ry =
+                        attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?
                 }
                 _ => (),
             }
@@ -630,7 +635,16 @@ impl NodeTrait for Ellipse {
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
         let values = cascaded.get();
+        Shape::new(
+            Rc::new(self.make_path_builder(values, draw_ctx)),
+            Markers::No,
+        )
+        .draw(node, values, draw_ctx, clipping)
+    }
+}
 
+impl Ellipse {
+    fn make_path_builder(&self, values: &ComputedValues, draw_ctx: &mut DrawingCtx) -> PathBuilder {
         let params = draw_ctx.get_view_params();
 
         let cx = self.cx.normalize(values, &params);
@@ -638,7 +652,7 @@ impl NodeTrait for Ellipse {
         let rx = self.rx.normalize(values, &params);
         let ry = self.ry.normalize(values, &params);
 
-        render_ellipse(cx, cy, rx, ry, draw_ctx, node, values, clipping)
+        make_ellipse(cx, cy, rx, ry)
     }
 }
 
