@@ -1,20 +1,20 @@
-use cairo::Rectangle;
+//! Structural elements in SVG: the `g`, `switch`, `svg`, `use`, `symbol`, `clip_path`, `mask`, `link` elements.
+
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
 
 use crate::allowed_url::Fragment;
 use crate::aspect_ratio::*;
 use crate::bbox::BoundingBox;
+use crate::coord_units::CoordUnits;
 use crate::dpi::Dpi;
 use crate::drawing_ctx::{ClipMode, DrawingCtx, ViewParams};
-use crate::error::{AcquireError, AttributeResultExt, RenderingError};
-use crate::float_eq_cairo::ApproxEqCairo;
+use crate::error::*;
 use crate::length::*;
 use crate::node::*;
 use crate::parsers::{Parse, ParseValue};
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
-use crate::property_defs::Overflow;
-use crate::rect::RectangleExt;
+use crate::rect::Rect;
 use crate::viewbox::*;
 
 #[derive(Default)]
@@ -108,7 +108,7 @@ impl Svg {
 
         match (w, h, self.vbox) {
             (w, h, Some(vbox)) => {
-                let params = ViewParams::new(dpi.x(), dpi.y(), vbox.width, vbox.height);
+                let params = ViewParams::new(dpi.x(), dpi.y(), vbox.0.width(), vbox.0.height());
 
                 Some((
                     w.normalize(values, &params).round() as i32,
@@ -162,15 +162,15 @@ impl Svg {
         (x, y, w, h)
     }
 
-    fn get_viewport(&self, values: &ComputedValues, params: &ViewParams) -> Rectangle {
+    fn get_viewport(&self, values: &ComputedValues, params: &ViewParams) -> Rect {
         let (x, y, w, h) = self.get_unnormalized_viewport();
 
-        Rectangle::new(
-            x.normalize(values, &params),
-            y.normalize(values, &params),
-            w.normalize(values, &params),
-            h.normalize(values, &params),
-        )
+        let nx = x.normalize(values, &params);
+        let ny = y.normalize(values, &params);
+        let nw = w.normalize(values, &params);
+        let nh = h.normalize(values, &params);
+
+        Rect::new(nx, ny, nx + nw, ny + nh)
     }
 }
 
@@ -182,20 +182,21 @@ impl NodeTrait for Svg {
 
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "preserveAspectRatio") => {
+                expanded_name!("", "preserveAspectRatio") => {
                     self.preserve_aspect_ratio = attr.parse(value)?
                 }
-                expanded_name!(svg "x") if is_inner_svg => self.x = Some(attr.parse(value)?),
-                expanded_name!(svg "y") if is_inner_svg => self.y = Some(attr.parse(value)?),
-                expanded_name!(svg "width") => {
-                    self.w =
-                        Some(attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?)
+                expanded_name!("", "x") if is_inner_svg => self.x = Some(attr.parse(value)?),
+                expanded_name!("", "y") if is_inner_svg => self.y = Some(attr.parse(value)?),
+                expanded_name!("", "width") => {
+                    self.w = Some(
+                        attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?,
+                    )
                 }
-                expanded_name!(svg "height") => {
+                expanded_name!("", "height") => {
                     self.h =
                         Some(attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?)
                 }
-                expanded_name!(svg "viewBox") => self.vbox = attr.parse(value).map(Some)?,
+                expanded_name!("", "viewBox") => self.vbox = attr.parse(value).map(Some)?,
                 _ => (),
             }
         }
@@ -244,19 +245,17 @@ impl NodeTrait for Svg {
                 // Use our viewBox if available, or try to derive one from
                 // the intrinsic dimensions.
                 self.vbox.or_else(|| {
-                    Some(ViewBox {
-                        x: 0.0,
-                        y: 0.0,
-                        width: svg_viewport.width,
-                        height: svg_viewport.height,
-                    })
+                    Some(ViewBox(Rect::from_size(
+                        svg_viewport.width(),
+                        svg_viewport.height(),
+                    )))
                 }),
             )
         };
 
         draw_ctx.with_discrete_layer(node, values, clipping, &mut |dc| {
             let _params =
-                dc.push_new_viewport(vbox, &viewport, self.preserve_aspect_ratio, clip_mode);
+                dc.push_new_viewport(vbox, viewport, self.preserve_aspect_ratio, clip_mode);
 
             node.draw_children(cascaded, dc, clipping)
         })
@@ -272,6 +271,33 @@ pub struct Use {
     h: Option<Length<Vertical>>,
 }
 
+impl Use {
+    pub fn get_link(&self) -> Option<&Fragment> {
+        self.link.as_ref()
+    }
+
+    pub fn get_rect(&self, values: &ComputedValues, params: &ViewParams) -> Rect {
+        let x = self.x.normalize(values, &params);
+        let y = self.y.normalize(values, &params);
+
+        // If attributes ‘width’ and/or ‘height’ are not specified,
+        // [...] use values of '100%' for these attributes.
+        // From https://www.w3.org/TR/SVG/struct.html#UseElement in
+        // "If the ‘use’ element references a ‘symbol’ element"
+
+        let w = self
+            .w
+            .unwrap_or_else(|| Length::<Horizontal>::parse_str("100%").unwrap())
+            .normalize(values, &params);
+        let h = self
+            .h
+            .unwrap_or_else(|| Length::<Vertical>::parse_str("100%").unwrap())
+            .normalize(values, &params);
+
+        Rect::new(x, y, x + w, y + h)
+    }
+}
+
 impl NodeTrait for Use {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
@@ -279,14 +305,14 @@ impl NodeTrait for Use {
                 expanded_name!(xlink "href") => {
                     self.link = Some(Fragment::parse(value).attribute(attr)?)
                 }
-                expanded_name!(svg "x") => self.x = attr.parse(value)?,
-                expanded_name!(svg "y") => self.y = attr.parse(value)?,
-                expanded_name!(svg "width") => {
+                expanded_name!("", "x") => self.x = attr.parse(value)?,
+                expanded_name!("", "y") => self.y = attr.parse(value)?,
+                expanded_name!("", "width") => {
                     self.w = attr
                         .parse_and_validate(value, Length::<Horizontal>::check_nonnegative)
                         .map(Some)?
                 }
-                expanded_name!(svg "height") => {
+                expanded_name!("", "height") => {
                     self.h = attr
                         .parse_and_validate(value, Length::<Vertical>::check_nonnegative)
                         .map(Some)?
@@ -305,117 +331,7 @@ impl NodeTrait for Use {
         draw_ctx: &mut DrawingCtx,
         clipping: bool,
     ) -> Result<BoundingBox, RenderingError> {
-        let values = cascaded.get();
-
-        if self.link.is_none() {
-            return Ok(draw_ctx.empty_bbox());
-        }
-
-        let link = self.link.as_ref().unwrap();
-
-        // <use> is an element that is used directly, unlike
-        // <pattern>, which is used through a fill="url(#...)"
-        // reference.  However, <use> will always reference another
-        // element, potentially itself or an ancestor of itself (or
-        // another <use> which references the first one, etc.).  So,
-        // we acquire the <use> element itself so that circular
-        // references can be caught.
-        let _self_acquired = draw_ctx.acquire_node_ref(node).map_err(|e| {
-            if let AcquireError::CircularReference(_) = e {
-                rsvg_log!("circular reference in element {}", node);
-                RenderingError::CircularReference
-            } else {
-                unreachable!();
-            }
-        })?;
-
-        let acquired = match draw_ctx.acquire_node(link, &[]) {
-            Ok(acquired) => acquired,
-
-            Err(AcquireError::CircularReference(node)) => {
-                rsvg_log!("circular reference in element {}", node);
-                return Err(RenderingError::CircularReference);
-            }
-
-            Err(AcquireError::MaxReferencesExceeded) => {
-                return Err(RenderingError::InstancingLimit);
-            }
-
-            Err(AcquireError::InvalidLinkType(_)) => unreachable!(),
-
-            Err(AcquireError::LinkNotFound(fragment)) => {
-                rsvg_log!("element {} references nonexistent \"{}\"", node, fragment);
-                return Ok(draw_ctx.empty_bbox());
-            }
-        };
-
-        let child = acquired.get();
-
-        let params = draw_ctx.get_view_params();
-
-        let nx = self.x.normalize(values, &params);
-        let ny = self.y.normalize(values, &params);
-
-        // If attributes ‘width’ and/or ‘height’ are not specified,
-        // [...] use values of '100%' for these attributes.
-        // From https://www.w3.org/TR/SVG/struct.html#UseElement in
-        // "If the ‘use’ element references a ‘symbol’ element"
-
-        let nw = self
-            .w
-            .unwrap_or_else(|| Length::<Horizontal>::parse_str("100%").unwrap())
-            .normalize(values, &params);
-        let nh = self
-            .h
-            .unwrap_or_else(|| Length::<Vertical>::parse_str("100%").unwrap())
-            .normalize(values, &params);
-
-        // width or height set to 0 disables rendering of the element
-        // https://www.w3.org/TR/SVG/struct.html#UseElementWidthAttribute
-        if nw.approx_eq_cairo(0.0) || nh.approx_eq_cairo(0.0) {
-            return Ok(draw_ctx.empty_bbox());
-        }
-
-        let viewport = Rectangle::new(nx, ny, nw, nh);
-
-        if child.borrow().get_type() != NodeType::Symbol {
-            let cr = draw_ctx.get_cairo_context();
-            cr.translate(viewport.x, viewport.y);
-
-            draw_ctx.with_discrete_layer(node, values, clipping, &mut |dc| {
-                dc.draw_node_from_stack(
-                    &CascadedValues::new_from_values(&child, values),
-                    &child,
-                    clipping,
-                )
-            })
-        } else {
-            let node_data = child.borrow();
-            let symbol = node_data.get_impl::<Symbol>();
-
-            let clip_mode = if !values.is_overflow()
-                || (values.overflow == Overflow::Visible && child.borrow().is_overflow())
-            {
-                Some(ClipMode::ClipToVbox)
-            } else {
-                None
-            };
-
-            draw_ctx.with_discrete_layer(node, values, clipping, &mut |dc| {
-                let _params = dc.push_new_viewport(
-                    symbol.vbox,
-                    &viewport,
-                    symbol.preserve_aspect_ratio,
-                    clip_mode,
-                );
-
-                child.draw_children(
-                    &CascadedValues::new_from_values(&child, values),
-                    dc,
-                    clipping,
-                )
-            })
-        }
+        draw_ctx.draw_from_use_node(node, cascaded, clipping)
     }
 }
 
@@ -425,14 +341,24 @@ pub struct Symbol {
     vbox: Option<ViewBox>,
 }
 
+impl Symbol {
+    pub fn get_viewbox(&self) -> Option<ViewBox> {
+        self.vbox
+    }
+
+    pub fn get_preserve_aspect_ratio(&self) -> AspectRatio {
+        self.preserve_aspect_ratio
+    }
+}
+
 impl NodeTrait for Symbol {
     fn set_atts(&mut self, _parent: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "preserveAspectRatio") => {
+                expanded_name!("", "preserveAspectRatio") => {
                     self.preserve_aspect_ratio = attr.parse(value)?
                 }
-                expanded_name!(svg "viewBox") => self.vbox = attr.parse(value).map(Some)?,
+                expanded_name!("", "viewBox") => self.vbox = attr.parse(value).map(Some)?,
                 _ => (),
             }
         }
@@ -444,3 +370,136 @@ impl NodeTrait for Symbol {
         true
     }
 }
+
+coord_units!(ClipPathUnits, CoordUnits::UserSpaceOnUse);
+
+#[derive(Default)]
+pub struct ClipPath {
+    units: ClipPathUnits,
+}
+
+impl ClipPath {
+    pub fn get_units(&self) -> CoordUnits {
+        CoordUnits::from(self.units)
+    }
+}
+
+impl NodeTrait for ClipPath {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        for (attr, value) in pbag.iter() {
+            match attr.expanded() {
+                expanded_name!("", "clipPathUnits") => self.units = attr.parse(value)?,
+                _ => (),
+            }
+        }
+
+        Ok(())
+    }
+}
+
+coord_units!(MaskUnits, CoordUnits::ObjectBoundingBox);
+coord_units!(MaskContentUnits, CoordUnits::UserSpaceOnUse);
+
+pub struct Mask {
+    x: Length<Horizontal>,
+    y: Length<Vertical>,
+    width: Length<Horizontal>,
+    height: Length<Vertical>,
+
+    units: MaskUnits,
+    content_units: MaskContentUnits,
+}
+
+impl Default for Mask {
+    fn default() -> Mask {
+        Mask {
+            // these values are per the spec
+            x: Length::<Horizontal>::parse_str("-10%").unwrap(),
+            y: Length::<Vertical>::parse_str("-10%").unwrap(),
+            width: Length::<Horizontal>::parse_str("120%").unwrap(),
+            height: Length::<Vertical>::parse_str("120%").unwrap(),
+
+            units: MaskUnits::default(),
+            content_units: MaskContentUnits::default(),
+        }
+    }
+}
+
+impl Mask {
+    pub fn get_units(&self) -> CoordUnits {
+        CoordUnits::from(self.content_units)
+    }
+
+    pub fn get_content_units(&self) -> CoordUnits {
+        CoordUnits::from(self.content_units)
+    }
+
+    pub fn get_rect(&self, values: &ComputedValues, params: &ViewParams) -> Rect {
+        let x = self.x.normalize(&values, &params);
+        let y = self.y.normalize(&values, &params);
+        let w = self.width.normalize(&values, &params);
+        let h = self.height.normalize(&values, &params);
+
+        Rect::new(x, y, x + w, y + h)
+    }
+}
+
+impl NodeTrait for Mask {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        for (attr, value) in pbag.iter() {
+            match attr.expanded() {
+                expanded_name!("", "x") => self.x = attr.parse(value)?,
+                expanded_name!("", "y") => self.y = attr.parse(value)?,
+                expanded_name!("", "width") => {
+                    self.width =
+                        attr.parse_and_validate(value, Length::<Horizontal>::check_nonnegative)?
+                }
+                expanded_name!("", "height") => {
+                    self.height =
+                        attr.parse_and_validate(value, Length::<Vertical>::check_nonnegative)?
+                }
+                expanded_name!("", "maskUnits") => self.units = attr.parse(value)?,
+                expanded_name!("", "maskContentUnits") => self.content_units = attr.parse(value)?,
+                _ => (),
+            }
+        }
+
+        Ok(())
+    }
+}
+
+#[derive(Default)]
+pub struct Link {
+    link: Option<String>,
+}
+
+impl NodeTrait for Link {
+    fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
+        for (attr, value) in pbag.iter() {
+            match attr.expanded() {
+                expanded_name!(xlink "href") => self.link = Some(value.to_owned()),
+                _ => (),
+            }
+        }
+
+        Ok(())
+    }
+
+    fn draw(
+        &self,
+        node: &RsvgNode,
+        cascaded: &CascadedValues<'_>,
+        draw_ctx: &mut DrawingCtx,
+        clipping: bool,
+    ) -> Result<BoundingBox, RenderingError> {
+        let cascaded = CascadedValues::new(cascaded, node);
+        let values = cascaded.get();
+
+        draw_ctx.with_discrete_layer(node, values, clipping, &mut |dc| match self.link.as_ref() {
+            Some(l) if !l.is_empty() => {
+                dc.with_link_tag(l, &mut |dc| node.draw_children(&cascaded, dc, clipping))
+            }
+            _ => node.draw_children(&cascaded, dc, clipping),
+        })
+    }
+}
