@@ -1,14 +1,14 @@
+//! Text elements: `text`, `tspan`, `tref`.
+
 use glib::translate::*;
 use markup5ever::{expanded_name, local_name, namespace_url, ns};
-use pango::{self, FontMapExt};
-use pango_sys;
-use pangocairo;
+use pango::FontMapExt;
 use std::cell::RefCell;
 
 use crate::allowed_url::Fragment;
 use crate::bbox::BoundingBox;
 use crate::drawing_ctx::DrawingCtx;
-use crate::error::{AttributeResultExt, RenderingError};
+use crate::error::*;
 use crate::float_eq_cairo::ApproxEqCairo;
 use crate::font_props::FontWeightSpec;
 use crate::length::*;
@@ -17,17 +17,10 @@ use crate::parsers::ParseValue;
 use crate::properties::ComputedValues;
 use crate::property_bag::PropertyBag;
 use crate::property_defs::{
-    Direction,
-    FontStretch,
-    FontStyle,
-    FontVariant,
-    TextAnchor,
-    TextRendering,
-    UnicodeBidi,
-    WritingMode,
-    XmlLang,
-    XmlSpace,
+    Direction, FontStretch, FontStyle, FontVariant, TextAnchor, TextRendering, UnicodeBidi,
+    WritingMode, XmlLang, XmlSpace,
 };
+use crate::rect::Rect;
 use crate::space::{xml_space_normalize, NormalizeDefault, XmlSpaceNormalize};
 
 /// An absolutely-positioned array of `Span`s
@@ -283,7 +276,7 @@ impl PositionedSpan {
             let affine = cr.get_matrix();
 
             let gravity = self.layout.get_context().unwrap().get_gravity();
-            let bbox = self.compute_text_bbox(&affine, gravity);
+            let bbox = self.compute_text_bbox(affine, gravity);
             if bbox.is_none() {
                 return Ok(dc.empty_bbox());
             }
@@ -350,7 +343,9 @@ impl PositionedSpan {
                     pangocairo::functions::layout_path(&cr, &self.layout);
 
                     if !clipping {
-                        let ib = BoundingBox::new(&affine).with_ink_extents(cr.stroke_extents());
+                        let (x0, y0, x1, y1) = cr.stroke_extents();
+                        let r = Rect::new(x0, y0, x1, y1);
+                        let ib = BoundingBox::new().with_affine(affine).with_ink_rect(r);
                         cr.stroke();
                         bbox.insert(&ib);
                     }
@@ -363,7 +358,7 @@ impl PositionedSpan {
 
     fn compute_text_bbox(
         &self,
-        affine: &cairo::Matrix,
+        affine: cairo::Matrix,
         gravity: pango::Gravity,
     ) -> Option<BoundingBox> {
         let (ink, _) = self.layout.get_extents();
@@ -378,25 +373,27 @@ impl PositionedSpan {
         let ink_height = f64::from(ink.height);
         let pango_scale = f64::from(pango::SCALE);
 
-        let rect = if gravity_is_vertical(gravity) {
-            cairo::Rectangle {
-                x: x + (ink_x - ink_height) / pango_scale,
-                y: y + ink_y / pango_scale,
-                width: ink_height / pango_scale,
-                height: ink_width / pango_scale,
-            }
+        let (x, y, w, h) = if gravity_is_vertical(gravity) {
+            (
+                x + (ink_x - ink_height) / pango_scale,
+                y + ink_y / pango_scale,
+                ink_height / pango_scale,
+                ink_width / pango_scale,
+            )
         } else {
-            cairo::Rectangle {
-                x: x + ink_x / pango_scale,
-                y: y + ink_y / pango_scale,
-                width: ink_width / pango_scale,
-                height: ink_height / pango_scale,
-            }
+            (
+                x + ink_x / pango_scale,
+                y + ink_y / pango_scale,
+                ink_width / pango_scale,
+                ink_height / pango_scale,
+            )
         };
 
-        let bbox = BoundingBox::new(affine)
-            .with_rect(Some(rect))
-            .with_ink_rect(Some(rect));
+        let r = Rect::new(x, y, x + w, y + h);
+        let bbox = BoundingBox::new()
+            .with_affine(affine)
+            .with_rect(r)
+            .with_ink_rect(r);
 
         Some(bbox)
     }
@@ -598,10 +595,10 @@ impl NodeTrait for Text {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "x") => self.x = attr.parse(value)?,
-                expanded_name!(svg "y") => self.y = attr.parse(value)?,
-                expanded_name!(svg "dx") => self.dx = attr.parse(value).map(Some)?,
-                expanded_name!(svg "dy") => self.dy = attr.parse(value).map(Some)?,
+                expanded_name!("", "x") => self.x = attr.parse(value)?,
+                expanded_name!("", "y") => self.y = attr.parse(value)?,
+                expanded_name!("", "dx") => self.dx = attr.parse(value).map(Some)?,
+                expanded_name!("", "dy") => self.dy = attr.parse(value).map(Some)?,
                 _ => (),
             }
         }
@@ -758,10 +755,10 @@ impl NodeTrait for TSpan {
     fn set_atts(&mut self, _: Option<&RsvgNode>, pbag: &PropertyBag<'_>) -> NodeResult {
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                expanded_name!(svg "x") => self.x = attr.parse(value).map(Some)?,
-                expanded_name!(svg "y") => self.y = attr.parse(value).map(Some)?,
-                expanded_name!(svg "dx") => self.dx = attr.parse(value).map(Some)?,
-                expanded_name!(svg "dy") => self.dy = attr.parse(value).map(Some)?,
+                expanded_name!("", "x") => self.x = attr.parse(value).map(Some)?,
+                expanded_name!("", "y") => self.y = attr.parse(value).map(Some)?,
+                expanded_name!("", "dx") => self.dx = attr.parse(value).map(Some)?,
+                expanded_name!("", "dy") => self.dy = attr.parse(value).map(Some)?,
                 _ => (),
             }
         }
