@@ -1,8 +1,11 @@
-use crate::xml_rs::{reader::XmlEvent, ParserConfig};
+//! The main XML parser.
+
 use encoding::label::encoding_from_whatwg_label;
 use encoding::DecoderTrap;
 use libc;
-use markup5ever::{expanded_name, local_name, namespace_url, ns, ExpandedName, LocalName, Namespace, QualName};
+use markup5ever::{
+    expanded_name, local_name, namespace_url, ns, ExpandedName, LocalName, Namespace, QualName,
+};
 use std::cell::RefCell;
 use std::collections::HashMap;
 use std::rc::{Rc, Weak};
@@ -17,7 +20,8 @@ use crate::node::{NodeType, RsvgNode};
 use crate::property_bag::PropertyBag;
 use crate::style::{Style, StyleType};
 use crate::text::NodeChars;
-use crate::xml2_load::{ParseFromStreamError, Xml2Parser};
+use crate::xml2_load::Xml2Parser;
+use crate::xml_rs::{reader::XmlEvent, ParserConfig};
 
 #[derive(Clone)]
 enum Context {
@@ -43,7 +47,7 @@ enum Context {
     XIncludeFallback(XIncludeContext),
 
     // An XML parsing error was found.  We will no-op upon any further XML events.
-    FatalError(ParseFromStreamError),
+    FatalError(LoadingError),
 }
 
 #[derive(Clone)]
@@ -140,7 +144,7 @@ impl XmlState {
         }
     }
 
-    fn check_last_error(&self) -> Result<(), ParseFromStreamError> {
+    fn check_last_error(&self) -> Result<(), LoadingError> {
         let inner = self.inner.borrow();
 
         match inner.context() {
@@ -151,7 +155,7 @@ impl XmlState {
 
     fn check_limits(&self) -> Result<(), ()> {
         if self.inner.borrow().num_loaded_elements > MAX_LOADED_ELEMENTS {
-            self.error(ParseFromStreamError::XmlParseError(format!(
+            self.error(LoadingError::XmlParseError(format!(
                 "cannot load more than {} XML elements",
                 MAX_LOADED_ELEMENTS
             )));
@@ -269,19 +273,22 @@ impl XmlState {
                     .append_stylesheet_from_xml_processing_instruction(alternate, type_, &href)
                     .is_err()
                 {
-                    rsvg_log!("invalid xml-stylesheet {} in XML processing instruction", href);
+                    rsvg_log!(
+                        "invalid xml-stylesheet {} in XML processing instruction",
+                        href
+                    );
                 }
             } else {
                 rsvg_log!("xml-stylesheet processing instruction does not have href; ignoring");
             }
         } else {
-            self.error(ParseFromStreamError::XmlParseError(String::from(
+            self.error(LoadingError::XmlParseError(String::from(
                 "invalid processing instruction data in xml-stylesheet",
             )));
         }
     }
 
-    pub fn error(&self, e: ParseFromStreamError) {
+    pub fn error(&self, e: LoadingError) {
         self.inner
             .borrow_mut()
             .context_stack
@@ -361,7 +368,8 @@ impl XmlState {
             .unwrap_or(StyleType::TextCss);
 
         if style_type == StyleType::TextCss {
-            let stylesheet_text = current_node.children()
+            let stylesheet_text = current_node
+                .children()
                 .map(|child| {
                     let child_borrow = child.borrow();
 
@@ -388,11 +396,13 @@ impl XmlState {
         let mut parse = None;
         let mut encoding = None;
 
+        let ln_parse = LocalName::from("parse");
+
         for (attr, value) in pbag.iter() {
             match attr.expanded() {
-                ref n if *n == xinclude_name!("href") => href = Some(value),
-                ref n if *n == xinclude_name!("parse") => parse = Some(value),
-                ref n if *n == xinclude_name!("encoding") => encoding = Some(value),
+                expanded_name!("", "href") => href = Some(value),
+                ref v if *v == ExpandedName { ns: &ns!(), local: &ln_parse } => parse = Some(value),
+                expanded_name!("", "encoding") => encoding = Some(value),
                 _ => (),
             }
         }
@@ -401,7 +411,7 @@ impl XmlState {
             Ok(()) => false,
             Err(AcquireError::ResourceError) => true,
             Err(AcquireError::FatalError(s)) => {
-                return Context::FatalError(ParseFromStreamError::XmlParseError(s))
+                return Context::FatalError(LoadingError::XmlParseError(s))
             }
         };
 
@@ -539,11 +549,12 @@ impl XmlState {
 
         // FIXME: pass a cancellable
         self.parse_from_stream(&stream, None).map_err(|e| match e {
-            ParseFromStreamError::CouldNotCreateXmlParser => {
+            LoadingError::CouldNotCreateXmlParser => {
                 AcquireError::FatalError(String::from("could not create XML parser"))
             }
-            ParseFromStreamError::IoError(_) => AcquireError::ResourceError,
-            ParseFromStreamError::XmlParseError(s) => AcquireError::FatalError(s),
+            LoadingError::Glib(_) => AcquireError::ResourceError,
+            LoadingError::XmlParseError(s) => AcquireError::FatalError(s),
+            _ => AcquireError::FatalError(String::from("unknown error")),
         })
     }
 
@@ -555,7 +566,7 @@ impl XmlState {
         &self,
         stream: &gio::InputStream,
         cancellable: Option<&gio::Cancellable>,
-    ) -> Result<(), ParseFromStreamError> {
+    ) -> Result<(), LoadingError> {
         let strong = self
             .inner
             .borrow()
@@ -644,8 +655,7 @@ pub fn xml_load_from_possibly_compressed_stream(
 
     state.inner.borrow_mut().weak = Some(Rc::downgrade(&state));
 
-    let stream =
-        get_input_stream_for_loading(stream, cancellable).map_err(ParseFromStreamError::IoError)?;
+    let stream = get_input_stream_for_loading(stream, cancellable)?;
 
     state.build_document(&stream, cancellable)
 }
